@@ -295,14 +295,25 @@ static pj_bool_t jb_framelist_get(jb_framelist_t *framelist,
 		if (bit_info)
 		    *bit_info = 0;
 	    } else {
+		pj_size_t frm_size = framelist->content_len[framelist->head];
+		pj_size_t max_size = size? *size : frm_size;
+		pj_size_t copy_size = PJ_MIN(max_size, frm_size);
+
+		/* Buffer size should not be smaller than frame size. */
+		if (max_size < frm_size) {
+		    pj_assert(!"Buffer too small");
+		    PJ_LOG(4, (THIS_FILE, "Warning: buffer too small for the "
+					  "retrieved frame!"));
+		}
+
 		pj_memcpy(frame,
 			  framelist->content +
 			  framelist->head * framelist->frame_size,
-			  framelist->frame_size);
+			  copy_size);
 		*p_type = (pjmedia_jb_frame_type)
 			  framelist->frame_type[framelist->head];
 		if (size)
-		    *size   = framelist->content_len[framelist->head];
+		    *size = copy_size;
 		if (bit_info)
 		    *bit_info = framelist->bit_info[framelist->head];
 	    }
@@ -585,6 +596,20 @@ PJ_DEF(pj_status_t) pjmedia_jbuf_create(pj_pool_t *pool,
 }
 
 
+PJ_DEF(pj_status_t) pjmedia_jbuf_set_ptime( pjmedia_jbuf *jb,
+					    unsigned ptime)
+{
+    PJ_ASSERT_RETURN(jb, PJ_EINVAL);
+
+    jb->jb_frame_ptime    = ptime;
+    jb->jb_min_shrink_gap = PJMEDIA_JBUF_DISC_MIN_GAP / ptime;
+    jb->jb_max_burst	  = PJ_MAX(MAX_BURST_MSEC / ptime,
+    				   jb->jb_max_count*3/4);
+
+    return PJ_SUCCESS;
+}
+
+
 /*
  * Set the jitter buffer to fixed delay mode. The default behavior
  * is to adapt the delay with actual packet delay.
@@ -973,6 +998,12 @@ PJ_DEF(void) pjmedia_jbuf_put_frame3(pjmedia_jbuf *jb,
 
     cur_size = jb_framelist_eff_size(&jb->jb_framelist);
 
+    /* Check if frame size is larger than JB frame size */
+    if (frame_size > jb->jb_frame_size) {
+	PJ_LOG(4, (THIS_FILE, "Warning: frame too large for jitter buffer, "
+		   "it will be truncated!"));
+    }
+
     /* Attempt to store the frame */
     min_frame_size = PJ_MIN(frame_size, jb->jb_frame_size);
     status = jb_framelist_put_at(&jb->jb_framelist, frame_seq, frame,
@@ -1129,6 +1160,7 @@ PJ_DEF(pj_status_t) pjmedia_jbuf_get_state( const pjmedia_jbuf *jb,
     state->frame_size = (unsigned)jb->jb_frame_size;
     state->min_prefetch = jb->jb_min_prefetch;
     state->max_prefetch = jb->jb_max_prefetch;
+    state->max_count = jb->jb_max_count;
 
     state->burst = jb->jb_eff_level;
     state->prefetch = jb->jb_prefetch;
