@@ -1,5 +1,5 @@
 /* $Id$ */
-/* 
+/*
  * Copyright (C) 2011 Teluu Inc. (http://www.teluu.com)
  *
  * This program is free software; you can redistribute it and/or modify
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 #include <pjmedia/vid_stream.h>
 #include <pjmedia/errno.h>
@@ -22,7 +22,6 @@
 #include <pjmedia/rtp.h>
 #include <pjmedia/rtcp.h>
 #include <pjmedia/jbuf.h>
-#include <pjmedia/stream_common.h>
 #include <pj/array.h>
 #include <pj/assert.h>
 #include <pj/compat/socket.h>
@@ -41,7 +40,7 @@
 
 #define THIS_FILE			"vid_stream.c"
 #define ERRLEVEL			1
-#define LOGERR_(expr)			stream_perror expr
+#define LOGERR_(expr)			PJ_PERROR(4,expr)
 #define TRC_(expr)			PJ_LOG(5,expr)
 #define SIGNATURE			PJMEDIA_SIG_PORT_VID_STREAM
 
@@ -75,12 +74,8 @@
  */
 #define MIN_CHUNKS_PER_FRM	30
 
-/* Video stream keep-alive feature is currently disabled. */
-#if defined(PJMEDIA_STREAM_ENABLE_KA) && PJMEDIA_STREAM_ENABLE_KA != 0
-#   undef PJMEDIA_STREAM_ENABLE_KA
-#   define PJMEDIA_STREAM_ENABLE_KA 0
-#endif
-
+/*  Number of send error before repeat the report. */
+#define SEND_ERR_COUNT_TO_REPORT	50
 
 /**
  * Media channel.
@@ -121,7 +116,6 @@ struct pjmedia_vid_stream
     pj_str_t		     cname;	    /**< SDES CNAME		    */
 
     pjmedia_transport	    *transport;	    /**< Stream transport.	    */
-    unsigned		     send_err_cnt;  /**< Send error count.          */
 
     pj_mutex_t		    *jb_mutex;
     pjmedia_jbuf	    *jb;	    /**< Jitter buffer.		    */
@@ -140,9 +134,10 @@ struct pjmedia_vid_stream
     unsigned		     dec_max_size;  /**< Size of decoded/raw picture*/
     pjmedia_ratio	     dec_max_fps;   /**< Max fps of decoding dir.   */
     pjmedia_frame            dec_frame;	    /**< Current decoded frame.     */
+    unsigned		     dec_delay_cnt; /**< Decoding delay (in frames).*/
     pjmedia_event            fmt_event;	    /**< Buffered fmt_changed event
                                                  to avoid deadlock	    */
-    pjmedia_event            miss_keyframe_event; 
+    pjmedia_event            miss_keyframe_event;
 					    /**< Buffered missing keyframe
                                                  event for delayed republish*/
 
@@ -152,8 +147,14 @@ struct pjmedia_vid_stream
     unsigned		     rx_frame_cnt;  /**< # of array in rx_frames    */
     pjmedia_frame	    *rx_frames;	    /**< Temp. buffer for incoming
 					         frame assembly.	    */
-
     pj_bool_t		     force_keyframe;/**< Forced to encode keyframe? */
+    unsigned		     num_keyframe;  /**< The number of keyframe needed
+						 to be sent, after the stream
+						 is created. */
+    pj_timestamp	     last_keyframe_tx;
+					    /**< Timestamp of the last
+						 keyframe. */
+
 
 #if defined(PJMEDIA_STREAM_ENABLE_KA) && PJMEDIA_STREAM_ENABLE_KA!=0
     pj_bool_t		     use_ka;	       /**< Stream keep-alive with non-
@@ -171,10 +172,21 @@ struct pjmedia_vid_stream
     pjmedia_vid_codec	    *codec;	    /**< Codec instance being used. */
     pj_uint32_t		     last_dec_ts;   /**< Last decoded timestamp.    */
     int			     last_dec_seq;  /**< Last decoded sequence.     */
-
+    pj_uint32_t		     rtp_tx_err_cnt;/**< The number of RTP
+						 send() error		    */
+    pj_uint32_t		     rtcp_tx_err_cnt;/**< The number of RTCP
+						  send() error		    */
 
     pj_timestamp	     ts_freq;	    /**< Timestamp frequency.	    */
 
+    pj_sockaddr		     rem_rtp_addr;  /**< Remote RTP address	    */
+    unsigned		     rem_rtp_flag;  /**< Indicator flag about
+						 packet from this addr.
+						 0=no pkt, 1=good ssrc pkts,
+						 2=bad ssrc pkts	    */
+    pj_sockaddr		     rtp_src_addr;  /**< Actual packet src addr.    */
+    unsigned		     rtp_src_cnt;   /**< How many pkt from this addr*/
+
 #if TRACE_RC
     unsigned		     rc_total_sleep;
     unsigned		     rc_total_pkt;
@@ -188,23 +200,14 @@ struct pjmedia_vid_stream
 static pj_status_t decode_frame(pjmedia_vid_stream *stream,
                                 pjmedia_frame *frame);
 
-/*
- * Print error.
- */
-static void stream_perror(const char *sender, const char *title,
-			  pj_status_t status)
-{
-    char errmsg[PJ_ERR_MSG_SIZE];
-
-    pj_strerror(status, errmsg, sizeof(errmsg));
-    PJ_LOG(4,(sender, "%s: %s [err:%d]", title, errmsg, status));
-}
-
 
 static pj_status_t send_rtcp(pjmedia_vid_stream *stream,
 			     pj_bool_t with_sdes,
 			     pj_bool_t with_bye);
 
+static void on_rx_rtcp( void *data,
+                        void *pkt,
+                        pj_ssize_t bytes_read);
 
 #if TRACE_JB
 
@@ -233,7 +236,7 @@ PJ_INLINE(int) trace_jb_print_timestamp(char **buf, pj_ssize_t len)
     return 0;
 }
 
-PJ_INLINE(int) trace_jb_print_state(pjmedia_vid_stream *stream, 
+PJ_INLINE(int) trace_jb_print_state(pjmedia_vid_stream *stream,
 				    char **buf, pj_ssize_t len)
 {
     char *p = *buf;
@@ -405,6 +408,32 @@ static pj_status_t stream_event_cb(pjmedia_event *event,
     return pjmedia_event_publish(NULL, stream, event, 0);
 }
 
+/**
+ * Publish transport error event.
+ */
+static void publish_tp_event(pjmedia_event_type event_type,
+			     pj_status_t status,
+			     pj_bool_t is_rtp,
+			     pjmedia_dir dir,
+			     pjmedia_vid_stream *stream)
+{
+    pjmedia_event ev;
+    pj_timestamp ts_now;
+
+    pj_get_timestamp(&ts_now);
+    pj_bzero(&ev.data.med_tp_err, sizeof(ev.data.med_tp_err));
+
+    /* Publish event. */
+    pjmedia_event_init(&ev, event_type,
+		       &ts_now, stream);
+    ev.data.med_tp_err.type = PJMEDIA_TYPE_VIDEO;
+    ev.data.med_tp_err.is_rtp = is_rtp;
+    ev.data.med_tp_err.dir = dir;
+    ev.data.med_tp_err.status = status;
+
+    pjmedia_event_publish(NULL, stream, &ev, 0);
+}
+
 #if defined(PJMEDIA_STREAM_ENABLE_KA) && PJMEDIA_STREAM_ENABLE_KA != 0
 /*
  * Send keep-alive packet using non-codec frame.
@@ -438,6 +467,9 @@ static void send_keep_alive_packet(pjmedia_vid_stream *stream)
     /* Send RTCP */
     send_rtcp(stream, PJ_TRUE, PJ_FALSE);
 
+    /* Update stats in case the stream is paused */
+    stream->rtcp.stat.rtp_tx_last_seq = pj_ntohs(stream->enc->rtp.out_hdr.seq);
+
 #elif PJMEDIA_STREAM_ENABLE_KA == PJMEDIA_STREAM_KA_USER
 
     /* Keep-alive packet is defined in PJMEDIA_STREAM_KA_USER_PKT */
@@ -459,7 +491,7 @@ static void send_keep_alive_packet(pjmedia_vid_stream *stream)
 			        pkt_len);
 
 #else
-    
+
     PJ_UNUSED_ARG(stream);
 
 #endif
@@ -475,6 +507,7 @@ static pj_status_t send_rtcp(pjmedia_vid_stream *stream,
     pj_uint8_t *pkt;
     int len, max_len;
     pj_status_t status;
+    pjmedia_vid_channel *channel = stream->enc;
 
     /* Build RTCP RR/SR packet */
     pjmedia_rtcp_build_rtcp(&stream->rtcp, &sr_rr_pkt, &len);
@@ -523,7 +556,15 @@ static pj_status_t send_rtcp(pjmedia_vid_stream *stream,
 
     /* Send! */
     status = pjmedia_transport_send_rtcp(stream->transport, pkt, len);
-
+    if (status != PJ_SUCCESS) {
+	if (stream->rtcp_tx_err_cnt++ == 0) {
+	    LOGERR_((channel->port.info.name.ptr, status,
+		     "Error sending RTCP"));
+	}
+	if (stream->rtcp_tx_err_cnt > SEND_ERR_COUNT_TO_REPORT) {
+	    stream->rtcp_tx_err_cnt = 0;
+	}
+    }
     return status;
 }
 
@@ -536,19 +577,19 @@ static pj_status_t send_rtcp(pjmedia_vid_stream *stream,
  */
 static void check_tx_rtcp(pjmedia_vid_stream *stream, pj_uint32_t timestamp)
 {
-    /* Note that timestamp may represent local or remote timestamp, 
+    /* Note that timestamp may represent local or remote timestamp,
      * depending on whether this function is called from put_frame()
      * or get_frame().
      */
 
 
     if (stream->rtcp_last_tx == 0) {
-	
+
 	stream->rtcp_last_tx = timestamp;
 
     } else if (timestamp - stream->rtcp_last_tx >= stream->rtcp_interval) {
 	pj_status_t status;
-	
+
 	status = send_rtcp(stream, !stream->rtcp_sdes_bye_disabled, PJ_FALSE);
 	if (status != PJ_SUCCESS) {
 	    PJ_PERROR(4,(stream->name.ptr, status,
@@ -588,14 +629,13 @@ static void dump_bin(const char *buf, unsigned len)
 
 /*
  * This callback is called by stream transport on receipt of packets
- * in the RTP socket. 
+ * in the RTP socket.
  */
-static void on_rx_rtp( void *data, 
-		       void *pkt,
-                       pj_ssize_t bytes_read)
-
+static void on_rx_rtp( pjmedia_tp_cb_param *param)
 {
-    pjmedia_vid_stream *stream = (pjmedia_vid_stream*) data;
+    pjmedia_vid_stream *stream = (pjmedia_vid_stream*) param->user_data;
+    void *pkt = param->pkt;
+    pj_ssize_t bytes_read = param->size;
     pjmedia_vid_channel *channel = stream->dec;
     const pjmedia_rtp_hdr *hdr;
     const void *payload;
@@ -606,8 +646,19 @@ static void on_rx_rtp( void *data,
 
     /* Check for errors */
     if (bytes_read < 0) {
-	LOGERR_((channel->port.info.name.ptr, "RTP recv() error", 
-		(pj_status_t)-bytes_read));
+	status = (pj_status_t)-bytes_read;
+	if (status == PJ_STATUS_FROM_OS(OSERR_EWOULDBLOCK)) {
+	    return;
+	}
+
+	LOGERR_((channel->port.info.name.ptr, status,
+		 "Unable to receive RTP packet"));
+
+	if (status == PJ_ESOCKETSTOP) {
+	    /* Publish receive error event. */
+	    publish_tp_event(PJMEDIA_EVENT_MEDIA_TP_ERR, status, PJ_TRUE,
+			     PJMEDIA_DIR_DECODING, stream);
+	}
 	return;
     }
 
@@ -619,11 +670,17 @@ static void on_rx_rtp( void *data,
     status = pjmedia_rtp_decode_rtp(&channel->rtp, pkt, (int)bytes_read,
 				    &hdr, &payload, &payloadlen);
     if (status != PJ_SUCCESS) {
-	LOGERR_((channel->port.info.name.ptr, "RTP decode error", status));
+	LOGERR_((channel->port.info.name.ptr, status, "RTP decode error"));
 	stream->rtcp.stat.rx.discard++;
 	return;
     }
 
+    /* Check if multiplexing is allowed and the payload indicates RTCP. */
+    if (stream->info.rtcp_mux && hdr->pt >= 64 && hdr->pt <= 95) {
+    	on_rx_rtcp(stream, pkt, bytes_read);
+    	return;
+    }
+
     /* Ignore the packet if decoder is paused */
     if (channel->paused)
 	goto on_return;
@@ -633,9 +690,9 @@ static void on_rx_rtp( void *data,
      */
     pjmedia_rtp_session_update2(&channel->rtp, hdr, &seq_st, PJ_TRUE);
     if (seq_st.status.value) {
-	TRC_  ((channel->port.info.name.ptr, 
+	TRC_  ((channel->port.info.name.ptr,
 		"RTP status: badpt=%d, badssrc=%d, dup=%d, "
-		"outorder=%d, probation=%d, restart=%d", 
+		"outorder=%d, probation=%d, restart=%d",
 		seq_st.status.flag.badpt,
 		seq_st.status.flag.badssrc,
 		seq_st.status.flag.dup,
@@ -649,7 +706,7 @@ static void on_rx_rtp( void *data,
 		      hdr->pt, channel->rtp.out_pt));
 	}
 
-	if (seq_st.status.flag.badssrc) {
+	if (!stream->info.has_rem_ssrc && seq_st.status.flag.badssrc) {
 	    PJ_LOG(4,(channel->port.info.name.ptr,
 		      "Changed RTP peer SSRC %d (previously %d)",
 		      channel->rtp.peer_ssrc, stream->rtcp.peer_ssrc));
@@ -671,6 +728,62 @@ static void on_rx_rtp( void *data,
 	goto on_return;
     }
 
+    /* See if source address of RTP packet is different than the
+     * configured address, and check if we need to tell the
+     * media transport to switch RTP remote address.
+     */
+    if (param->src_addr) {
+        pj_bool_t badssrc = (stream->info.has_rem_ssrc &&
+        		     seq_st.status.flag.badssrc);
+
+	if (pj_sockaddr_cmp(&stream->rem_rtp_addr, param->src_addr) == 0) {
+	    /* We're still receiving from rem_rtp_addr. */
+	    stream->rtp_src_cnt = 0;
+	    stream->rem_rtp_flag = badssrc? 2: 1;
+	} else {
+	    stream->rtp_src_cnt++;
+
+	    if (stream->rtp_src_cnt < PJMEDIA_RTP_NAT_PROBATION_CNT) {
+	    	if (stream->rem_rtp_flag == 1 ||
+	    	    (stream->rem_rtp_flag == 2 && badssrc))
+	    	{
+		    /* Only discard if:
+		     * - we have ever received packet with good ssrc from
+		     *   remote address (rem_rtp_addr), or
+		     * - we have ever received packet with bad ssrc from
+		     *   remote address and this packet also has bad ssrc.
+		     */
+	    	    pkt_discarded = PJ_TRUE;
+	    	    goto on_return;
+	    	}
+	    	if (stream->info.has_rem_ssrc && !seq_st.status.flag.badssrc
+	    	    && stream->rem_rtp_flag != 1)
+	    	{
+	    	    /* Immediately switch if we receive packet with the
+	    	     * correct ssrc AND we never receive packets with
+	    	     * good ssrc from rem_rtp_addr.
+	    	     */
+	    	    param->rem_switch = PJ_TRUE;
+	    	}
+	    } else {
+	        /* Switch. We no longer receive packets from rem_rtp_addr. */
+	        param->rem_switch = PJ_TRUE;
+	    }
+
+	    if (param->rem_switch) {
+		/* Set remote RTP address to source address */
+		pj_sockaddr_cp(&stream->rem_rtp_addr, param->src_addr);
+
+		/* Reset counter and flag */
+		stream->rtp_src_cnt = 0;
+		stream->rem_rtp_flag = badssrc? 2: 1;
+
+		/* Update RTCP peer ssrc */
+	    	stream->rtcp.peer_ssrc = pj_ntohl(hdr->ssrc);
+	    }
+	}
+    }
+
     pj_mutex_lock( stream->jb_mutex );
 
     /* Quickly see if there may be a full picture in the jitter buffer, and
@@ -717,7 +830,7 @@ static void on_rx_rtp( void *data,
 	PJ_LOG(4,(channel->port.info.name.ptr, "Jitter buffer reset"));
     } else {
 	/* Just put the payload into jitter buffer */
-	pjmedia_jbuf_put_frame3(stream->jb, payload, payloadlen, 0, 
+	pjmedia_jbuf_put_frame3(stream->jb, payload, payloadlen, 0,
 				pj_ntohs(hdr->seq), pj_ntohl(hdr->ts), NULL);
 
 #if TRACE_JB
@@ -729,16 +842,17 @@ static void on_rx_rtp( void *data,
 
 
     /* Check if now is the time to transmit RTCP SR/RR report.
-     * We only do this when stream direction is "decoding only", 
+     * We only do this when stream direction is "decoding only" or
+     * if the encoder is paused,
      * because otherwise check_tx_rtcp() will be handled by put_frame()
      */
-    if (stream->dir == PJMEDIA_DIR_DECODING) {
+    if (stream->dir == PJMEDIA_DIR_DECODING || stream->enc->paused) {
 	check_tx_rtcp(stream, pj_ntohl(hdr->ts));
     }
 
     if (status != 0) {
-	LOGERR_((channel->port.info.name.ptr, "Jitter buffer put() error", 
-		status));
+	LOGERR_((channel->port.info.name.ptr, status,
+		 "Jitter buffer put() error"));
 	pkt_discarded = PJ_TRUE;
 	goto on_return;
     }
@@ -767,18 +881,27 @@ on_return:
 
 /*
  * This callback is called by stream transport on receipt of packets
- * in the RTCP socket. 
+ * in the RTCP socket.
  */
 static void on_rx_rtcp( void *data,
-                        void *pkt, 
+                        void *pkt,
                         pj_ssize_t bytes_read)
 {
     pjmedia_vid_stream *stream = (pjmedia_vid_stream*) data;
+    pj_status_t status;
 
     /* Check for errors */
     if (bytes_read < 0) {
-	LOGERR_((stream->cname.ptr, "RTCP recv() error", 
-		 (pj_status_t)-bytes_read));
+	status = (pj_status_t)-bytes_read;
+	if (status == PJ_STATUS_FROM_OS(OSERR_EWOULDBLOCK)) {
+	    return;
+	}
+	LOGERR_((stream->cname.ptr, status, "Unable to receive RTCP packet"));
+	if (status == PJ_ESOCKETSTOP) {
+	    /* Publish receive error event. */
+	    publish_tp_event(PJMEDIA_EVENT_MEDIA_TP_ERR, status, PJ_FALSE,
+			     PJMEDIA_DIR_DECODING, stream);
+	}
 	return;
     }
 
@@ -800,6 +923,7 @@ static pj_status_t put_frame(pjmedia_port *port,
     pjmedia_vid_encode_opt enc_opt;
     unsigned pkt_cnt = 0;
     pj_timestamp initial_time;
+    pj_timestamp null_ts ={{0}};
 
 #if defined(PJMEDIA_STREAM_ENABLE_KA) && PJMEDIA_STREAM_ENABLE_KA != 0
     /* If the interval since last sending packet is greater than
@@ -809,18 +933,15 @@ static pj_status_t put_frame(pjmedia_port *port,
     {
 	pj_uint32_t dtx_duration;
 
-	dtx_duration = pj_timestamp_diff32(&stream->last_frm_ts_sent, 
+	dtx_duration = pj_timestamp_diff32(&stream->last_frm_ts_sent,
 					   &frame->timestamp);
-        /* Video stream keep-alive feature is currently disabled. */
-        /*
+
         if (dtx_duration >
-	    PJMEDIA_STREAM_KA_INTERVAL *
-            PJMEDIA_PIA_SRATE(&channel->port.info))
+	    PJMEDIA_STREAM_KA_INTERVAL * stream->info.codec_info.clock_rate)
 	{
 	    send_keep_alive_packet(stream);
 	    stream->last_frm_ts_sent = frame->timestamp;
 	}
-        */
     }
 #endif
 
@@ -832,9 +953,35 @@ static pj_status_t put_frame(pjmedia_port *port,
     /* Get frame length in timestamp unit */
     rtp_ts_len = stream->frame_ts_len;
 
+    /* Empty video frame? Just update RTP timestamp for now */
+    if (frame->type==PJMEDIA_FRAME_TYPE_VIDEO && frame->size==0) {
+	pjmedia_rtp_encode_rtp(&channel->rtp, channel->pt, 1, 0,
+			       rtp_ts_len,  (const void**)&rtphdr,
+			       &rtphdrlen);
+	return PJ_SUCCESS;
+    }
+
     /* Init frame_out buffer. */
+    pj_bzero(&frame_out, sizeof(frame_out));
     frame_out.buf = ((char*)channel->buf) + sizeof(pjmedia_rtp_hdr);
-    frame_out.size = 0;
+
+    /* Check if need to send keyframe. */
+    if (stream->num_keyframe &&
+	(pj_cmp_timestamp(&null_ts, &stream->last_keyframe_tx) != 0))
+    {
+	unsigned elapse_time;
+	pj_timestamp now;
+
+	pj_get_timestamp(&now);
+
+	elapse_time = pj_elapsed_msec(&stream->last_keyframe_tx, &now);
+
+	if (elapse_time > stream->info.sk_cfg.interval)
+	{
+	    stream->force_keyframe = PJ_TRUE;
+	    --stream->num_keyframe;
+	}
+    }
 
     /* Init encoding option */
     pj_bzero(&enc_opt, sizeof(enc_opt));
@@ -853,8 +1000,8 @@ static pj_status_t put_frame(pjmedia_port *port,
                                             &frame_out,
                                             &has_more_data);
     if (status != PJ_SUCCESS) {
-	LOGERR_((channel->port.info.name.ptr,
-		"Codec encode_begin() error", status));
+	LOGERR_((channel->port.info.name.ptr, status,
+		"Codec encode_begin() error"));
 
 	/* Update RTP timestamp */
 	pjmedia_rtp_encode_rtp(&channel->rtp, channel->pt, 1, 0,
@@ -862,9 +1009,16 @@ static pj_status_t put_frame(pjmedia_port *port,
 			       &rtphdrlen);
 	return status;
     }
-    
+
     pj_get_timestamp(&initial_time);
 
+    if ((stream->num_keyframe) &&
+	((frame_out.bit_info & PJMEDIA_VID_FRM_KEYFRAME)
+						  == PJMEDIA_VID_FRM_KEYFRAME))
+    {
+	stream->last_keyframe_tx = initial_time;
+    }
+
     /* Loop while we have frame to send */
     for (;;) {
 	status = pjmedia_rtp_encode_rtp(&channel->rtp,
@@ -875,8 +1029,8 @@ static pj_status_t put_frame(pjmedia_port *port,
 	                                (const void**)&rtphdr,
 	                                &rtphdrlen);
 	if (status != PJ_SUCCESS) {
-	    LOGERR_((channel->port.info.name.ptr,
-		    "RTP encode_rtp() error", status));
+	    LOGERR_((channel->port.info.name.ptr, status,
+		    "RTP encode_rtp() error"));
 	    return status;
 	}
 
@@ -893,17 +1047,14 @@ static pj_status_t put_frame(pjmedia_port *port,
 						frame_out.size +
 						    sizeof(pjmedia_rtp_hdr));
 	    if (status != PJ_SUCCESS) {
-		enum { COUNT_TO_REPORT = 20 };
-		if (stream->send_err_cnt++ == 0) {
-		    LOGERR_((channel->port.info.name.ptr,
-			     "Transport send_rtp() error",
-			     status));
+		if (stream->rtp_tx_err_cnt++ == 0) {
+		    LOGERR_((channel->port.info.name.ptr, status,
+			     "Error sending RTP"));
+		}
+		if (stream->rtp_tx_err_cnt > SEND_ERR_COUNT_TO_REPORT) {
+		    stream->rtp_tx_err_cnt = 0;
 		}
-		if (stream->send_err_cnt > COUNT_TO_REPORT)
-		    stream->send_err_cnt = 0;
-		/* Ignore this error */
 	    }
-
 	    pjmedia_rtcp_tx_rtp(&stream->rtcp, (unsigned)frame_out.size);
 	    total_sent += frame_out.size;
 	    pkt_cnt++;
@@ -924,8 +1075,8 @@ static pj_status_t put_frame(pjmedia_port *port,
 				               &frame_out,
 					       &has_more_data);
 	if (status != PJ_SUCCESS) {
-	    LOGERR_((channel->port.info.name.ptr,
-		     "Codec encode_more() error", status));
+	    LOGERR_((channel->port.info.name.ptr, status,
+		     "Codec encode_more() error"));
 	    /* Ignore this error (?) */
 	    break;
 	}
@@ -973,7 +1124,7 @@ static pj_status_t put_frame(pjmedia_port *port,
     }
 #endif
 
-    /* Check if now is the time to transmit RTCP SR/RR report. 
+    /* Check if now is the time to transmit RTCP SR/RR report.
      * We only do this when stream direction is not "decoding only", because
      * when it is, check_tx_rtcp() will be handled by get_frame().
      */
@@ -1007,10 +1158,10 @@ static pj_status_t decode_frame(pjmedia_vid_stream *stream,
                                 pjmedia_frame *frame)
 {
     pjmedia_vid_channel *channel = stream->dec;
-    pj_uint32_t last_ts = 0;
+    pj_uint32_t last_ts = 0, frm_ts = 0;
     int frm_first_seq = 0, frm_last_seq = 0;
     pj_bool_t got_frame = PJ_FALSE;
-    unsigned cnt;
+    unsigned cnt, frm_pkt_cnt = 0, frm_cnt = 0;
     pj_status_t status;
 
     /* Repeat get payload from the jitter buffer until all payloads with same
@@ -1018,7 +1169,7 @@ static pj_status_t decode_frame(pjmedia_vid_stream *stream,
      */
 
     /* Check if we got a decodable frame */
-    for (cnt=0; ; ++cnt) {
+    for (cnt=0; ; ) {
 	char ptype;
 	pj_uint32_t ts;
 	int seq;
@@ -1027,38 +1178,68 @@ static pj_status_t decode_frame(pjmedia_vid_stream *stream,
 	pjmedia_jbuf_peek_frame(stream->jb, cnt, NULL, NULL,
 				&ptype, NULL, &ts, &seq);
 	if (ptype == PJMEDIA_JB_NORMAL_FRAME) {
+	    if (stream->last_dec_ts ==  ts) {
+		/* Remove any late packet (the frame has been decoded) */
+		pjmedia_jbuf_remove_frame(stream->jb, 1);
+		continue;
+	    }
+
 	    if (last_ts == 0) {
 		last_ts = ts;
+
+		/* Init timestamp and first seq of the first frame */
+		frm_ts = ts;
 		frm_first_seq = seq;
 	    }
 	    if (ts != last_ts) {
-		got_frame = PJ_TRUE;
-		break;
+		last_ts = ts;
+		if (frm_pkt_cnt == 0)
+		    frm_pkt_cnt = cnt;
+
+		/* Is it time to decode? Check with minimum delay setting */
+		if (++frm_cnt == stream->dec_delay_cnt) {
+		    got_frame = PJ_TRUE;
+		    break;
+		}
 	    }
-	    frm_last_seq = seq;
 	} else if (ptype == PJMEDIA_JB_ZERO_EMPTY_FRAME) {
 	    /* No more packet in the jitter buffer */
 	    break;
 	}
+
+	++cnt;
     }
 
     if (got_frame) {
 	unsigned i;
 
-	/* Generate frame bitstream from the payload */
-	if (cnt > stream->rx_frame_cnt) {
-	    PJ_LOG(1,(channel->port.info.name.ptr,
-		      "Discarding %u frames because array is full!",
-		      cnt - stream->rx_frame_cnt));
-	    pjmedia_jbuf_remove_frame(stream->jb, cnt - stream->rx_frame_cnt);
-	    cnt = stream->rx_frame_cnt;
+	/* Exclude any MISSING frames in the end of the packets array, as
+	 * it may be part of the next video frame (late packets).
+	 */
+	for (; frm_pkt_cnt > 1; --frm_pkt_cnt) {
+	    char ptype;
+	    pjmedia_jbuf_peek_frame(stream->jb, frm_pkt_cnt, NULL, NULL, &ptype,
+				    NULL, NULL, NULL);
+	    if (ptype == PJMEDIA_JB_NORMAL_FRAME)
+		break;
 	}
 
-	for (i = 0; i < cnt; ++i) {
+	/* Check if the packet count for this frame exceeds the limit */
+	if (frm_pkt_cnt > stream->rx_frame_cnt) {
+	    PJ_LOG(1,(channel->port.info.name.ptr,
+		      "Discarding %u frames because array is full!",
+		      frm_pkt_cnt - stream->rx_frame_cnt));
+	    pjmedia_jbuf_remove_frame(stream->jb,
+				      frm_pkt_cnt - stream->rx_frame_cnt);
+	    frm_pkt_cnt = stream->rx_frame_cnt;
+	}
+
+	/* Generate frame bitstream from the payload */
+	for (i = 0; i < frm_pkt_cnt; ++i) {
 	    char ptype;
 
 	    stream->rx_frames[i].type = PJMEDIA_FRAME_TYPE_VIDEO;
-	    stream->rx_frames[i].timestamp.u64 = last_ts;
+	    stream->rx_frames[i].timestamp.u64 = frm_ts;
 	    stream->rx_frames[i].bit_info = 0;
 
 	    /* We use jbuf_peek_frame() as it will returns the pointer of
@@ -1067,7 +1248,7 @@ static pj_status_t decode_frame(pjmedia_vid_stream *stream,
 	    pjmedia_jbuf_peek_frame(stream->jb, i,
 				    (const void**)&stream->rx_frames[i].buf,
 				    &stream->rx_frames[i].size, &ptype,
-				    NULL, NULL, NULL);
+				    NULL, NULL, &frm_last_seq);
 
 	    if (ptype != PJMEDIA_JB_NORMAL_FRAME) {
 		/* Packet lost, must set payload to NULL and keep going */
@@ -1079,55 +1260,73 @@ static pj_status_t decode_frame(pjmedia_vid_stream *stream,
 	}
 
 	/* Decode */
-	status = pjmedia_vid_codec_decode(stream->codec, cnt,
+	status = pjmedia_vid_codec_decode(stream->codec, frm_pkt_cnt,
 	                                  stream->rx_frames,
 	                                  (unsigned)frame->size, frame);
 	if (status != PJ_SUCCESS) {
-	    LOGERR_((channel->port.info.name.ptr, "codec decode() error",
-		     status));
+	    LOGERR_((channel->port.info.name.ptr, status,
+		     "codec decode() error"));
 	    frame->type = PJMEDIA_FRAME_TYPE_NONE;
 	    frame->size = 0;
 	}
 
-	pjmedia_jbuf_remove_frame(stream->jb, cnt);
+	pjmedia_jbuf_remove_frame(stream->jb, frm_pkt_cnt);
     }
 
     /* Learn remote frame rate after successful decoding */
-    if (frame->type == PJMEDIA_FRAME_TYPE_VIDEO && frame->size)
+    if (got_frame && frame->type == PJMEDIA_FRAME_TYPE_VIDEO && frame->size)
     {
 	/* Only check remote frame rate when timestamp is not wrapping and
 	 * sequence is increased by 1.
 	 */
-	if (last_ts > stream->last_dec_ts &&
+	if (frm_ts > stream->last_dec_ts &&
 	    frm_first_seq - stream->last_dec_seq == 1)
 	{
 	    pj_uint32_t ts_diff;
-	    pjmedia_video_format_detail *vfd;
+	    pjmedia_ratio new_fps;
 
-	    ts_diff = last_ts - stream->last_dec_ts;
-	    vfd = pjmedia_format_get_video_format_detail(
-				    &channel->port.info.fmt, PJ_TRUE);
-	    if (stream->info.codec_info.clock_rate * vfd->fps.denum !=
-		vfd->fps.num * ts_diff)
+	    ts_diff = frm_ts - stream->last_dec_ts;
+
+	    /* Calculate new FPS based on RTP timestamp diff */
+	    if (stream->info.codec_info.clock_rate % ts_diff == 0) {
+		new_fps.num = stream->info.codec_info.clock_rate/ts_diff;
+		new_fps.denum = 1;
+	    } else {
+		new_fps.num = stream->info.codec_info.clock_rate;
+		new_fps.denum = ts_diff;
+	    }
+
+	    /* Only apply the new FPS when it is >0, <=100, and increasing */
+	    if (new_fps.num/new_fps.denum <= 100 &&
+		new_fps.num/new_fps.denum > 0 &&
+		new_fps.num*1.0/new_fps.denum >
+		stream->dec_max_fps.num*1.0/stream->dec_max_fps.denum)
 	    {
-		/* Frame rate changed, update decoding port info */
-		if (stream->info.codec_info.clock_rate % ts_diff == 0) {
-		    vfd->fps.num = stream->info.codec_info.clock_rate/ts_diff;
-		    vfd->fps.denum = 1;
-		} else {
-		    vfd->fps.num = stream->info.codec_info.clock_rate;
-		    vfd->fps.denum = ts_diff;
+		pjmedia_video_format_detail *vfd;
+		vfd = pjmedia_format_get_video_format_detail(
+					&channel->port.info.fmt, PJ_TRUE);
+
+		/* Update FPS in channel & stream info */
+		vfd->fps = new_fps;
+		stream->info.codec_param->dec_fmt.det.vid.fps = new_fps;
+
+		/* Update the decoding delay */
+		{
+		    pjmedia_jb_state jb_state;
+		    pjmedia_jbuf_get_state(stream->jb, &jb_state);
+
+		    stream->dec_delay_cnt =
+				    ((PJMEDIA_VID_STREAM_DECODE_MIN_DELAY_MSEC *
+				      vfd->fps.num) +
+				     (1000 * vfd->fps.denum) - 1) /
+				    (1000 * vfd->fps.denum);
+		    if (stream->dec_delay_cnt < 1)
+			stream->dec_delay_cnt = 1;
+		    if (stream->dec_delay_cnt > jb_state.max_count * 4/5)
+			stream->dec_delay_cnt = jb_state.max_count * 4/5;
 		}
 
-		/* Update stream info */
-		stream->info.codec_param->dec_fmt.det.vid.fps = vfd->fps;
-
-		/* Publish PJMEDIA_EVENT_FMT_CHANGED event if frame rate
-		 * increased and not exceeding 100fps.
-		 */
-		if (vfd->fps.num/vfd->fps.denum <= 100 &&
-		    vfd->fps.num * stream->dec_max_fps.denum >
-		    stream->dec_max_fps.num * vfd->fps.denum)
+		/* Publish PJMEDIA_EVENT_FMT_CHANGED event */
 		{
 		    pjmedia_event *event = &stream->fmt_event;
 
@@ -1154,7 +1353,7 @@ static pj_status_t decode_frame(pjmedia_vid_stream *stream,
 
 	/* Update last frame seq and timestamp */
 	stream->last_dec_seq = frm_last_seq;
-	stream->last_dec_ts = last_ts;
+	stream->last_dec_ts = frm_ts;
     }
 
     return got_frame ? PJ_SUCCESS : PJ_ENOTFOUND;
@@ -1268,7 +1467,7 @@ static pj_status_t create_channel( pj_pool_t *pool,
     pjmedia_format *fmt;
     char fourcc_name[5];
     pjmedia_port_info *pi;
-    
+
     pj_assert(info->type == PJMEDIA_TYPE_VIDEO);
     pj_assert(dir == PJMEDIA_DIR_DECODING || dir == PJMEDIA_DIR_ENCODING);
 
@@ -1293,7 +1492,7 @@ static pj_status_t create_channel( pj_pool_t *pool,
     channel->dir = dir;
     channel->paused = 1;
     channel->pt = pt;
-    
+
     /* Allocate buffer for outgoing packet. */
     if (dir == PJMEDIA_DIR_ENCODING) {
 	channel->buf_size = sizeof(pjmedia_rtp_hdr) + stream->frame_size;
@@ -1312,14 +1511,14 @@ static pj_status_t create_channel( pj_pool_t *pool,
     }
 
     /* Create RTP and RTCP sessions: */
-    if (info->rtp_seq_ts_set == 0) {
-	status = pjmedia_rtp_session_init(&channel->rtp, pt, info->ssrc);
-    } else {
+    {
 	pjmedia_rtp_session_setting settings;
 
-	settings.flags = (pj_uint8_t)((info->rtp_seq_ts_set << 2) | 3);
+	settings.flags = (pj_uint8_t)((info->rtp_seq_ts_set << 2) |
+				      (info->has_rem_ssrc << 4) | 3);
 	settings.default_pt = pt;
 	settings.sender_ssrc = info->ssrc;
+	settings.peer_ssrc = info->rem_ssrc;
 	settings.seq = info->rtp_seq;
 	settings.ts = info->rtp_ts;
 	status = pjmedia_rtp_session_init2(&channel->rtp, settings);
@@ -1375,6 +1574,7 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_create(
     pjmedia_video_format_detail *vfd_enc, *vfd_dec;
     char *p;
     pj_status_t status;
+    pjmedia_transport_attach_param att_param;
 
     if (!pool) {
 	own_pool = pjmedia_endpt_create_pool( endpt, "vstrm%p",
@@ -1395,11 +1595,11 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_create(
 
     /* Init stream/port name */
     stream->name.ptr = (char*) pj_pool_alloc(pool, M);
-    stream->name.slen = pj_ansi_snprintf(stream->name.ptr, M, 
+    stream->name.slen = pj_ansi_snprintf(stream->name.ptr, M,
 					 "vstrm%p", stream);
 
     /* Create and initialize codec: */
-    status = pjmedia_vid_codec_mgr_alloc_codec(stream->codec_mgr, 
+    status = pjmedia_vid_codec_mgr_alloc_codec(stream->codec_mgr,
 					       &info->codec_info,
 					       &stream->codec);
     if (status != PJ_SUCCESS)
@@ -1409,7 +1609,7 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_create(
     if (!info->codec_param) {
 	pjmedia_vid_codec_param def_param;
 
-	status = pjmedia_vid_codec_mgr_get_default_param(stream->codec_mgr, 
+	status = pjmedia_vid_codec_mgr_get_default_param(stream->codec_mgr,
 						         &info->codec_info,
 						         &def_param);
 	if (status != PJ_SUCCESS)
@@ -1445,17 +1645,20 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_create(
 #if defined(PJMEDIA_STREAM_ENABLE_KA) && PJMEDIA_STREAM_ENABLE_KA!=0
     stream->use_ka = info->use_ka;
 #endif
+    stream->num_keyframe = info->sk_cfg.count;
 
-    /* Build random RTCP CNAME. CNAME has user@host format */
-    stream->cname.ptr = p = (char*) pj_pool_alloc(pool, 20);
-    pj_create_random_string(p, 5);
-    p += 5;
-    *p++ = '@'; *p++ = 'p'; *p++ = 'j';
-    pj_create_random_string(p, 6);
-    p += 6;
-    *p++ = '.'; *p++ = 'o'; *p++ = 'r'; *p++ = 'g';
-    stream->cname.slen = p - stream->cname.ptr;
-
+    stream->cname = info->cname;
+    if (stream->cname.slen == 0) {
+	/* Build random RTCP CNAME. CNAME has user@host format */
+    	stream->cname.ptr = p = (char*) pj_pool_alloc(pool, 20);
+    	pj_create_random_string(p, 5);
+    	p += 5;
+    	*p++ = '@'; *p++ = 'p'; *p++ = 'j';
+    	pj_create_random_string(p, 6);
+    	p += 6;
+    	*p++ = '.'; *p++ = 'o'; *p++ = 'r'; *p++ = 'g';
+    	stream->cname.slen = p - stream->cname.ptr;
+    }
 
     /* Create mutex to protect jitter buffer: */
 
@@ -1481,7 +1684,7 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_create(
 #if 0
     stream->frame_size = vfd_enc->max_bps/8 * vfd_enc->fps.denum /
 			 vfd_enc->fps.num;
-    
+
     /* As the maximum frame_size is not represented directly by maximum bps
      * (which includes intra and predicted frames), let's increase the
      * frame size value for safety.
@@ -1490,7 +1693,7 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_create(
 #endif
 
     /* Validate the frame size */
-    if (stream->frame_size == 0 || 
+    if (stream->frame_size == 0 ||
 	stream->frame_size > PJMEDIA_MAX_VIDEO_ENC_FRAME_SIZE)
     {
 	stream->frame_size = PJMEDIA_MAX_VIDEO_ENC_FRAME_SIZE;
@@ -1525,13 +1728,13 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_create(
     stream->dec_max_fps = vfd_dec->fps;
 
     /* Create decoder channel */
-    status = create_channel( pool, stream, PJMEDIA_DIR_DECODING, 
+    status = create_channel( pool, stream, PJMEDIA_DIR_DECODING,
 			     info->rx_pt, info, &stream->dec);
     if (status != PJ_SUCCESS)
 	return status;
 
     /* Create encoder channel */
-    status = create_channel( pool, stream, PJMEDIA_DIR_ENCODING, 
+    status = create_channel( pool, stream, PJMEDIA_DIR_ENCODING,
 			     info->tx_pt, info, &stream->enc);
     if (status != PJ_SUCCESS)
 	return status;
@@ -1570,6 +1773,16 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_create(
     else
 	jb_init  = 0;
 
+    /* Calculate the decoding delay (in number of frames) based on FPS */
+    stream->dec_delay_cnt = ((PJMEDIA_VID_STREAM_DECODE_MIN_DELAY_MSEC *
+			      vfd_dec->fps.num) +
+			     (1000 * vfd_dec->fps.denum) - 1) /
+			    (1000 * vfd_dec->fps.denum);
+    if (stream->dec_delay_cnt < 1)
+	stream->dec_delay_cnt = 1;
+    if (stream->dec_delay_cnt > jb_max * 4/5)
+	stream->dec_delay_cnt = jb_max * 4/5;
+
     /* Allocate array for temporary storage for assembly of incoming
      * frames. Add more just in case.
      */
@@ -1615,12 +1828,23 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_create(
 	stream->out_rtcp_pkt_size = PJMEDIA_MAX_MTU;
 
     stream->out_rtcp_pkt = pj_pool_alloc(pool, stream->out_rtcp_pkt_size);
+    pj_bzero(&att_param, sizeof(att_param));
+    att_param.stream = stream;
+    att_param.media_type = PJMEDIA_TYPE_VIDEO;
+    att_param.user_data = stream;
+    pj_sockaddr_cp(&att_param.rem_addr, &info->rem_addr);
+    pj_sockaddr_cp(&stream->rem_rtp_addr, &info->rem_addr);
+    if (info->rtcp_mux) {
+	pj_sockaddr_cp(&att_param.rem_rtcp, &info->rem_addr);
+    } else if (pj_sockaddr_has_addr(&info->rem_rtcp.addr)) {
+	pj_sockaddr_cp(&att_param.rem_rtcp, &info->rem_rtcp);
+    }
+    att_param.addr_len = pj_sockaddr_get_len(&info->rem_addr);
+    att_param.rtp_cb2 = &on_rx_rtp;
+    att_param.rtcp_cb = &on_rx_rtcp;
 
     /* Only attach transport when stream is ready. */
-    status = pjmedia_transport_attach(tp, stream, &info->rem_addr, 
-				      &info->rem_rtcp, 
-				      pj_sockaddr_get_len(&info->rem_addr), 
-                                      &on_rx_rtp, &on_rx_rtcp);
+    status = pjmedia_transport_attach2(tp, &att_param);
     if (status != PJ_SUCCESS)
 	return status;
 
@@ -1642,15 +1866,16 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_create(
 	char trace_name[PJ_MAXPATH];
 	pj_ssize_t len;
 
-	pj_ansi_snprintf(trace_name, sizeof(trace_name), 
+	pj_ansi_snprintf(trace_name, sizeof(trace_name),
 			 TRACE_JB_PATH_PREFIX "%s.csv",
 			 channel->port.info.name.ptr);
 	status = pj_file_open(pool, trace_name, PJ_O_RDWR,
 			      &stream->trace_jb_fd);
 	if (status != PJ_SUCCESS) {
 	    stream->trace_jb_fd = TRACE_JB_INVALID_FD;
-	    PJ_LOG(3,(THIS_FILE, "Failed creating RTP trace file '%s'", 
-		      trace_name));
+	    PJ_PERROR(3,(THIS_FILE, status,
+			 "Failed creating RTP trace file '%s'",
+			 trace_name));
 	} else {
 	    stream->trace_jb_buf = (char*)pj_pool_alloc(pool, PJ_LOG_MAX_SIZE);
 
@@ -1693,7 +1918,7 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_destroy( pjmedia_vid_stream *stream )
 	unsigned total_time;
 
 	total_time = pj_elapsed_msec(&stream->tx_start, &stream->tx_end);
-	PJ_LOG(5, (stream->name.ptr, 
+	PJ_LOG(5, (stream->name.ptr,
 		   "RC stat: pkt_cnt=%.2f/image, sleep=%.2fms/s, fps=%.2f",
 		   stream->rc_total_pkt*1.0/stream->rc_total_img,
 		   stream->rc_total_sleep*1000.0/total_time,
@@ -1706,7 +1931,7 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_destroy( pjmedia_vid_stream *stream )
 	send_rtcp(stream, PJ_TRUE, PJ_TRUE);
     }
 
-    /* Detach from transport 
+    /* Detach from transport
      * MUST NOT hold stream mutex while detaching from transport, as
      * it may cause deadlock. See ticket #460 for the details.
      */
@@ -1730,7 +1955,7 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_destroy( pjmedia_vid_stream *stream )
     }
 
     /* Free mutex */
-    
+
     if (stream->jb_mutex) {
 	pj_mutex_destroy(stream->jb_mutex);
 	stream->jb_mutex = NULL;
@@ -1749,11 +1974,7 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_destroy( pjmedia_vid_stream *stream )
     }
 #endif
 
-    if (stream->own_pool) {
-	pj_pool_t *pool = stream->own_pool;
-	stream->own_pool = NULL;
-	pj_pool_release(pool);
-    }
+    pj_pool_safe_release(&stream->own_pool);
 
     return PJ_SUCCESS;
 }
@@ -1927,11 +2148,13 @@ PJ_DEF(pj_status_t) pjmedia_vid_stream_resume(pjmedia_vid_stream *stream,
 
     if ((dir & PJMEDIA_DIR_ENCODING) && stream->enc) {
 	stream->enc->paused = 0;
+	stream->force_keyframe = PJ_TRUE;
 	PJ_LOG(4,(stream->enc->port.info.name.ptr, "Encoder stream resumed"));
     }
 
     if ((dir & PJMEDIA_DIR_DECODING) && stream->dec) {
 	stream->dec->paused = 0;
+	stream->last_dec_seq = 0;
 	PJ_LOG(4,(stream->dec->port.info.name.ptr, "Decoder stream resumed"));
     }
 
@@ -1995,4 +2218,30 @@ pjmedia_vid_stream_rc_config_default(pjmedia_vid_stream_rc_config *cfg)
 }
 
 
+/*
+ * Initialize the video stream send keyframe with default settings.
+ */
+PJ_DEF(void)
+pjmedia_vid_stream_sk_config_default(pjmedia_vid_stream_sk_config *cfg)
+{
+    pj_bzero(cfg, sizeof(*cfg));
+    cfg->count = PJMEDIA_VID_STREAM_START_KEYFRAME_CNT;
+    cfg->interval = PJMEDIA_VID_STREAM_START_KEYFRAME_INTERVAL_MSEC;
+}
+
+
+/**
+ * Get RTP session information from video stream.
+ */
+PJ_DEF(pj_status_t)
+pjmedia_vid_stream_get_rtp_session_info(pjmedia_vid_stream *stream,
+				    pjmedia_stream_rtp_sess_info *session_info)
+{
+    session_info->rx_rtp = &stream->dec->rtp;
+    session_info->tx_rtp = &stream->enc->rtp;
+    session_info->rtcp = &stream->rtcp;
+    return PJ_SUCCESS;
+}
+
+
 #endif /* PJMEDIA_HAS_VIDEO */
