@@ -69,6 +69,17 @@ struct AccountRegConfig : public PersistentObject
      */
     SipHeaderVector	headers;
 
+    /**
+     * Additional parameters that will be appended in the Contact header
+     * of the registration requests. This will be appended after
+     * \a AccountSipConfig.contactParams;
+     *
+     * The parameters should be preceeded by semicolon, and all strings must
+     * be properly escaped. Example:
+     *	 ";my-param=X;another-param=Hi%20there"
+     */
+    string	    	contactParams;
+
     /**
      * Optional interval for registration, in seconds. If the value is zero,
      * default interval will be used (PJSUA_REG_INTERVAL, 300 seconds).
@@ -163,14 +174,14 @@ public:
      *
      * @param node		Container to read values from.
      */
-    virtual void readObject(const ContainerNode &node) throw(Error);
+    virtual void readObject(const ContainerNode &node) PJSUA2_THROW(Error);
 
     /**
      * Write this object to a container node.
      *
      * @param node		Container to write values to.
      */
-    virtual void writeObject(ContainerNode &node) const throw(Error);
+    virtual void writeObject(ContainerNode &node) const PJSUA2_THROW(Error);
 
 };
 
@@ -262,14 +273,14 @@ public:
      *
      * @param node		Container to read values from.
      */
-    virtual void readObject(const ContainerNode &node) throw(Error);
+    virtual void readObject(const ContainerNode &node) PJSUA2_THROW(Error);
 
     /**
      * Write this object to a container node.
      *
      * @param node		Container to write values to.
      */
-    virtual void writeObject(ContainerNode &node) const throw(Error);
+    virtual void writeObject(ContainerNode &node) const PJSUA2_THROW(Error);
 };
 
 /**
@@ -320,14 +331,14 @@ public:
      *
      * @param node		Container to read values from.
      */
-    virtual void readObject(const ContainerNode &node) throw(Error);
+    virtual void readObject(const ContainerNode &node) PJSUA2_THROW(Error);
 
     /**
      * Write this object to a container node.
      *
      * @param node		Container to write values to.
      */
-    virtual void writeObject(ContainerNode &node) const throw(Error);
+    virtual void writeObject(ContainerNode &node) const PJSUA2_THROW(Error);
 };
 
 /**
@@ -387,14 +398,14 @@ public:
      *
      * @param node		Container to read values from.
      */
-    virtual void readObject(const ContainerNode &node) throw(Error);
+    virtual void readObject(const ContainerNode &node) PJSUA2_THROW(Error);
 
     /**
      * Write this object to a container node.
      *
      * @param node		Container to write values to.
      */
-    virtual void writeObject(ContainerNode &node) const throw(Error);
+    virtual void writeObject(ContainerNode &node) const PJSUA2_THROW(Error);
 };
 
 /**
@@ -427,14 +438,14 @@ public:
      *
      * @param node		Container to read values from.
      */
-    virtual void readObject(const ContainerNode &node) throw(Error);
+    virtual void readObject(const ContainerNode &node) PJSUA2_THROW(Error);
 
     /**
      * Write this object to a container node.
      *
      * @param node		Container to write values to.
      */
-    virtual void writeObject(ContainerNode &node) const throw(Error);
+    virtual void writeObject(ContainerNode &node) const PJSUA2_THROW(Error);
 };
 
 /**
@@ -457,6 +468,13 @@ struct AccountNatConfig : public PersistentObject
      */
     pjsua_stun_use 	mediaStunUse;
 
+    /**
+     * Specify NAT64 options.
+     *
+     * Default: PJSUA_NAT64_DISABLED
+     */
+    pjsua_nat64_opt 	nat64Opt;
+
     /**
      * Enable ICE for the media transport.
      *
@@ -674,14 +692,206 @@ public:
      *
      * @param node		Container to read values from.
      */
-    virtual void readObject(const ContainerNode &node) throw(Error);
+    virtual void readObject(const ContainerNode &node) PJSUA2_THROW(Error);
 
     /**
      * Write this object to a container node.
      *
      * @param node		Container to write values to.
      */
-    virtual void writeObject(ContainerNode &node) const throw(Error);
+    virtual void writeObject(ContainerNode &node) const PJSUA2_THROW(Error);
+};
+
+/**
+ * SRTP crypto.
+ */
+struct SrtpCrypto
+{
+    /**
+     * Optional key. If empty, a random key will be autogenerated.
+     */
+    string	key;
+
+    /**
+     * Crypto name.
+     */
+    string	name;
+
+    /**
+     * Flags, bitmask from #pjmedia_srtp_crypto_option
+     */
+    unsigned	flags;
+
+public:
+    /**
+     * Convert from pjsip
+     */
+    void fromPj(const pjmedia_srtp_crypto &prm);
+
+    /**
+     * Convert to pjsip
+     */
+    pjmedia_srtp_crypto toPj() const;
+};
+
+/* Array of SRTP cryptos. */
+typedef std::vector<SrtpCrypto> SrtpCryptoVector;
+
+/**
+ * SRTP settings.
+ */
+struct SrtpOpt : public PersistentObject
+{
+    /**
+     * Specify SRTP cryptos. If empty, all crypto will be enabled.
+     * Available crypto can be enumerated using Endpoint::srtpCryptoEnum().
+     *
+     * Default: empty.
+     */
+    SrtpCryptoVector		cryptos;
+
+    /**
+     * Specify SRTP keying methods, valid keying method is defined in
+     * pjmedia_srtp_keying_method. If empty, all keying methods will be
+     * enabled with priority order: SDES, DTLS-SRTP.
+     *
+     * Default: empty.
+     */
+    IntVector			keyings;
+
+public:
+    /**
+     * Default constructor initializes with default values.
+     */
+    SrtpOpt();
+
+    /**
+     * Convert from pjsip
+     */
+    void fromPj(const pjsua_srtp_opt &prm);
+
+    /**
+     * Convert to pjsip
+     */
+    pjsua_srtp_opt toPj() const;
+
+public:
+    /**
+     * Read this object from a container node.
+     *
+     * @param node		Container to read values from.
+     */
+    virtual void readObject(const ContainerNode &node) PJSUA2_THROW(Error);
+
+    /**
+     * Write this object to a container node.
+     *
+     * @param node		Container to write values to.
+     */
+    virtual void writeObject(ContainerNode &node) const PJSUA2_THROW(Error);
+};
+
+/**
+ * RTCP Feedback capability.
+ */
+struct RtcpFbCap
+{
+    /**
+     * Specify the codecs to which the capability is applicable. Codec ID is
+     * using the same format as in pjmedia_codec_mgr_find_codecs_by_id() and
+     * pjmedia_vid_codec_mgr_find_codecs_by_id(), e.g: "L16/8000/1", "PCMU",
+     * "H264". This can also be an asterisk ("*") to represent all codecs.
+     */
+    string		    codecId;
+
+    /**
+     * Specify the RTCP Feedback type.
+     */
+    pjmedia_rtcp_fb_type    type;
+
+    /**
+     * Specify the type name if RTCP Feedback type is PJMEDIA_RTCP_FB_OTHER.
+     */
+    string		    typeName;
+
+    /**
+     * Specify the RTCP Feedback parameters.
+     */
+    string		    param;
+
+public:
+    /**
+     * Constructor.
+     */
+    RtcpFbCap() : type(PJMEDIA_RTCP_FB_OTHER)
+    {}
+
+    /**
+     * Convert from pjsip
+     */
+    void fromPj(const pjmedia_rtcp_fb_cap &prm);
+
+    /**
+     * Convert to pjsip
+     */
+    pjmedia_rtcp_fb_cap toPj() const;
+};
+
+/* Array of RTCP Feedback capabilities. */
+typedef std::vector<RtcpFbCap> RtcpFbCapVector;
+
+
+/**
+ * RTCP Feedback settings.
+ */
+struct RtcpFbConfig : public PersistentObject
+{
+    /**
+     * Specify whether transport protocol in SDP media description uses
+     * RTP/AVP instead of RTP/AVPF. Note that the standard mandates to signal
+     * AVPF profile, but it may cause SDP negotiation failure when negotiating
+     * with endpoints that does not support RTCP Feedback (including older
+     * version of PJSIP).
+     *
+     * Default: false.
+     */
+    bool		    dontUseAvpf;
+
+    /**
+     * RTCP Feedback capabilities.
+     */
+    RtcpFbCapVector	    caps;
+
+public:
+    /**
+     * Constructor.
+     */
+    RtcpFbConfig();
+
+    /**
+     * Convert from pjsip
+     */
+    void fromPj(const pjmedia_rtcp_fb_setting &prm);
+
+    /**
+     * Convert to pjsip
+     */
+    pjmedia_rtcp_fb_setting toPj() const;
+
+public:
+    /**
+     * Read this object from a container node.
+     *
+     * @param node		Container to read values from.
+     */
+    virtual void readObject(const ContainerNode &node) PJSUA2_THROW(Error);
+
+    /**
+     * Write this object to a container node.
+     *
+     * @param node		Container to write values to.
+     */
+    virtual void writeObject(ContainerNode &node) const PJSUA2_THROW(Error);
 };
 
 /**
@@ -735,25 +945,40 @@ struct AccountMediaConfig : public PersistentObject
      */
     int			srtpSecureSignaling;
 
+    /**
+     * Specify SRTP settings, like cryptos and keying methods.
+     */
+    SrtpOpt		srtpOpt;
+
     /**
      * Specify whether IPv6 should be used on media. Default is not used.
      */
     pjsua_ipv6_use	ipv6Use;
 
+    /**
+     * Enable RTP and RTCP multiplexing.
+     */
+    bool		rtcpMuxEnabled;
+
+    /**
+     * RTCP Feedback settings.
+     */
+    RtcpFbConfig	rtcpFbConfig;
+
 public:
     /**
      * Read this object from a container node.
      *
      * @param node		Container to read values from.
      */
-    virtual void readObject(const ContainerNode &node) throw(Error);
+    virtual void readObject(const ContainerNode &node) PJSUA2_THROW(Error);
 
     /**
      * Write this object to a container node.
      *
      * @param node		Container to write values to.
      */
-    virtual void writeObject(ContainerNode &node) const throw(Error);
+    virtual void writeObject(ContainerNode &node) const PJSUA2_THROW(Error);
 };
 
 /**
@@ -831,22 +1056,93 @@ struct AccountVideoConfig : public PersistentObject
      */
     unsigned			rateControlBandwidth;
 
+    /**
+     * The number of keyframe to be sent after the stream is created.
+     *
+     * Default: PJMEDIA_VID_STREAM_START_KEYFRAME_CNT
+     */
+    unsigned			    startKeyframeCount;
+
+    /**
+     * The keyframe sending interval after the stream is created.
+     *
+     * Default: PJMEDIA_VID_STREAM_START_KEYFRAME_INTERVAL_MSEC
+     */
+    unsigned			    startKeyframeInterval;
+
+
 public:
     /**
      * Read this object from a container node.
      *
      * @param node		Container to read values from.
      */
-    virtual void readObject(const ContainerNode &node) throw(Error);
+    virtual void readObject(const ContainerNode &node) PJSUA2_THROW(Error);
 
     /**
      * Write this object to a container node.
      *
      * @param node		Container to write values to.
      */
-    virtual void writeObject(ContainerNode &node) const throw(Error);
+    virtual void writeObject(ContainerNode &node) const PJSUA2_THROW(Error);
 };
 
+/**
+ * Account config specific to IP address change.
+ */
+typedef struct AccountIpChangeConfig
+{    
+    /**
+     * Shutdown the transport used for account registration. If this is set to
+     * PJ_TRUE, the transport will be shutdown altough it's used by multiple
+     * account. Shutdown transport will be followed by re-Registration if
+     * AccountConfig.natConfig.contactRewriteUse is enabled.
+     *
+     * Default: true
+     */
+    bool    		shutdownTp;
+
+    /**
+     * Hangup active calls associated with the acount. If this is set to true, 
+     * then the calls will be hang up.
+     *
+     * Default: false
+     */
+    bool		hangupCalls;
+
+    /**
+     * Specify the call flags used in the re-INVITE when \a hangupCalls is set 
+     * to false. If this is set to 0, no re-INVITE will be sent. The 
+     * re-INVITE will be sent after re-Registration is finished.
+     *
+     * Default: PJSUA_CALL_REINIT_MEDIA | PJSUA_CALL_UPDATE_CONTACT |
+     *          PJSUA_CALL_UPDATE_VIA
+     */
+    unsigned		reinviteFlags;
+
+public:
+    /**
+     * Virtual destructor
+     */
+    virtual ~AccountIpChangeConfig()
+    {}
+
+    /**
+     * Read this object from a container node.
+     *
+     * @param node		Container to read values from.
+     */
+    virtual void readObject(const ContainerNode &node) PJSUA2_THROW(Error);
+
+    /**
+     * Write this object to a container node.
+     *
+     * @param node		Container to write values to.
+     */
+    virtual void writeObject(ContainerNode &node) const PJSUA2_THROW(Error);
+    
+} AccountIpChangeConfig;
+
 /**
  * Account configuration.
  */
@@ -908,6 +1204,11 @@ struct AccountConfig : public PersistentObject
      */
     AccountVideoConfig	videoConfig;
 
+    /**
+     * IP Change settings.
+     */
+    AccountIpChangeConfig ipChangeConfig;
+
 public:
     /**
      * Default constructor will initialize with default values.
@@ -933,14 +1234,14 @@ public:
      *
      * @param node		Container to read values from.
      */
-    virtual void readObject(const ContainerNode &node) throw(Error);
+    virtual void readObject(const ContainerNode &node) PJSUA2_THROW(Error);
 
     /**
      * Write this object to a container node.
      *
      * @param node		Container to write values to.
      */
-    virtual void writeObject(ContainerNode &node) const throw(Error);
+    virtual void writeObject(ContainerNode &node) const PJSUA2_THROW(Error);
 };
 
 
@@ -1319,17 +1620,44 @@ public:
     /**
      * Destructor. Note that if the account is deleted, it will also delete
      * the corresponding account in the PJSUA-LIB.
+     *
+     * If application implements a derived class, the derived class should
+     * call shutdown() in the beginning stage in its destructor, or
+     * alternatively application should call shutdown() before deleting
+     * the derived class instance. This is to avoid race condition between
+     * the derived class destructor and Account callbacks.
      */
     virtual ~Account();
 
     /**
      * Create the account.
      *
+     * If application implements a derived class, the derived class should
+     * call shutdown() in the beginning stage in its destructor, or
+     * alternatively application should call shutdown() before deleting
+     * the derived class instance. This is to avoid race condition between
+     * the derived class destructor and Account callbacks.
+     *
      * @param cfg		The account config.
      * @param make_default	Make this the default account.
      */
     void create(const AccountConfig &cfg,
-                bool make_default=false) throw(Error);
+                bool make_default=false) PJSUA2_THROW(Error);
+
+    /**
+     * Shutdown the account. This will initiate unregistration if needed,
+     * and delete the corresponding account in the PJSUA-LIB.
+     *
+     * Note that application must delete all Buddy instances belong to this
+     * account before shutting down the account.
+     *
+     * If application implements a derived class, the derived class should
+     * call this method in the beginning stage in its destructor, or
+     * alternatively application should call this method before deleting
+     * the derived class instance. This is to avoid race condition between
+     * the derived class destructor and Account callbacks.
+     */
+    void shutdown();
 
     /**
      * Modify the account to use the specified account configuration.
@@ -1339,7 +1667,7 @@ public:
      * @param cfg 		New account config to be applied to the
      * 				account.
      */
-    void modify(const AccountConfig &cfg) throw(Error);
+    void modify(const AccountConfig &cfg) PJSUA2_THROW(Error);
 
     /**
      * Check if this account is still valid.
@@ -1351,10 +1679,8 @@ public:
     /**
      * Set this as default account to be used when incoming and outgoing
      * requests don't match any accounts.
-     *
-     * @return			PJ_SUCCESS on success.
      */
-    void setDefault() throw(Error);
+    void setDefault() PJSUA2_THROW(Error);
 
     /**
      * Check if this account is the default account. Default account will be
@@ -1386,7 +1712,7 @@ public:
      *
      * @return			Account info.
      */
-    AccountInfo getInfo() const throw(Error);
+    AccountInfo getInfo() const PJSUA2_THROW(Error);
 
     /**
      * Update registration or perform unregistration. Application normally
@@ -1396,7 +1722,7 @@ public:
      * @param renew		If False, this will start unregistration
      * 				process.
      */
-    void setRegistration(bool renew) throw(Error);
+    void setRegistration(bool renew) PJSUA2_THROW(Error);
 
     /**
      * Set or modify account's presence online status to be advertised to
@@ -1407,7 +1733,7 @@ public:
      *
      * @param pres_st		Presence online status.
      */
-    void setOnlineStatus(const PresenceStatus &pres_st) throw(Error);
+    void setOnlineStatus(const PresenceStatus &pres_st) PJSUA2_THROW(Error);
 
     /**
      * Lock/bind this account to a specific transport/listener. Normally
@@ -1423,7 +1749,7 @@ public:
      *
      * @param tp_id		The transport ID.
      */
-    void setTransport(TransportId tp_id) throw(Error);
+    void setTransport(TransportId tp_id) PJSUA2_THROW(Error);
 
     /**
      * Send NOTIFY to inform account presence status or to terminate server
@@ -1433,16 +1759,29 @@ public:
      *
      * @param prm		The sending NOTIFY parameter.
      */
-    void presNotify(const PresNotifyParam &prm) throw(Error);
+    void presNotify(const PresNotifyParam &prm) PJSUA2_THROW(Error);
     
+    /**
+     * Warning: deprecated, use enumBuddies2() instead. This function is not
+     * safe in multithreaded environment.
+     *
+     * Enumerate all buddies of the account.
+     *
+     * @return			The buddy list.
+     */
+    const BuddyVector& enumBuddies() const PJSUA2_THROW(Error);
+
     /**
      * Enumerate all buddies of the account.
      *
      * @return			The buddy list.
      */
-    const BuddyVector& enumBuddies() const throw(Error);
+    BuddyVector2 enumBuddies2() const PJSUA2_THROW(Error);
 
     /**
+     * Warning: deprecated, use findBuddy2 instead. This function is not
+     * safe in multithreaded environment.
+     *
      * Find a buddy in the buddy list with the specified URI. 
      *
      * Exception: if buddy is not found, PJ_ENOTFOUND will be thrown.
@@ -1453,19 +1792,18 @@ public:
      * @return			The pointer to buddy.
      */
     Buddy* findBuddy(string uri, FindBuddyMatch *buddy_match = NULL) const
-		    throw(Error);
+		    PJSUA2_THROW(Error);
 
     /**
-     * An internal function to add a Buddy to Account buddy list.
-     * This function must never be used by application.
+     * Find a buddy in the buddy list with the specified URI. 
+     *
+     * Exception: if buddy is not found, PJ_ENOTFOUND will be thrown.
+     *
+     * @param uri		The buddy URI.
+     *
+     * @return			The pointer to buddy.
      */
-    void addBuddy(Buddy *buddy);
-
-    /**
-     * An internal function to remove a Buddy from Account buddy list.
-     * This function must never be used by application.
-     */
-    void removeBuddy(Buddy *buddy);
+    Buddy findBuddy2(string uri) const PJSUA2_THROW(Error);
 
 public:
     /*
@@ -1572,8 +1910,23 @@ public:
     virtual void onMwiInfo(OnMwiInfoParam &prm)
     { PJ_UNUSED_ARG(prm); }
 
-protected:
+
+private:
     friend class Endpoint;
+    friend class Buddy;
+
+    /**
+     * An internal function to add a Buddy to Account buddy list.
+     * This method is used by Buddy::create().
+     */
+    void addBuddy(Buddy *buddy);
+
+    /**
+     * An internal function to remove a Buddy from Account buddy list.
+     * This method is used by Buddy::~Buddy().
+     */
+    void removeBuddy(Buddy *buddy);
+
 
 private:
     pjsua_acc_id 	 id;
