@@ -46,12 +46,6 @@ using std::vector;
 
 //////////////////////////////////////////////////////////////////////////////
 
-/**
- * Codec parameters, corresponds to pjmedia_codec_param or
- * pjmedia_vid_codec_param.
- */
-typedef void *CodecParam;
-
 /**
  * Media stream, corresponds to pjmedia_stream
  */
@@ -233,75 +227,37 @@ public:
     void fromPj(const pjmedia_sdp_session &sdp);
 };
 
-/**
- * This structure describes a media format changed event.
- */
-struct MediaFmtChangedEvent
-{
-    unsigned newWidth;      /**< The new width.     */
-    unsigned newHeight;     /**< The new height.    */
-};
-
-/**
- * Media event data.
- */
-typedef union MediaEventData {
-    /**
-     * Media format changed event data.
-     */
-    MediaFmtChangedEvent    fmtChanged;
-    
-    /**
-     * Pointer to storage to user event data, if it's outside
-     * this struct
-     */
-    GenericData		ptr;
-
-} MediaEventData;
-
-/**
- * This structure describes a media event. It corresponds to the
- * pjmedia_event structure.
- */
-struct MediaEvent
-{
-    /**
-     * The event type.
-     */
-    pjmedia_event_type          type;
-
-    /**
-     * Additional data/parameters about the event. The type of data
-     * will be specific to the event type being reported.
-     */
-    MediaEventData              data;
-    
-    /**
-     * Pointer to original pjmedia_event. Only valid when the struct
-     * is converted from PJSIP's pjmedia_event.
-     */
-    void                       *pjMediaEvent;
-
-public:
-    /**
-     * Convert from pjsip
-     */
-    void fromPj(const pjmedia_event &ev);
-};
 
 /**
  * This structure describes media transport informations. It corresponds to the
- * pjmedia_transport_info structure.
+ * pjmedia_transport_info structure. The address name field can be empty string
+ * if the address in the pjmedia_transport_info is invalid.
  */
 struct MediaTransportInfo
 {
     /**
-     * Remote address where RTP originated from.
+     * Address to be advertised as the local address for the RTP socket, 
+     * which does not need to be equal as the bound address (for example, 
+     * this address can be the address resolved with STUN).
+     */
+    SocketAddress   localRtpName;
+
+    /**
+     * Address to be advertised as the local address for the RTCP socket, 
+     * which does not need to be equal as the bound address (for example, 
+     * this address can be the address resolved with STUN).
+     */
+    SocketAddress   localRtcpName;
+
+    /**
+     * Remote address where RTP originated from. This can be empty string if 
+     * no data is received from the remote.
      */
     SocketAddress   srcRtpName;
 
     /**
-     * Remote address where RTCP originated from.
+     * Remote address where RTCP originated from. This can be empty string if 
+     * no data is recevied from the remote.
      */
     SocketAddress   srcRtcpName;
     
@@ -355,7 +311,7 @@ public:
     /**
      * Default constructor initializes with empty or default values.
      */
-    CallSetting(pj_bool_t useDefaultValues = false);
+    CallSetting(bool useDefaultValues = false);
 
     /**
      * Check if the settings are set with empty values.
@@ -377,6 +333,11 @@ public:
 
 /**
  * Call media information.
+ *
+ * Application can query conference bridge port of this media using
+ * Call::getAudioMedia() if the media type is audio,
+ * or Call::getEncodingVideoMedia()/Call::getDecodingVideoMedia()
+ * if the media type is video.
  */
 struct CallMediaInfo
 {
@@ -401,6 +362,9 @@ struct CallMediaInfo
     pjsua_call_media_status status;
     
     /**
+     * Warning: this is deprecated, application can query conference bridge
+     * port of this media using Call::getAudioMedia().
+     *
      * The conference port number for the call. Only valid if the media type
      * is audio.
      */
@@ -418,7 +382,7 @@ struct CallMediaInfo
      * the media type is video.
      */
     VideoWindow	    	    videoWindow;
-    
+
     /**
      * The video capture device for outgoing transmission, if any,
      * or PJMEDIA_VID_INVALID_DEV. Only valid if the media type is video.
@@ -607,9 +571,14 @@ struct StreamInfo
     unsigned            codecClockRate;
     
     /**
-     * Optional codec param.
+     * Optional audio codec param.
      */
-    CodecParam          codecParam;
+    CodecParam          audCodecParam;
+
+    /**
+     * Optional video codec param.
+     */
+    VidCodecParam       vidCodecParam;
 
 public:
     /**
@@ -692,15 +661,25 @@ struct OnCallSdpCreatedParam
 struct OnStreamCreatedParam
 {
     /**
-     * Media stream.
+     * Media stream, read-only.
      */
     MediaStream stream;
     
     /**
-     * Stream index in the media session.
+     * Stream index in the media session, read-only.
      */
     unsigned    streamIdx;
     
+    /**
+     * Specify if PJSUA2 should take ownership of the port returned in
+     * the pPort parameter below. If set to PJ_TRUE,
+     * pjmedia_port_destroy() will be called on the port when it is
+     * no longer needed.
+     *
+     * Default: PJ_FALSE
+     */
+    bool 	destroyPort;
+
     /**
      * On input, it specifies the media port of the stream. Application
      * may modify this pointer to point to different media port to be
@@ -732,10 +711,21 @@ struct OnStreamDestroyedParam
  */
 struct OnDtmfDigitParam
 {
+    /**
+     * DTMF sending method.
+     */
+    pjsua_dtmf_method	method;
+
     /**
      * DTMF ASCII digit.
      */
-    string      digit;
+    string		digit;
+
+    /**
+     * DTMF signal duration which might be included when sending DTMF using 
+     * SIP INFO.
+     */
+    unsigned		duration;
 };
 
 /**
@@ -747,19 +737,25 @@ struct OnCallTransferRequestParam
     /**
      * The destination where the call will be transferred to.
      */
-    string              dstUri;
+    string               dstUri;
     
     /**
      * Status code to be returned for the call transfer request. On input,
-     * it contains status code 200.
+     * it contains status code 202.
      */
-    pjsip_status_code   statusCode;
+    pjsip_status_code    statusCode;
     
     /**
      * The current call setting, application can update this setting
      * for the call being transferred.
      */
-    CallSetting         opt;
+    CallSetting          opt;
+
+    /**
+     * New Call derived object instantiated by application when the call
+     * transfer is about to be accepted.
+     */
+    Call		*newCall;
 };
 
 /**
@@ -829,7 +825,12 @@ struct OnCallReplacedParam
     /**
      * The new call id.
      */
-    pjsua_call_id       newCallId;
+    pjsua_call_id        newCallId;
+
+    /**
+     * New Call derived object instantiated by application.
+     */
+    Call		*newCall;
 };
 
 /**
@@ -856,6 +857,54 @@ struct OnCallRxOfferParam
     CallSetting         opt;
 };
 
+/**
+ * This structure contains parameters for Call::onCallRxReinvite() callback.
+ */
+struct OnCallRxReinviteParam
+{
+    /**
+     * The new offer received.
+     */
+    SdpSession          offer;
+
+    /**
+     * The incoming re-INVITE.
+     */
+    SipRxData           rdata;
+    
+    /**
+     * On input, it is false. Set to true if app wants to manually answer
+     * the re-INVITE.
+     */
+    bool		async;
+    
+    /**
+     * Status code to be returned for answering the offer. On input,
+     * it contains status code 200. Currently, valid values are only
+     * 200 and 488.
+     */
+    pjsip_status_code   statusCode;
+    
+    /**
+     * The current call setting, application can update this setting for
+     * answering the offer.
+     */
+    CallSetting         opt;
+};
+
+/**
+ * This structure contains parameters for Call::onCallTxOffer() callback.
+ */
+struct OnCallTxOfferParam
+{
+    /**
+     * The current call setting, application can update this setting for
+     * generating the offer. Note that application should maintain any
+     * active media to avoid the need for the peer to reject the offer.
+     */
+    CallSetting         opt;
+};
+
 /**
  * This structure contains parameters for Call::onCallRedirected() callback.
  */
@@ -944,6 +993,33 @@ struct OnCreateMediaTransportParam
     unsigned        flags;
 };
 
+/**
+ * This structure contains parameters for Call::onCreateMediaTransportSrtp()
+ * callback.
+ */
+struct OnCreateMediaTransportSrtpParam
+{
+    /**
+     * The media index in the SDP for which the SRTP media transport
+     * will be used.
+     */
+    unsigned        		mediaIdx;
+    
+    /**
+     * Specify whether secure media transport should be used. Application
+     * can modify this only for initial INVITE.
+     * Valid values are PJMEDIA_SRTP_DISABLED, PJMEDIA_SRTP_OPTIONAL, and
+     * PJMEDIA_SRTP_MANDATORY.
+     */
+    pjmedia_srtp_use		srtpUse;
+    
+    /**
+     * Application can modify this to specify the cryptos and keys
+     * which are going to be used.
+     */
+    SrtpCryptoVector		cryptos;
+};
+
 /**
  * @}  // PJSUA2_Call_Data_Structure
  */
@@ -986,6 +1062,11 @@ struct CallOpParam
      * answers/responses for this INVITE request.
      */
     SipTxOption         txOption;
+
+    /**
+     * SDP answer. Currently only used for Call::answer().
+     */
+    SdpSession		sdp;
     
 public:
     /**
@@ -1066,6 +1147,47 @@ public:
     CallVidSetStreamParam();
 };
 
+/**
+ * This structure contains parameters for Call::sendDtmf()
+ */
+struct CallSendDtmfParam
+{
+    /**
+     * The method used to send DTMF.
+     * 
+     * Default: PJSUA_DTMF_METHOD_RFC2833
+     */
+    pjsua_dtmf_method method;    
+
+    /**
+     * The signal duration used for the DTMF.
+     *
+     * Default: PJSUA_CALL_SEND_DTMF_DURATION_DEFAULT
+     */
+    unsigned duration;
+
+    /**
+     * The DTMF digits to be sent.
+     */
+    string digits;
+
+public:
+    /**
+     * Default constructor initialize with default value.     
+     */
+    CallSendDtmfParam();
+
+    /**
+     * Convert to pjsip.
+     */
+    pjsua_call_send_dtmf_param toPj() const;
+
+    /**
+     * Convert from pjsip.
+     */
+    void fromPj(const pjsua_call_send_dtmf_param &param);
+};
+
 /**
  * Call.
  */
@@ -1087,7 +1209,7 @@ public:
      *
      * @return              Call info.
      */
-    CallInfo getInfo() const throw(Error);
+    CallInfo getInfo() const PJSUA2_THROW(Error);
     
     /**
      * Check if this call has active INVITE session and the INVITE
@@ -1121,6 +1243,9 @@ public:
     bool hasMedia() const;
     
     /**
+     * Warning: deprecated, use getAudioMedia() instead. This function is not
+     * safe in multithreaded environment.
+     *
      * Get media for the specified media index.
      *
      * @param med_idx       Media index.
@@ -1129,6 +1254,43 @@ public:
      */
     Media *getMedia(unsigned med_idx) const;
 
+    /**
+     * Get audio media for the specified media index. If the specified media
+     * index is not audio or invalid or inactive, exception will be thrown.
+     *
+     * @param med_idx       Media index, or -1 to specify any first audio
+     *                      media registered in the conference bridge.
+     *
+     * @return              The audio media.
+     */
+    AudioMedia getAudioMedia(int med_idx) const PJSUA2_THROW(Error);
+
+    /**
+     * Get video media in encoding direction for the specified media index.
+     * If the specified media index is not video or invalid or the direction
+     * is receive only, exception will be thrown.
+     *
+     * @param med_idx       Media index, or -1 to specify any first video
+     *			    media with encoding direction registered in the
+     *			    conference bridge.
+     *
+     * @return              The video media.
+     */
+    VideoMedia getEncodingVideoMedia(int med_idx) const PJSUA2_THROW(Error);
+
+    /**
+     * Get video media in decoding direction for the specified media index.
+     * If the specified media index is not video or invalid or the direction
+     * is send only, exception will be thrown.
+     *
+     * @param med_idx       Media index, or -1 to specify any first video
+     *			    media with decoding direction registered in the
+     *			    conference bridge.
+     *
+     * @return              The video media.
+     */
+    VideoMedia getDecodingVideoMedia(int med_idx) const PJSUA2_THROW(Error);
+
     /**
      * Check if remote peer support the specified capability.
      *
@@ -1186,7 +1348,7 @@ public:
      *
      * @see Endpoint::natGetType(), natTypeInSdp
      */
-    pj_stun_nat_type getRemNatType() throw(Error);
+    pj_stun_nat_type getRemNatType() PJSUA2_THROW(Error);
 
     /**
      * Make outgoing call to the specified URI.
@@ -1197,7 +1359,8 @@ public:
      * @param prm.txOption  Optional headers etc to be added to outgoing INVITE
      *                      request.
      */
-    void makeCall(const string &dst_uri, const CallOpParam &prm) throw(Error);
+    void makeCall(const string &dst_uri, const CallOpParam &prm)
+		  PJSUA2_THROW(Error);
 
     /**
      * Send response to incoming INVITE request with call setting param.
@@ -1221,7 +1384,7 @@ public:
      *                      be persistent in all next answers/responses for this
      *                      INVITE request.
      */
-    void answer(const CallOpParam &prm) throw(Error);
+    void answer(const CallOpParam &prm) PJSUA2_THROW(Error);
     
     /**
      * Hangup call by using method that is appropriate according to the
@@ -1241,7 +1404,7 @@ public:
      * @param prm.txOption  Optional list of headers etc to be added to outgoing
      *                      request/response message.
      */
-    void hangup(const CallOpParam &prm) throw(Error);
+    void hangup(const CallOpParam &prm) PJSUA2_THROW(Error);
     
     /**
      * Put the specified call on hold. This will send re-INVITE with the
@@ -1255,7 +1418,7 @@ public:
      * @param prm.txOption  Optional message components to be sent with
      *                      the request.
      */
-    void setHold(const CallOpParam &prm) throw(Error);
+    void setHold(const CallOpParam &prm) PJSUA2_THROW(Error);
     
     /**
      * Send re-INVITE.
@@ -1270,7 +1433,7 @@ public:
      * @param prm.txOption  Optional message components to be sent with
      *                      the request.
      */
-    void reinvite(const CallOpParam &prm) throw(Error);
+    void reinvite(const CallOpParam &prm) PJSUA2_THROW(Error);
     
     /**
      * Send UPDATE request.
@@ -1280,7 +1443,7 @@ public:
      * @param prm.txOption  Optional message components to be sent with
      *                      the request.
      */
-    void update(const CallOpParam &prm) throw(Error);
+    void update(const CallOpParam &prm) PJSUA2_THROW(Error);
     
     /**
      * Initiate call transfer to the specified address. This function will send
@@ -1297,7 +1460,7 @@ public:
      * @param prm.txOption  Optional message components to be sent with
      *                      the request.
      */
-    void xfer(const string &dest, const CallOpParam &prm) throw(Error);
+    void xfer(const string &dest, const CallOpParam &prm) PJSUA2_THROW(Error);
 
     /**
      * Initiate attended call transfer. This function will send REFER request
@@ -1315,7 +1478,7 @@ public:
      *                      the request.
      */
     void xferReplaces(const Call& dest_call,
-                      const CallOpParam &prm) throw(Error);
+                      const CallOpParam &prm) PJSUA2_THROW(Error);
     
     /**
      * Accept or reject redirection response. Application MUST call this
@@ -1339,14 +1502,21 @@ public:
      *                      callback, except that the PJSIP_REDIRECT_PENDING is
      *                      not accepted here.
      */
-    void processRedirect(pjsip_redirect_op cmd) throw(Error);
+    void processRedirect(pjsip_redirect_op cmd) PJSUA2_THROW(Error);
 
     /**
      * Send DTMF digits to remote using RFC 2833 payload formats.
      *
      * @param digits        DTMF string digits to be sent.
      */
-    void dialDtmf(const string &digits) throw(Error);
+    void dialDtmf(const string &digits) PJSUA2_THROW(Error);
+
+    /**
+     * Send DTMF digits to remote.
+     *
+     * @param param	The send DTMF parameter.
+     */
+    void sendDtmf(const CallSendDtmfParam &param) PJSUA2_THROW(Error);
     
     /**
      * Send instant messaging inside INVITE session.
@@ -1360,7 +1530,8 @@ public:
      * @param prm.userData  Optional user data, which will be given back when
      *                      the IM callback is called.
      */
-    void sendInstantMessage(const SendInstantMessageParam& prm) throw(Error);
+    void sendInstantMessage(const SendInstantMessageParam& prm)
+			    PJSUA2_THROW(Error);
     
     /**
      * Send IM typing indication inside INVITE session.
@@ -1371,7 +1542,7 @@ public:
      *                      outgoing request.
      */
     void sendTypingIndication(const SendTypingIndicationParam &prm)
-         throw(Error);
+         PJSUA2_THROW(Error);
     
     /**
      * Send arbitrary request with the call. This is useful for example to send
@@ -1383,7 +1554,7 @@ public:
      * @param prm.txOption  Optional message body and/or list of headers to be
      *                      included in outgoing request.
      */
-    void sendRequest(const CallSendRequestParam &prm) throw(Error);
+    void sendRequest(const CallSendRequestParam &prm) PJSUA2_THROW(Error);
     
     /**
      * Dump call and media statistics to string.
@@ -1393,7 +1564,7 @@ public:
      *
      * @return              Call dump and media statistics string.
      */
-    string dump(bool with_media, const string indent) throw(Error);
+    string dump(bool with_media, const string indent) PJSUA2_THROW(Error);
     
     /**
      * Get the media stream index of the default video stream in the call.
@@ -1431,7 +1602,7 @@ public:
      *                      (see CallVidSetStreamParam).
      */
     void vidSetStream(pjsua_call_vid_strm_op op,
-                      const CallVidSetStreamParam &param) throw(Error);
+                      const CallVidSetStreamParam &param) PJSUA2_THROW(Error);
 
     /**
      * Get media stream info for the specified media index.
@@ -1440,7 +1611,7 @@ public:
      *
      * @return              The stream info.
      */
-    StreamInfo getStreamInfo(unsigned med_idx) const throw(Error);
+    StreamInfo getStreamInfo(unsigned med_idx) const PJSUA2_THROW(Error);
     
     /**
      * Get media stream statistic for the specified media index.
@@ -1449,7 +1620,7 @@ public:
      *
      * @return              The stream statistic.
      */
-    StreamStat getStreamStat(unsigned med_idx) const throw(Error);
+    StreamStat getStreamStat(unsigned med_idx) const PJSUA2_THROW(Error);
     
     /**
      * Get media transport info for the specified media index.
@@ -1458,7 +1629,8 @@ public:
      *
      * @return              The transport info.
      */
-    MediaTransportInfo getMedTransportInfo(unsigned med_idx) const throw(Error);
+    MediaTransportInfo getMedTransportInfo(unsigned med_idx) const
+					   PJSUA2_THROW(Error);
 
     /**
      * Internal function (callled by Endpoint( to process update to call
@@ -1551,10 +1723,21 @@ public:
     
     /**
      * Notify application on call being transferred (i.e. REFER is received).
-     * Application can decide to accept/reject transfer request
-     * by setting the code (default is 202). When this callback
-     * is not implemented, the default behavior is to accept the
-     * transfer.
+     * Application can decide to accept/reject transfer request by setting
+     * the code (default is 202). When this callback is not implemented,
+     * the default behavior is to accept the transfer.
+     *
+     * If application decides to accept the transfer request, it must also
+     * instantiate the new Call object for the transfer operation and return
+     * this new Call object to prm.newCall.
+     * 
+     * If application does not specify new Call object, library will reuse the
+     * existing Call object for initiating the new call (to the transfer
+     * destination). In this case, any events from both calls (transferred and
+     * transferring) will be delivered to the same Call object, where the call
+     * ID will be switched back and forth between callbacks. Application must
+     * be careful to not destroy the Call object when receiving disconnection
+     * event of the transferred call after the transfer process is completed.
      *
      * @param prm	Callback parameter.
      */
@@ -1587,7 +1770,11 @@ public:
      * request with Replaces header.
      *
      * After this callback is called, normally PJSUA-API will disconnect
-     * this call and establish a new call \a newCallId.
+     * this call and establish a new call. To be able to control the call,
+     * e.g: hold, transfer, change media parameters, application must
+     * instantiate a new Call object for the new call using call ID
+     * specified in prm.newCallId, and return the Call object via
+     * prm.newCall.
      *
      * @param prm	Callback parameter.
      */
@@ -1608,6 +1795,49 @@ public:
     virtual void onCallRxOffer(OnCallRxOfferParam &prm)
     { PJ_UNUSED_ARG(prm); }
     
+    /**
+     * Notify application when call has received a re-INVITE offer from
+     * the peer. It allows more fine-grained control over the response to
+     * a re-INVITE. If application sets async to PJ_TRUE, it can send
+     * the reply manually using the function #Call::answer() and setting
+     * the SDP answer. Otherwise, by default the re-INVITE will be
+     * answered automatically after the callback returns.
+     *
+     * Currently, this callback is only called for re-INVITE with
+     * SDP, but app should be prepared to handle the case of re-INVITE
+     * without SDP.
+     *
+     * Remarks: If manually answering at a later timing, application may
+     * need to monitor onCallTsxState() callback to check whether
+     * the re-INVITE is already answered automatically with 487 due to
+     * being cancelled.
+     *
+     * Note: onCallRxOffer() will still be called after this callback,
+     * but only if prm.async is false and prm.code is 200. 
+     */
+    virtual void onCallRxReinvite(OnCallRxReinviteParam &prm)
+    { PJ_UNUSED_ARG(prm); }
+
+
+    /**
+     * Notify application when call has received INVITE with no SDP offer.
+     * Application can update the call setting (e.g: add audio/video), or
+     * enable/disable codecs, or update other media session settings from
+     * within the callback, however, as mandated by the standard (RFC3261
+     * section 14.2), it must ensure that the update overlaps with the
+     * existing media session (in codecs, transports, or other parameters)
+     * that require support from the peer, this is to avoid the need for
+     * the peer to reject the offer.
+     *
+     * When this callback is not implemented, the default behavior is to send
+     * SDP offer using current active media session (with all enabled codecs
+     * on each media type).
+     *
+     * @param prm	Callback parameter.
+     */
+    virtual void onCallTxOffer(OnCallTxOfferParam &prm)
+    { PJ_UNUSED_ARG(prm); }
+    
     /**
      * Notify application on incoming MESSAGE request.
      *
@@ -1717,11 +1947,34 @@ public:
     onCreateMediaTransport(OnCreateMediaTransportParam &prm)
     { PJ_UNUSED_ARG(prm); }
 
+    /**
+     * Warning: deprecated and may be removed in future release.
+     * Application can set SRTP crypto settings (including keys) and
+     * keying methods via AccountConfig.mediaConfig.srtpOpt.
+     * See also ticket #2100.
+     *
+     * This callback is called when SRTP media transport is created.
+     * Application can modify the SRTP setting \a srtpOpt to specify
+     * the cryptos and keys which are going to be used. Note that
+     * application should not modify the field
+     * \a pjmedia_srtp_setting.close_member_tp and can only modify
+     * the field \a pjmedia_srtp_setting.use for initial INVITE.
+     *
+     * @param prm	Callback parameter.
+     */
+    virtual void
+    onCreateMediaTransportSrtp(OnCreateMediaTransportSrtpParam &prm)
+    { PJ_UNUSED_ARG(prm); }
+
 private:
+    friend class Endpoint;
+
     Account             &acc;
     pjsua_call_id 	 id;
     Token                userData;
     std::vector<Media *> medias;
+    pj_pool_t		*sdp_pool;
+    Call		*child;	    /* New outgoing call in call transfer.  */
 };
 
 /**
