@@ -39,6 +39,23 @@
 #   define PJSUA2_ERROR_HAS_EXTRA_INFO		1
 #endif
 
+/**
+ * Maximum buffer length to print SDP content for SdpSession. Set this to 0
+ * if the printed SDP is not needed.
+ */
+#ifndef PJSUA2_MAX_SDP_BUF_LEN
+#   define PJSUA2_MAX_SDP_BUF_LEN		1024
+#endif
+
+
+/*
+ * C++11 deprecated dynamic exception specification, but SWIG needs it.
+ */
+#ifndef SWIG
+#   define PJSUA2_THROW(x)
+#else
+#   define PJSUA2_THROW(x) throw(x)
+#endif
 
 /**
  * @}  PJSUA2_CFG
