@@ -222,6 +222,8 @@ PJ_DEF(pj_status_t) pjsip_publishc_destroy(pjsip_publishc *pubc)
 
 	if (pubc->mutex)
 	    pj_mutex_destroy(pubc->mutex);
+
+	pjsip_auth_clt_deinit(&pubc->auth_sess);
 	pjsip_endpt_release_pool(pubc->endpt, pubc->pool);
     }
 
@@ -434,14 +436,14 @@ static pj_status_t create_request(pjsip_publishc *pubc,
 
     /* Add user headers */
     if (!pj_list_empty(&pubc->usr_hdr)) {
-	const pjsip_hdr *hdr;
+	const pjsip_hdr *uhdr;
 
-	hdr = pubc->usr_hdr.next;
-	while (hdr != &pubc->usr_hdr) {
+	uhdr = pubc->usr_hdr.next;
+	while (uhdr != &pubc->usr_hdr) {
 	    pjsip_hdr *new_hdr = (pjsip_hdr*)
-	    			 pjsip_hdr_shallow_clone(tdata->pool, hdr);
+	    			 pjsip_hdr_shallow_clone(tdata->pool, uhdr);
 	    pjsip_msg_add_hdr(tdata->msg, new_hdr);
-	    hdr = hdr->next;
+	    uhdr = uhdr->next;
 	}
     }
 
@@ -801,7 +803,7 @@ PJ_DEF(pj_status_t) pjsip_publishc_send(pjsip_publishc *pubc,
 	// already decremented pending_tsx. Decrementing this here may 
 	// cause accessing freed memory location.
 	//--pubc->pending_tsx;
-	PJ_LOG(4,(THIS_FILE, "Error sending request, status=%d", status));
+	PJ_PERROR(4,(THIS_FILE, status, "Error sending request"));
     }
 
     return status;
