@@ -1,4 +1,4 @@
-/* $Id: sip_transport.c 5097 2015-05-18 04:42:42Z ming $ */
+/* $Id$ */
 /* 
  * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
  * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
@@ -54,6 +54,25 @@ static const char *addr_string(const pj_sockaddr_t *addr)
 #   define TRACE_(x)
 #endif
 
+/* Specify the initial size of the transport manager's pool. */
+#ifndef  TPMGR_POOL_INIT_SIZE
+#   define TPMGR_POOL_INIT_SIZE	64
+#endif
+
+/* Specify the increment size of the transport manager's pool. */
+#ifndef TPMGR_POOL_INC_SIZE
+    #define TPMGR_POOL_INC_SIZE	64
+#endif
+
+/* Specify transport entry allocation count. When registering a new transport,
+ * a new entry will be picked from a free list. This setting will determine
+ * the size of the free list size. If all entry is used, then the same number
+ * of entry will be allocated.
+ */
+#ifndef PJSIP_TRANSPORT_ENTRY_ALLOC_CNT
+#   define PJSIP_TRANSPORT_ENTRY_ALLOC_CNT  16
+#endif
+
 /* Prototype. */
 static pj_status_t mod_on_tx_msg(pjsip_tx_data *tdata);
 
@@ -81,6 +100,7 @@ static pjsip_module mod_msg_print =
 typedef struct transport
 {
     PJ_DECL_LIST_MEMBER(struct transport);
+    pj_hash_entry_buf tp_buf;
     pjsip_transport *tp;
 } transport;
 
@@ -93,6 +113,7 @@ struct pjsip_tpmgr
     pj_lock_t	    *lock;
     pjsip_endpoint  *endpt;
     pjsip_tpfactory  factory_list;
+    pj_pool_t	    *pool;
 #if defined(PJ_DEBUG) && PJ_DEBUG!=0
     pj_atomic_t	    *tdata_counter;
 #endif
@@ -105,12 +126,9 @@ struct pjsip_tpmgr
      * is destroyed.
      */
     pjsip_tx_data    tdata_list;
-    
-    /* List of transports which are NOT stored in the hash table, so
-     * that it can be properly cleaned up when transport manager
-     * is destroyed.
-     */
-    transport        tp_list;
+
+    /* List of free transport entry. */
+    transport	     tp_entry_freelist;
 };
 
 
@@ -253,6 +271,7 @@ PJ_DEF(pj_status_t) pjsip_transport_register_type( unsigned tp_flag,
 						   int *p_tp_type)
 {
     unsigned i;
+    pjsip_transport_type_e parent = 0;
 
     PJ_ASSERT_RETURN(tp_flag && tp_name && def_port, PJ_EINVAL);
     PJ_ASSERT_RETURN(pj_ansi_strlen(tp_name) < 
@@ -260,6 +279,11 @@ PJ_DEF(pj_status_t) pjsip_transport_register_type( unsigned tp_flag,
 		     PJ_ENAMETOOLONG);
 
     for (i=1; i<PJ_ARRAY_SIZE(transport_names); ++i) {
+        if (tp_flag & PJSIP_TRANSPORT_IPV6 && 
+            pj_stricmp2(&transport_names[i].name, tp_name) == 0)
+        {
+	    parent = transport_names[i].type;
+        }
 	if (transport_names[i].type == 0)
 	    break;
     }
@@ -267,14 +291,19 @@ PJ_DEF(pj_status_t) pjsip_transport_register_type( unsigned tp_flag,
     if (i == PJ_ARRAY_SIZE(transport_names))
 	return PJ_ETOOMANY;
 
-    transport_names[i].type = (pjsip_transport_type_e)i;
+    if (tp_flag & PJSIP_TRANSPORT_IPV6 && parent) {
+        transport_names[i].type = parent | PJSIP_TRANSPORT_IPV6;
+    } else {
+        transport_names[i].type = (pjsip_transport_type_e)i;
+    }
+
     transport_names[i].port = (pj_uint16_t)def_port;
     pj_ansi_strcpy(transport_names[i].name_buf, tp_name);
     transport_names[i].name = pj_str(transport_names[i].name_buf);
     transport_names[i].flag = tp_flag;
 
     if (p_tp_type)
-	*p_tp_type = i;
+	*p_tp_type = transport_names[i].type;
 
     return PJ_SUCCESS;
 }
@@ -422,7 +451,8 @@ PJ_DEF(pj_status_t) pjsip_tx_data_create( pjsip_tpmgr *mgr,
     tdata = PJ_POOL_ZALLOC_T(pool, pjsip_tx_data);
     tdata->pool = pool;
     tdata->mgr = mgr;
-    pj_memcpy(tdata->obj_name, pool->obj_name, PJ_MAX_OBJ_NAME);
+    pj_ansi_snprintf(tdata->obj_name, sizeof(tdata->obj_name), "tdta%p", tdata);
+    pj_memcpy(pool->obj_name, tdata->obj_name, sizeof(pool->obj_name));
 
     status = pj_atomic_create(tdata->pool, 0, &tdata->ref_cnt);
     if (status != PJ_SUCCESS) {
@@ -491,8 +521,13 @@ static void tx_data_destroy(pjsip_tx_data *tdata)
  */
 PJ_DEF(pj_status_t) pjsip_tx_data_dec_ref( pjsip_tx_data *tdata )
 {
-    pj_assert( pj_atomic_get(tdata->ref_cnt) > 0);
-    if (pj_atomic_dec_and_get(tdata->ref_cnt) <= 0) {
+    pj_atomic_value_t ref_cnt;
+    
+    PJ_ASSERT_RETURN(tdata && tdata->ref_cnt, PJ_EINVAL);
+
+    ref_cnt = pj_atomic_dec_and_get(tdata->ref_cnt);
+    pj_assert( ref_cnt >= 0);
+    if (ref_cnt == 0) {
 	tx_data_destroy(tdata);
 	return PJSIP_EBUFDESTROYED;
     } else {
@@ -626,6 +661,53 @@ PJ_DEF(pj_status_t) pjsip_tx_data_set_transport(pjsip_tx_data *tdata,
     return PJ_SUCCESS;
 }
 
+/* Clone pjsip_tx_data. */
+PJ_DEF(pj_status_t) pjsip_tx_data_clone(const pjsip_tx_data *src,
+                                        unsigned flags,
+				  	pjsip_tx_data ** p_tdata)
+{
+    pjsip_tx_data *dst;
+    const pjsip_hdr *hsrc;
+    pjsip_msg *msg;
+    pj_status_t status;
+
+    PJ_UNUSED_ARG(flags);
+
+    status = pjsip_tx_data_create(src->mgr, p_tdata);
+    if (status != PJ_SUCCESS)
+	return status;
+
+    dst = *p_tdata;
+
+    msg = pjsip_msg_create(dst->pool, PJSIP_RESPONSE_MSG);
+    dst->msg = msg;
+    pjsip_tx_data_add_ref(dst);
+
+    /* Duplicate status line */
+    msg->line.status.code = src->msg->line.status.code;
+    pj_strdup(dst->pool, &msg->line.status.reason,
+	      &src->msg->line.status.reason);
+
+    /* Duplicate all headers */
+    hsrc = src->msg->hdr.next;
+    while (hsrc != &src->msg->hdr) {
+	pjsip_hdr *h = (pjsip_hdr*) pjsip_hdr_clone(dst->pool, hsrc);
+	pjsip_msg_add_hdr(msg, h);
+	hsrc = hsrc->next;
+    }
+
+    /* Duplicate message body */
+    if (src->msg->body)
+	msg->body = pjsip_msg_body_clone(dst->pool, src->msg->body);
+
+    dst->is_pending = src->is_pending;
+
+    PJ_LOG(5,(THIS_FILE,
+	     "Tx data %s cloned",
+	     pjsip_tx_data_get_info(dst)));
+
+    return PJ_SUCCESS;
+}
 
 PJ_DEF(char*) pjsip_rx_data_get_info(pjsip_rx_data *rdata)
 {
@@ -948,27 +1030,69 @@ static void transport_idle_callback(pj_timer_heap_t *timer_heap,
 
     PJ_UNUSED_ARG(timer_heap);
 
+    if (entry->id == PJ_FALSE)
+	return;
+
     entry->id = PJ_FALSE;
     pjsip_transport_destroy(tp);
 }
 
+
+static pj_bool_t is_transport_valid(pjsip_transport *tp, pjsip_tpmgr *tpmgr,
+				    const pjsip_transport_key *key,
+				    int key_len)
+{
+    transport *tp_entry;
+
+    tp_entry = (transport *)pj_hash_get(tpmgr->table, key, key_len, NULL);
+    if (tp_entry != NULL) {
+
+	transport *tp_iter = tp_entry;
+	do {
+	    if (tp_iter->tp == tp) {
+		return PJ_TRUE;
+	    }
+	    tp_iter = tp_iter->next;
+	} while (tp_iter != tp_entry);
+    }
+
+    return PJ_FALSE;
+}
+
 /*
  * Add ref.
  */
 PJ_DEF(pj_status_t) pjsip_transport_add_ref( pjsip_transport *tp )
 {
+    pjsip_tpmgr *tpmgr;
+    pjsip_transport_key key;
+    int key_len;
+
     PJ_ASSERT_RETURN(tp != NULL, PJ_EINVAL);
 
+    /* Add ref transport group lock, if any */
+    if (tp->grp_lock)
+	pj_grp_lock_add_ref(tp->grp_lock);
+
+    /* Cache some vars for checking transport validity later */
+    tpmgr = tp->tpmgr;
+    key_len = sizeof(tp->key.type) + tp->addr_len;
+    pj_memcpy(&key, &tp->key, key_len);
+
     if (pj_atomic_inc_and_get(tp->ref_cnt) == 1) {
-	pj_lock_acquire(tp->tpmgr->lock);
-	/* Verify again. */
-	if (pj_atomic_get(tp->ref_cnt) == 1) {
+	pj_lock_acquire(tpmgr->lock);
+	/* Verify again. But first, make sure transport is still valid
+	 * (see #1883).
+	 */
+	if (is_transport_valid(tp, tpmgr, &key, key_len) &&
+	    pj_atomic_get(tp->ref_cnt) == 1)
+	{
 	    if (tp->idle_timer.id != PJ_FALSE) {
-		pjsip_endpt_cancel_timer(tp->tpmgr->endpt, &tp->idle_timer);
 		tp->idle_timer.id = PJ_FALSE;
+		pjsip_endpt_cancel_timer(tp->tpmgr->endpt, &tp->idle_timer);
 	    }
 	}
-	pj_lock_release(tp->tpmgr->lock);
+	pj_lock_release(tpmgr->lock);
     }
 
     return PJ_SUCCESS;
@@ -979,16 +1103,27 @@ PJ_DEF(pj_status_t) pjsip_transport_add_ref( pjsip_transport *tp )
  */
 PJ_DEF(pj_status_t) pjsip_transport_dec_ref( pjsip_transport *tp )
 {
-    PJ_ASSERT_RETURN(tp != NULL, PJ_EINVAL);
+    pjsip_tpmgr *tpmgr;
+    pjsip_transport_key key;
+    int key_len;
 
+    PJ_ASSERT_RETURN(tp != NULL, PJ_EINVAL);
     pj_assert(pj_atomic_get(tp->ref_cnt) > 0);
 
+    /* Cache some vars for checking transport validity later */
+    tpmgr = tp->tpmgr;
+    key_len = sizeof(tp->key.type) + tp->addr_len;
+    pj_memcpy(&key, &tp->key, key_len);
+
     if (pj_atomic_dec_and_get(tp->ref_cnt) == 0) {
-	pj_lock_acquire(tp->tpmgr->lock);
+	pj_lock_acquire(tpmgr->lock);
 	/* Verify again. Do not register timer if the transport is
-	 * being destroyed.
+	 * being destroyed. But first, make sure transport is still valid
+	 * (see #1883).
 	 */
-	if (pj_atomic_get(tp->ref_cnt) == 0 && !tp->is_destroying) {
+	if (is_transport_valid(tp, tpmgr, &key, key_len) &&
+	    !tp->is_destroying && pj_atomic_get(tp->ref_cnt) == 0)
+	{
 	    pj_time_val delay;
 	    
 	    /* If transport is in graceful shutdown, then this is the
@@ -1004,12 +1139,22 @@ PJ_DEF(pj_status_t) pjsip_transport_dec_ref( pjsip_transport *tp )
 		delay.msec = 0;
 	    }
 
-	    pj_assert(tp->idle_timer.id == 0);
-	    tp->idle_timer.id = PJ_TRUE;
-	    pjsip_endpt_schedule_timer(tp->tpmgr->endpt, &tp->idle_timer, 
-				       &delay);
+	    /* Avoid double timer entry scheduling */
+	    if (pj_timer_entry_running(&tp->idle_timer))
+		pjsip_endpt_cancel_timer(tp->tpmgr->endpt, &tp->idle_timer);
+
+	    pjsip_endpt_schedule_timer_w_grp_lock(tp->tpmgr->endpt,
+						  &tp->idle_timer,
+						  &delay,
+						  PJ_TRUE,
+						  tp->grp_lock);
 	}
-	pj_lock_release(tp->tpmgr->lock);
+	pj_lock_release(tpmgr->lock);
+    }
+
+    /* Dec ref transport group lock, if any */
+    if (tp->grp_lock) {
+	pj_grp_lock_dec_ref(tp->grp_lock);
     }
 
     return PJ_SUCCESS;
@@ -1024,7 +1169,8 @@ PJ_DEF(pj_status_t) pjsip_transport_register( pjsip_tpmgr *mgr,
 {
     int key_len;
     pj_uint32_t hval;
-    void *entry;
+    transport *tp_ref = NULL;
+    transport *tp_add = NULL;
 
     /* Init. */
     tp->tpmgr = mgr;
@@ -1032,39 +1178,57 @@ PJ_DEF(pj_status_t) pjsip_transport_register( pjsip_tpmgr *mgr,
     tp->idle_timer.user_data = tp;
     tp->idle_timer.cb = &transport_idle_callback;
 
-    /* 
+    /*
      * Register to hash table (see Trac ticket #42).
      */
     key_len = sizeof(tp->key.type) + tp->addr_len;
     pj_lock_acquire(mgr->lock);
 
-    /* If entry already occupied, unregister previous entry */
     hval = 0;
-    entry = pj_hash_get(mgr->table, &tp->key, key_len, &hval);
-    if (entry != NULL) {
-        transport *tp_ref;
-        
-        tp_ref = PJ_POOL_ZALLOC_T(((pjsip_transport *)entry)->pool, transport);
-        
-        /*
-         * Add transport to the list before removing it from the hash table.
-         * See ticket #1774 for more details.
-         */
-        tp_ref->tp = (pjsip_transport *)entry;
-        pj_list_push_back(&mgr->tp_list, tp_ref);
-	pj_hash_set(NULL, mgr->table, &tp->key, key_len, hval, NULL);
+    tp_ref = (transport *)pj_hash_get(mgr->table, &tp->key, key_len, &hval);
+
+    /* Get an empty entry from the freelist. */
+    if (pj_list_empty(&mgr->tp_entry_freelist)) {
+	unsigned i = 0;
+
+	TRACE_((THIS_FILE, "Transport list is full, allocate new entry"));
+	/* Allocate new entry for the freelist. */
+	for (; i < PJSIP_TRANSPORT_ENTRY_ALLOC_CNT; ++i) {
+	    tp_add = PJ_POOL_ZALLOC_T(mgr->pool, transport);
+	    if (!tp_add)
+		return PJ_ENOMEM;
+	    pj_list_init(tp_add);
+	    pj_list_push_back(&mgr->tp_entry_freelist, tp_add);
+	}
+    }
+    tp_add = mgr->tp_entry_freelist.next;
+    tp_add->tp = tp;
+    pj_list_erase(tp_add);
+
+    if (tp_ref) {
+	/* There'a already a transport list from the hash table. Add the 
+	 * new transport to the list.
+	 */
+	pj_list_push_back(tp_ref, tp_add);
+    } else {
+	/* Transport list not found, add it to the hash table. */
+	pj_hash_set_np(mgr->table, &tp->key, key_len, hval, tp_add->tp_buf,
+		       tp_add);
     }
 
-    /* Register new entry */
-    pj_hash_set(tp->pool, mgr->table, &tp->key, key_len, hval, tp);
+    /* Add ref transport group lock, if any */
+    if (tp->grp_lock)
+	pj_grp_lock_add_ref(tp->grp_lock);
 
     pj_lock_release(mgr->lock);
 
-    TRACE_((THIS_FILE,"Transport %s registered: type=%s, remote=%s:%d",
-		       tp->obj_name,
-		       pjsip_transport_get_type_name(tp->key.type),
-		       addr_string(&tp->key.rem_addr),
-		       pj_sockaddr_get_port(&tp->key.rem_addr)));
+    TRACE_((THIS_FILE, "Transport %s registered: type=%s, remote=%s:%d",
+	    tp->obj_name,
+	    pjsip_transport_get_type_name(tp->key.type),
+	    pj_sockaddr_has_addr(&tp->key.rem_addr)?
+				addr_string(&tp->key.rem_addr):"",
+	    pj_sockaddr_has_addr(&tp->key.rem_addr)?
+				pj_sockaddr_get_port(&tp->key.rem_addr):0));
 
     return PJ_SUCCESS;
 }
@@ -1077,20 +1241,20 @@ static pj_status_t destroy_transport( pjsip_tpmgr *mgr,
     pj_uint32_t hval;
     void *entry;
 
+    tp->is_destroying = PJ_TRUE;
+
     TRACE_((THIS_FILE, "Transport %s is being destroyed", tp->obj_name));
 
     pj_lock_acquire(tp->lock);
     pj_lock_acquire(mgr->lock);
 
-    tp->is_destroying = PJ_TRUE;
-
     /*
      * Unregister timer, if any.
      */
     //pj_assert(tp->idle_timer.id == PJ_FALSE);
     if (tp->idle_timer.id != PJ_FALSE) {
-	pjsip_endpt_cancel_timer(mgr->endpt, &tp->idle_timer);
 	tp->idle_timer.id = PJ_FALSE;
+	pjsip_endpt_cancel_timer(mgr->endpt, &tp->idle_timer);
     }
 
     /*
@@ -1099,23 +1263,47 @@ static pj_status_t destroy_transport( pjsip_tpmgr *mgr,
     key_len = sizeof(tp->key.type) + tp->addr_len;
     hval = 0;
     entry = pj_hash_get(mgr->table, &tp->key, key_len, &hval);
-    if (entry == (void*)tp) {
-	pj_hash_set(NULL, mgr->table, &tp->key, key_len, hval, NULL);
-    } else {
-        /* If not found in hash table, remove from the tranport list. */
-        transport *tp_iter = mgr->tp_list.next;
-        while (tp_iter != &mgr->tp_list) {
-            if (tp_iter->tp == tp) {
-                pj_list_erase(tp_iter);
-                break;
-            }
-            tp_iter = tp_iter->next;
-        }
+    if (entry) {
+	transport *tp_ref = (transport *)entry;
+	transport *tp_iter = tp_ref;
+	/* Search the matching entry from the transport list. */
+	do {
+	    if (tp_iter->tp == tp) {
+		transport *tp_next = tp_iter->next;
+
+		/* Update hash table :
+		 * - transport list only contain single element, or
+		 * - the entry is the first element of the transport list.
+		 */
+		if (tp_iter == tp_ref) {
+		    pj_hash_set(NULL, mgr->table, &tp->key, key_len, hval,
+				NULL);
+
+		    if (tp_ref->next != tp_ref) {
+			/* The transport list has multiple entry. */
+			pj_hash_set_np(mgr->table, &tp_next->tp->key, key_len,
+				       hval, tp_next->tp_buf, tp_next);
+		    }
+		}
+
+		pj_list_erase(tp_iter);
+		/* Put back to the transport freelist. */
+		pj_list_push_back(&mgr->tp_entry_freelist, tp_iter);
+
+		break;
+	    }
+	    tp_iter = tp_iter->next;
+	} while (tp_iter != tp_ref);
     }
 
     pj_lock_release(mgr->lock);
     pj_lock_release(tp->lock);
 
+    /* Dec ref transport group lock, if any */
+    if (tp->grp_lock) {
+	pj_grp_lock_dec_ref(tp->grp_lock);
+    }
+
     /* Destroy. */
     return tp->destroy(tp);
 }
@@ -1125,12 +1313,23 @@ static pj_status_t destroy_transport( pjsip_tpmgr *mgr,
  * Start graceful shutdown procedure for this transport. 
  */
 PJ_DEF(pj_status_t) pjsip_transport_shutdown(pjsip_transport *tp)
+{
+    return pjsip_transport_shutdown2(tp, PJ_FALSE);
+}
+
+
+/*
+ * Start shutdown procedure for this transport. 
+ */
+PJ_DEF(pj_status_t) pjsip_transport_shutdown2(pjsip_transport *tp,
+					      pj_bool_t force)
 {
     pjsip_tpmgr *mgr;
     pj_status_t status;
     pjsip_tp_state_callback state_cb;
 
-    TRACE_((THIS_FILE, "Transport %s shutting down", tp->obj_name));
+    PJ_LOG(4, (THIS_FILE, "Transport %s shutting down, force=%d",
+			  tp->obj_name, force));
 
     pj_lock_acquire(tp->lock);
 
@@ -1150,19 +1349,20 @@ PJ_DEF(pj_status_t) pjsip_transport_shutdown(pjsip_transport *tp)
     if (tp->do_shutdown)
 	status = tp->do_shutdown(tp);
 
+    if (status == PJ_SUCCESS)
+	tp->is_shutdown = PJ_TRUE;
+
     /* Notify application of transport shutdown */
     state_cb = pjsip_tpmgr_get_state_cb(tp->tpmgr);
     if (state_cb) {
 	pjsip_transport_state_info state_info;
 
 	pj_bzero(&state_info, sizeof(state_info));
-	state_info.status = status;
-        (*state_cb)(tp, PJSIP_TP_STATE_SHUTDOWN, &state_info);
+	state_info.status = PJ_ECANCELLED;
+	(*state_cb)(tp, (force? PJSIP_TP_STATE_DISCONNECTED:
+		    PJSIP_TP_STATE_SHUTDOWN), &state_info);
     }
 
-    if (status == PJ_SUCCESS)
-	tp->is_shutdown = PJ_TRUE;
-
     /* If transport reference count is zero, start timer count-down */
     if (pj_atomic_get(tp->ref_cnt) == 0) {
 	pjsip_transport_add_ref(tp);
@@ -1216,13 +1416,9 @@ PJ_DEF(pj_status_t) pjsip_tpmgr_register_tpfactory( pjsip_tpmgr *mgr,
 
     pj_lock_acquire(mgr->lock);
 
-    /* Check that no factory with the same type has been registered. */
+    /* Check that no same factory has been registered. */
     status = PJ_SUCCESS;
     for (p=mgr->factory_list.next; p!=&mgr->factory_list; p=p->next) {
-	if (p->type == tpf->type) {
-	    status = PJSIP_ETYPEEXISTS;
-	    break;
-	}
 	if (p == tpf) {
 	    status = PJ_EEXISTS;
 	    break;
@@ -1263,6 +1459,23 @@ PJ_DECL(void) pjsip_tpmgr_fla2_param_default(pjsip_tpmgr_fla2_param *prm)
     pj_bzero(prm, sizeof(*prm));
 }
 
+static pj_bool_t pjsip_tpmgr_is_tpfactory_valid(pjsip_tpmgr *mgr,
+						pjsip_tpfactory *tpf)
+{
+    pjsip_tpfactory *p;
+
+    pj_lock_acquire(mgr->lock);
+    for (p=mgr->factory_list.next; p!=&mgr->factory_list; p=p->next) {
+	if (p == tpf) {
+	    pj_lock_release(mgr->lock);
+	    return PJ_TRUE;
+	}
+    }
+    pj_lock_release(mgr->lock);
+
+    return PJ_FALSE;
+}
+
 /*****************************************************************************
  *
  * TRANSPORT MANAGER
@@ -1280,6 +1493,8 @@ PJ_DEF(pj_status_t) pjsip_tpmgr_create( pj_pool_t *pool,
 {
     pjsip_tpmgr *mgr;
     pj_status_t status;
+    unsigned i = 0;
+    pj_pool_t *mgr_pool;
 
     PJ_ASSERT_RETURN(pool && endpt && rx_cb && p_mgr, PJ_EINVAL);
 
@@ -1289,24 +1504,42 @@ PJ_DEF(pj_status_t) pjsip_tpmgr_create( pj_pool_t *pool,
 	return status;
 
     /* Create and initialize transport manager. */
-    mgr = PJ_POOL_ZALLOC_T(pool, pjsip_tpmgr);
+    mgr_pool = pjsip_endpt_create_pool(endpt, "tpmgr",
+				       TPMGR_POOL_INIT_SIZE,
+				       TPMGR_POOL_INC_SIZE);
+    mgr = PJ_POOL_ZALLOC_T(mgr_pool, pjsip_tpmgr);
     mgr->endpt = endpt;
     mgr->on_rx_msg = rx_cb;
     mgr->on_tx_msg = tx_cb;
+    mgr->pool = mgr_pool;
+
+    if (!mgr->pool)
+	return PJ_ENOMEM;
+
     pj_list_init(&mgr->factory_list);
     pj_list_init(&mgr->tdata_list);
-    pj_list_init(&mgr->tp_list);
+    pj_list_init(&mgr->tp_entry_freelist);
 
-    mgr->table = pj_hash_create(pool, PJSIP_TPMGR_HTABLE_SIZE);
+    mgr->table = pj_hash_create(mgr->pool, PJSIP_TPMGR_HTABLE_SIZE);
     if (!mgr->table)
 	return PJ_ENOMEM;
 
-    status = pj_lock_create_recursive_mutex(pool, "tmgr%p", &mgr->lock);
+    status = pj_lock_create_recursive_mutex(mgr->pool, "tmgr%p", &mgr->lock);
     if (status != PJ_SUCCESS)
 	return status;
 
+    for (; i < PJSIP_TRANSPORT_ENTRY_ALLOC_CNT; ++i) {
+	transport *tp_add = NULL;
+
+	tp_add = PJ_POOL_ZALLOC_T(mgr->pool, transport);
+	if (!tp_add)
+	    return PJ_ENOMEM;
+	pj_list_init(tp_add);
+	pj_list_push_back(&mgr->tp_entry_freelist, tp_add);
+    }
+
 #if defined(PJ_DEBUG) && PJ_DEBUG!=0
-    status = pj_atomic_create(pool, 0, &mgr->tdata_counter);
+    status = pj_atomic_create(mgr->pool, 0, &mgr->tdata_counter);
     if (status != PJ_SUCCESS) {
     	pj_lock_destroy(mgr->lock);
     	return status;
@@ -1331,7 +1564,7 @@ static pj_status_t get_net_interface(pjsip_transport_type_e tp_type,
     pj_sockaddr itf_addr;
     pj_status_t status = -1;
 
-    af = (tp_type & PJSIP_TRANSPORT_IPV6)? PJ_AF_INET6 : PJ_AF_INET;
+    af = (tp_type & PJSIP_TRANSPORT_IPV6)? pj_AF_INET6() : pj_AF_INET();
 
     if (pjsip_cfg()->endpt.resolve_hostname_to_get_interface) {
 	status = pj_getipinterface(af, dst, &itf_addr, PJ_TRUE, NULL);
@@ -1343,8 +1576,9 @@ static pj_status_t get_net_interface(pjsip_transport_type_e tp_type,
 	    /* If it fails, e.g: on WM6(http://support.microsoft.com/kb/129065),
 	     * just fallback using pj_gethostip(), see ticket #1660.
 	     */
-	    PJ_LOG(5,(THIS_FILE,"Warning: unable to determine local "
-				"interface, fallback to default interface!"));
+	    PJ_PERROR(5,(THIS_FILE, status,
+			 "Warning: unable to determine local interface, "
+			 "fallback to default interface!"));
 	    status = pj_gethostip(af, &itf_addr);
 	    if (status != PJ_SUCCESS)
 		return status;
@@ -1533,15 +1767,16 @@ PJ_DEF(unsigned) pjsip_tpmgr_get_transport_count(pjsip_tpmgr *mgr)
     pj_hash_iterator_t itr_val;
     pj_hash_iterator_t *itr;
     int nr_of_transports = 0;
-    
+
     pj_lock_acquire(mgr->lock);
-    
+
     itr = pj_hash_first(mgr->table, &itr_val);
     while (itr) {
-	nr_of_transports++;
+	transport *tp_entry = (transport *)pj_hash_this(mgr->table, itr);
+	nr_of_transports += pj_list_size(tp_entry);
 	itr = pj_hash_next(mgr->table, itr);
     }
-    
+
     pj_lock_release(mgr->lock);
 
     return nr_of_transports;
@@ -1558,7 +1793,7 @@ PJ_DEF(pj_status_t) pjsip_tpmgr_destroy( pjsip_tpmgr *mgr )
     pj_hash_iterator_t *itr;
     pjsip_tpfactory *factory;
     pjsip_endpoint *endpt = mgr->endpt;
-    
+
     PJ_LOG(5, (THIS_FILE, "Destroying transport manager"));
 
     pj_lock_acquire(mgr->lock);
@@ -1566,39 +1801,21 @@ PJ_DEF(pj_status_t) pjsip_tpmgr_destroy( pjsip_tpmgr *mgr )
     /*
      * Destroy all transports in the hash table.
      */
-    itr = pj_hash_first(mgr->table, &itr_val);
-    while (itr != NULL) {
-	pj_hash_iterator_t *next;
-	pjsip_transport *transport;
-	
-	transport = (pjsip_transport*) pj_hash_this(mgr->table, itr);
-
-	next = pj_hash_next(mgr->table, itr);
-
-	destroy_transport(mgr, transport);
-
-	itr = next;
+    for (itr = pj_hash_first(mgr->table, &itr_val); itr;
+	 itr = pj_hash_first(mgr->table, &itr_val))
+    {
+	transport *tp_ref;
+	tp_ref = pj_hash_this(mgr->table, itr);
+	destroy_transport(mgr, tp_ref->tp);
     }
 
-    /*
-     * Destroy transports in the list.
-     */
-    if (!pj_list_empty(&mgr->tp_list)) {
-        transport *tp_iter = mgr->tp_list.next;
-        while (tp_iter != &mgr->tp_list) {
-	    transport *next = tp_iter->next;
-	    destroy_transport(mgr, tp_iter->tp);
-	    tp_iter = next;
-        }
-    }
-    
     /*
      * Destroy all factories/listeners.
      */
     factory = mgr->factory_list.next;
     while (factory != &mgr->factory_list) {
 	pjsip_tpfactory *next = factory->next;
-	
+
 	factory->destroy(factory);
 
 	factory = next;
@@ -1642,6 +1859,10 @@ PJ_DEF(pj_status_t) pjsip_tpmgr_destroy( pjsip_tpmgr *mgr )
 	pjsip_endpt_unregister_module(endpt, &mod_msg_print);
     }
 
+    if (mgr->pool) {
+	pjsip_endpt_release_pool( mgr->endpt, mgr->pool );
+    }
+
     return PJ_SUCCESS;
 }
 
@@ -1770,7 +1991,7 @@ PJ_DEF(pj_ssize_t) pjsip_tpmgr_receive_packet( pjsip_tpmgr *mgr,
 	/* Check for parsing syntax error */
 	if (msg==NULL || !pj_list_empty(&rdata->msg_info.parse_err)) {
 	    pjsip_parser_err_report *err;
-	    char buf[128];
+	    char buf[256];
 	    pj_str_t tmp;
 
 	    /* Gather syntax error information */
@@ -1784,7 +2005,10 @@ PJ_DEF(pj_ssize_t) pjsip_tpmgr_receive_packet( pjsip_tpmgr *mgr,
 				       pj_exception_id_name(err->except_code),
 				       (int)err->hname.slen, err->hname.ptr,
 				       err->line, err->col);
-		if (len > 0 && len < (int) (sizeof(buf)-tmp.slen)) {
+		if (len >= (int)sizeof(buf)-tmp.slen) {
+		    len = (int)sizeof(buf)-tmp.slen;
+		}
+		if (len > 0) {
 		    tmp.slen += len;
 		}
 		err = err->next;
@@ -1834,6 +2058,17 @@ PJ_DEF(pj_ssize_t) pjsip_tpmgr_receive_packet( pjsip_tpmgr *mgr,
 	    rdata->msg_info.cseq == NULL) 
 	{
 	    mgr->on_rx_msg(mgr->endpt, PJSIP_EMISSINGHDR, rdata);
+
+	    /* Notify application about the missing header. */
+	    if (mgr->tp_drop_data_cb) {
+		pjsip_tp_dropped_data dd;
+		pj_bzero(&dd, sizeof(dd));
+		dd.tp = tr;
+		dd.data = current_pkt;
+		dd.len = msg_fragment_size;
+		dd.status = PJSIP_EMISSINGHDR;
+		(*mgr->tp_drop_data_cb)(&dd);	    
+	    }
 	    goto finish_process_fragment;
 	}
 
@@ -1856,6 +2091,17 @@ PJ_DEF(pj_ssize_t) pjsip_tpmgr_receive_packet( pjsip_tpmgr *mgr,
 		rdata->msg_info.msg->line.status.code >= 700)
 	    {
 		mgr->on_rx_msg(mgr->endpt, PJSIP_EINVALIDSTATUS, rdata);
+
+		/* Notify application about the invalid status. */
+		if (mgr->tp_drop_data_cb) {
+		    pjsip_tp_dropped_data dd;
+		    pj_bzero(&dd, sizeof(dd));
+		    dd.tp = tr;
+		    dd.data = current_pkt;
+		    dd.len = msg_fragment_size;
+		    dd.status = PJSIP_EINVALIDSTATUS;
+		    (*mgr->tp_drop_data_cb)(&dd);	    
+		}
 		goto finish_process_fragment;
 	    }
 	}
@@ -1968,45 +2214,62 @@ PJ_DEF(pj_status_t) pjsip_tpmgr_acquire_transport2(pjsip_tpmgr *mgr,
 	TRACE_((THIS_FILE, "Transport %s acquired", seltp->obj_name));
 	return PJ_SUCCESS;
 
-
-    } else if (sel && sel->type == PJSIP_TPSELECTOR_LISTENER &&
-	       sel->u.listener)
-    {
-	/* Application has requested that a specific listener is to
-	 * be used. In this case, skip transport hash table lookup.
-	 */
-
-	/* Verify that the listener type matches the destination type */
-	if (sel->u.listener->type != type) {
-	    pj_lock_release(mgr->lock);
-	    return PJSIP_ETPNOTSUITABLE;
-	}
-
-	/* We'll use this listener to create transport */
-	factory = sel->u.listener;
-
     } else {
 
 	/*
 	 * This is the "normal" flow, where application doesn't specify
-	 * specific transport/listener to be used to send message to.
+	 * specific transport to be used to send message to.
 	 * In this case, lookup the transport from the hash table.
 	 */
 	pjsip_transport_key key;
 	int key_len;
-	pjsip_transport *transport;
+	pjsip_transport *tp_ref = NULL;
+	transport *tp_entry = NULL;
 
-	pj_bzero(&key, sizeof(key));
-	key_len = sizeof(key.type) + addr_len;
 
-	/* First try to get exact destination. */
-	key.type = type;
-	pj_memcpy(&key.rem_addr, remote, addr_len);
+	/* If listener is specified, verify that the listener type matches
+	 * the destination type.
+	 */
+	if (sel && sel->type == PJSIP_TPSELECTOR_LISTENER && sel->u.listener)
+	{
+	    if (sel->u.listener->type != type) {
+		pj_lock_release(mgr->lock);
+		return PJSIP_ETPNOTSUITABLE;
+	    }
+	}
 
-	transport = (pjsip_transport*)
-		    pj_hash_get(mgr->table, &key, key_len, NULL);
+	if (!sel || sel->disable_connection_reuse == PJ_FALSE) {
+	    pj_bzero(&key, sizeof(key));
+	    key_len = sizeof(key.type) + addr_len;
 
-	if (transport == NULL) {
+	    /* First try to get exact destination. */
+	    key.type = type;
+	    pj_memcpy(&key.rem_addr, remote, addr_len);
+
+	    tp_entry = (transport *)pj_hash_get(mgr->table, &key, key_len,
+						NULL);
+	    if (tp_entry) {
+		if (sel && sel->type == PJSIP_TPSELECTOR_LISTENER) {
+		    transport *tp_iter = tp_entry;
+		    do {
+			if (sel && sel->type == PJSIP_TPSELECTOR_LISTENER &&
+			    sel->u.listener &&
+			    tp_iter->tp->factory == sel->u.listener)
+			{
+			    tp_ref = tp_iter->tp;
+			    break;
+			}
+			tp_iter = tp_iter->next;
+		    } while (tp_iter != tp_entry);
+		} else {
+		    tp_ref = tp_entry->tp;
+		}
+	    }
+	}
+
+	if (tp_ref == NULL &&
+	    (!sel || sel->disable_connection_reuse == PJ_FALSE))
+	{
 	    unsigned flag = pjsip_transport_get_flag_from_type(type);
 	    const pj_sockaddr *remote_addr = (const pj_sockaddr*)remote;
 
@@ -2019,8 +2282,11 @@ PJ_DEF(pj_status_t) pjsip_tpmgr_acquire_transport2(pjsip_tpmgr *mgr,
 
 		pj_bzero(addr, addr_len);
 		key_len = sizeof(key.type) + addr_len;
-		transport = (pjsip_transport*) 
-			    pj_hash_get(mgr->table, &key, key_len, NULL);
+		tp_entry = (transport *) pj_hash_get(mgr->table, &key,
+						     key_len, NULL);
+		if (tp_entry) {
+		    tp_ref = tp_entry->tp;
+		}
 	    }
 	    /* For datagram transports, try lookup with zero address.
 	     */
@@ -2032,39 +2298,87 @@ PJ_DEF(pj_status_t) pjsip_tpmgr_acquire_transport2(pjsip_tpmgr *mgr,
 		addr->addr.sa_family = remote_addr->addr.sa_family;
 
 		key_len = sizeof(key.type) + addr_len;
-		transport = (pjsip_transport*)
-			    pj_hash_get(mgr->table, &key, key_len, NULL);
+		tp_entry = (transport *) pj_hash_get(mgr->table, &key,
+						     key_len, NULL);
+		if (tp_entry) {
+		    tp_ref = tp_entry->tp;
+		}
 	    }
 	}
 
-	if (transport!=NULL && !transport->is_shutdown) {
+	/* If transport is found and listener is specified, verify listener */
+	else if (sel && sel->type == PJSIP_TPSELECTOR_LISTENER &&
+		 sel->u.listener && tp_ref->factory != sel->u.listener)
+	{
+	    tp_ref = NULL;
+	    /* This will cause a new transport to be created which will be a
+	     * 'duplicate' of the existing transport (same type & remote addr,
+	     * but different factory).
+	     */
+	}
+
+	if (tp_ref!=NULL && !tp_ref->is_shutdown) {
 	    /*
 	     * Transport found!
 	     */
-	    pjsip_transport_add_ref(transport);
+	    pjsip_transport_add_ref(tp_ref);
 	    pj_lock_release(mgr->lock);
-	    *tp = transport;
+	    *tp = tp_ref;
 
-	    TRACE_((THIS_FILE, "Transport %s acquired", transport->obj_name));
+	    TRACE_((THIS_FILE, "Transport %s acquired", tp_ref->obj_name));
 	    return PJ_SUCCESS;
 	}
 
-	/*
-	 * Transport not found!
-	 * Find factory that can create such transport.
-	 */
-	factory = mgr->factory_list.next;
-	while (factory != &mgr->factory_list) {
-	    if (factory->type == type)
-		break;
-	    factory = factory->next;
-	}
 
-	if (factory == &mgr->factory_list) {
-	    /* No factory can create the transport! */
-	    pj_lock_release(mgr->lock);
-	    TRACE_((THIS_FILE, "No suitable factory was found either"));
-	    return PJSIP_EUNSUPTRANSPORT;
+	/*
+	 * Either transport not found, or we don't want to use the existing
+	 * transport (such as in the case of different factory or
+	 * if connection reuse is disabled). So we need to create one,
+	 * find factory that can create such transport.
+	 *
+	 * If there's an existing transport, its place in the hash table
+	 * will be replaced by this new one. And eventually the existing
+	 * transport will still be freed (by application or #1774).
+	 */
+	if (sel && sel->type == PJSIP_TPSELECTOR_LISTENER && sel->u.listener)
+	{
+	    /* Application has requested that a specific listener is to
+	     * be used.
+	     */
+
+	    /* Verify that the listener type matches the destination type */
+	    if (sel->u.listener->type != type) {
+		pj_lock_release(mgr->lock);
+		return PJSIP_ETPNOTSUITABLE;
+	    }
+
+	    /* We'll use this listener to create transport */
+	    factory = sel->u.listener;
+
+	    /* Verify if listener is still valid */
+	    if (!pjsip_tpmgr_is_tpfactory_valid(mgr, factory)) {
+		pj_lock_release(mgr->lock);
+		PJ_LOG(3,(THIS_FILE, "Specified factory for creating "
+				     "transport is not found"));
+		return PJ_ENOTFOUND;
+	    }
+
+	} else {
+
+	    /* Find factory with type matches the destination type */
+	    factory = mgr->factory_list.next;
+	    while (factory != &mgr->factory_list) {
+		if (factory->type == type)
+		    break;
+		factory = factory->next;
+	    }
+
+	    if (factory == &mgr->factory_list) {
+		/* No factory can create the transport! */
+		pj_lock_release(mgr->lock);
+		TRACE_((THIS_FILE, "No suitable factory was found either"));
+		return PJSIP_EUNSUPTRANSPORT;
+	    }
 	}
     }
 
@@ -2073,17 +2387,18 @@ PJ_DEF(pj_status_t) pjsip_tpmgr_acquire_transport2(pjsip_tpmgr *mgr,
     /* Request factory to create transport. */
     if (factory->create_transport2) {
 	status = factory->create_transport2(factory, mgr, mgr->endpt,
-					    (const pj_sockaddr*) remote, 
+					    (const pj_sockaddr*) remote,
 					    addr_len, tdata, tp);
     } else {
 	status = factory->create_transport(factory, mgr, mgr->endpt,
-					   (const pj_sockaddr*) remote, 
+					   (const pj_sockaddr*) remote,
 					   addr_len, tp);
     }
     if (status == PJ_SUCCESS) {
-	PJ_ASSERT_ON_FAIL(tp!=NULL, 
+	PJ_ASSERT_ON_FAIL(tp!=NULL,
 	    {pj_lock_release(mgr->lock); return PJ_EBUG;});
 	pjsip_transport_add_ref(*tp);
+	(*tp)->factory = factory;
     }
     pj_lock_release(mgr->lock);
     return status;
@@ -2123,15 +2438,25 @@ PJ_DEF(void) pjsip_tpmgr_dump_transports(pjsip_tpmgr *mgr)
 	PJ_LOG(3, (THIS_FILE, " Dumping transports:"));
 
 	do {
-	    pjsip_transport *t = (pjsip_transport*) 
-	    			 pj_hash_this(mgr->table, itr);
+	    transport *tp_entry = (transport *) pj_hash_this(mgr->table, itr);
+	    if (tp_entry) {
+		transport *tp_iter = tp_entry;
 
-	    PJ_LOG(3, (THIS_FILE, "  %s %s (refcnt=%d%s)", 
-		       t->obj_name,
-		       t->info,
-		       pj_atomic_get(t->ref_cnt),
-		       (t->idle_timer.id ? " [idle]" : "")));
+		do {
+		    pjsip_transport *tp_ref = tp_iter->tp;
 
+		    PJ_LOG(3, (THIS_FILE, "  %s %s%s%s%s(refcnt=%d%s)",
+			       tp_ref->obj_name,
+			       tp_ref->info,
+			       (tp_ref->factory)?" listener[":"",
+			       (tp_ref->factory)?tp_ref->factory->obj_name:"",
+			       (tp_ref->factory)?"]":"",
+			       pj_atomic_get(tp_ref->ref_cnt),
+			       (tp_ref->idle_timer.id ? " [idle]" : "")));
+
+		    tp_iter = tp_iter->next;
+		} while (tp_iter != tp_entry);
+	    }
 	    itr = pj_hash_next(mgr->table, itr);
 	} while (itr);
     }
@@ -2231,6 +2556,11 @@ PJ_DEF(pj_status_t) pjsip_transport_add_state_listener (
 
     PJ_ASSERT_RETURN(tp && cb && key, PJ_EINVAL);
 
+    if (tp->is_shutdown) {
+	*key = NULL;
+	return PJ_EINVALIDOP;
+    }
+
     pj_lock_acquire(tp->lock);
 
     /* Init transport data, if it hasn't */
