@@ -57,7 +57,8 @@ struct tls_listener
     pj_ssl_sock_t	    *ssock;
     pj_sockaddr		     bound_addr;
     pj_ssl_cert_t	    *cert;
-    pjsip_tls_setting	     tls_setting;
+    pjsip_tls_setting	     tls_setting;    
+    unsigned		     async_cnt;    
 
     /* Group lock to be used by TLS transport and ioqueue key */
     pj_grp_lock_t	    *grp_lock;
@@ -120,10 +121,11 @@ struct tls_transport
  */
 
 /* This callback is called when pending accept() operation completes. */
-static pj_bool_t on_accept_complete(pj_ssl_sock_t *ssock,
+static pj_bool_t on_accept_complete2(pj_ssl_sock_t *ssock,
 				    pj_ssl_sock_t *new_ssock,
 				    const pj_sockaddr_t *src_addr,
-				    int src_addr_len);
+				    int src_addr_len,
+				    pj_status_t status);
 
 /* Callback on incoming data */
 static pj_bool_t on_data_read(pj_ssl_sock_t *ssock,
@@ -160,17 +162,20 @@ static pj_status_t tls_create(struct tls_listener *listener,
 			      const pj_sockaddr *local,
 			      const pj_sockaddr *remote,
 			      const pj_str_t *remote_name,
+			      pj_grp_lock_t *glock,
 			      struct tls_transport **p_tls);
 
 
+/* Clean up TLS resources */
+static void tls_on_destroy(void *arg);
+
+static void wipe_buf(pj_str_t *buf);
+
+
 static void tls_perror(const char *sender, const char *title,
 		       pj_status_t status)
 {
-    char errmsg[PJ_ERR_MSG_SIZE];
-
-    pj_strerror(status, errmsg, sizeof(errmsg));
-
-    PJ_LOG(1,(sender, "%s: %s [code=%d]", title, errmsg, status));
+    PJ_PERROR(3,(sender, status, "%s: [code=%d]", title, status));
 }
 
 
@@ -282,64 +287,264 @@ static void tls_init_shutdown(struct tls_transport *tls, pj_status_t status)
  * The TLS listener/transport factory.
  */
 
+
+static void set_ssock_param(pj_ssl_sock_param *ssock_param,
+    struct tls_listener *listener)
+{
+    int af, sip_ssl_method;
+    pj_uint32_t sip_ssl_proto;
+
+    /* Build SSL socket param */
+    af = pjsip_transport_type_get_af(listener->factory.type);
+    pj_ssl_sock_param_default(ssock_param);
+    ssock_param->sock_af = af;
+    ssock_param->cb.on_accept_complete2 = &on_accept_complete2;
+    ssock_param->async_cnt = listener->async_cnt;
+    ssock_param->ioqueue = pjsip_endpt_get_ioqueue(listener->endpt);
+    ssock_param->timer_heap = pjsip_endpt_get_timer_heap(listener->endpt);
+    ssock_param->require_client_cert = listener->tls_setting.require_client_cert;
+    ssock_param->timeout = listener->tls_setting.timeout;
+    ssock_param->user_data = listener;
+    ssock_param->verify_peer = PJ_FALSE; /* avoid SSL socket closing the socket
+					  * due to verification error */
+    if (ssock_param->send_buffer_size < PJSIP_MAX_PKT_LEN)
+	ssock_param->send_buffer_size = PJSIP_MAX_PKT_LEN;
+    if (ssock_param->read_buffer_size < PJSIP_MAX_PKT_LEN)
+	ssock_param->read_buffer_size = PJSIP_MAX_PKT_LEN;
+    ssock_param->ciphers_num = listener->tls_setting.ciphers_num;
+    ssock_param->ciphers = listener->tls_setting.ciphers;
+    ssock_param->curves_num = listener->tls_setting.curves_num;
+    ssock_param->curves = listener->tls_setting.curves;
+    ssock_param->sigalgs = listener->tls_setting.sigalgs;
+    ssock_param->entropy_type = listener->tls_setting.entropy_type;
+    ssock_param->entropy_path = listener->tls_setting.entropy_path;
+    ssock_param->reuse_addr = listener->tls_setting.reuse_addr;
+    ssock_param->qos_type = listener->tls_setting.qos_type;
+    ssock_param->qos_ignore_error = listener->tls_setting.qos_ignore_error;
+    pj_memcpy(&ssock_param->qos_params, &listener->tls_setting.qos_params,
+	      sizeof(ssock_param->qos_params));
+
+    ssock_param->sockopt_ignore_error =
+				    listener->tls_setting.sockopt_ignore_error;
+    /* Copy the sockopt */
+    pj_memcpy(&ssock_param->sockopt_params,
+	      &listener->tls_setting.sockopt_params,
+	      sizeof(listener->tls_setting.sockopt_params));
+
+    sip_ssl_method = listener->tls_setting.method;
+    sip_ssl_proto = listener->tls_setting.proto;
+    ssock_param->proto = ssl_get_proto(sip_ssl_method, sip_ssl_proto);
+}
+
+static void update_bound_addr(struct tls_listener *listener,
+			      const pj_sockaddr *local)
+{
+    pj_sockaddr *listener_addr = &listener->factory.local_addr;
+    int af = pjsip_transport_type_get_af(listener->factory.type);
+
+    /* Bind address may be different than factory.local_addr because
+     * factory.local_addr will be resolved.
+     */
+    if (local) {
+	pj_sockaddr_cp(&listener->bound_addr, local);
+    } else {
+	pj_sockaddr_init(af, &listener->bound_addr, NULL, 0);
+    }
+    pj_sockaddr_cp(listener_addr, &listener->bound_addr);    
+}
+
+static pj_status_t update_factory_addr(struct tls_listener *listener,
+				       const pjsip_host_port *addr_name)
+{
+    pj_status_t status = PJ_SUCCESS;
+    pj_sockaddr *listener_addr = &listener->factory.local_addr;
+
+    if (addr_name && addr_name->host.slen) {
+	pj_sockaddr tmp;
+	int af = pjsip_transport_type_get_af(listener->factory.type);
+
+	status = pj_sockaddr_init(af, &tmp, &addr_name->host,
+				  (pj_uint16_t)addr_name->port);
+	if (status != PJ_SUCCESS || !pj_sockaddr_has_addr(&tmp) ||
+	    (af == pj_AF_INET() && tmp.ipv4.sin_addr.s_addr == PJ_INADDR_NONE))
+	{
+	    /* Invalid address */
+	    return PJ_EINVAL;
+	}
+
+	/* Copy the address */
+	listener->factory.addr_name = *addr_name;
+	pj_strdup(listener->factory.pool, &listener->factory.addr_name.host,
+		  &addr_name->host);
+	listener->factory.addr_name.port = addr_name->port;
+
+    }
+    else {
+	/* No published address is given, use the bound address */
+
+	/* If the address returns 0.0.0.0, use the default
+	* interface address as the transport's address.
+	*/
+	if (!pj_sockaddr_has_addr(listener_addr)) {
+	    pj_sockaddr hostip;
+
+	    status = pj_gethostip(listener->bound_addr.addr.sa_family,
+				  &hostip);
+	    if (status != PJ_SUCCESS)
+		return status;
+
+	    pj_sockaddr_copy_addr(listener_addr, &hostip);
+	}
+
+	/* Save the address name */
+	sockaddr_to_host_port(listener->factory.pool,
+			      &listener->factory.addr_name, listener_addr);
+    }
+
+    /* If port is zero, get the bound port */
+    if (listener->factory.addr_name.port == 0) {
+	listener->factory.addr_name.port = pj_sockaddr_get_port(listener_addr);
+    }
+
+    pj_ansi_snprintf(listener->factory.obj_name,
+		     sizeof(listener->factory.obj_name),
+		     "tlstp:%d", listener->factory.addr_name.port);
+    return status;
+}
+
+static void update_transport_info(struct tls_listener *listener)
+{
+    enum { INFO_LEN = 100 };
+    char local_addr[PJ_INET6_ADDRSTRLEN + 10];
+    char pub_addr[PJ_INET6_ADDRSTRLEN + 10];
+    pj_sockaddr *listener_addr = &listener->factory.local_addr;
+
+    if (listener->factory.info == NULL) {
+	listener->factory.info = (char*)pj_pool_alloc(listener->factory.pool,
+						      INFO_LEN);
+    }
+    pj_sockaddr_print(listener_addr, local_addr, sizeof(local_addr), 3);
+    pj_addr_str_print(&listener->factory.addr_name.host, 
+		      listener->factory.addr_name.port, pub_addr, 
+		      sizeof(pub_addr), 1);
+    pj_ansi_snprintf(
+	    listener->factory.info, INFO_LEN, "tls %s [published as %s]",
+	    local_addr, pub_addr);
+
+    if (listener->ssock) {
+	char addr[PJ_INET6_ADDRSTRLEN+10];
+
+	PJ_LOG(4, (listener->factory.obj_name,
+	       "SIP TLS listener is ready for incoming connections at %s",
+	       pj_addr_str_print(&listener->factory.addr_name.host,
+			         listener->factory.addr_name.port, addr,
+				 sizeof(addr), 1)));
+    } else {
+	PJ_LOG(4, (listener->factory.obj_name, "SIP TLS is ready "
+	       "(client only)"));
+    }
+}
+
+
 /*
  * This is the public API to create, initialize, register, and start the
  * TLS listener.
  */
-PJ_DEF(pj_status_t) pjsip_tls_transport_start (pjsip_endpoint *endpt,
-					       const pjsip_tls_setting *opt,
-					       const pj_sockaddr_in *local_in,
-					       const pjsip_host_port *a_name,
-					       unsigned async_cnt,
-					       pjsip_tpfactory **p_factory)
+PJ_DEF(pj_status_t) pjsip_tls_transport_start(pjsip_endpoint *endpt,
+    const pjsip_tls_setting *opt,
+    const pj_sockaddr_in *local_in,
+    const pjsip_host_port *a_name,
+    unsigned async_cnt,
+    pjsip_tpfactory **p_factory)
 {
     pj_sockaddr local;
 
     if (local_in)
 	pj_sockaddr_cp(&local, local_in);
 
-    return pjsip_tls_transport_start2(endpt, opt, (local_in? &local : NULL),
-                                      a_name, async_cnt, p_factory);
+    return pjsip_tls_transport_start2(endpt, opt, (local_in ? &local : NULL),
+				      a_name, async_cnt, p_factory);
 }
 
+
+PJ_DEF(pj_status_t) pjsip_tls_transport_lis_start(pjsip_tpfactory *factory,
+						const pj_sockaddr *local,
+						const pjsip_host_port *a_name)
+{
+    pj_status_t status = PJ_SUCCESS;
+    pj_ssl_sock_param ssock_param, newsock_param;
+    struct tls_listener *listener = (struct tls_listener *)factory;
+    pj_sockaddr *listener_addr = &listener->factory.local_addr;
+
+    if (listener->ssock)
+	return PJ_SUCCESS;
+
+    set_ssock_param(&ssock_param, listener);
+    update_bound_addr(listener, local);
+    ssock_param.grp_lock = listener->grp_lock;
+
+    /* Create SSL socket */
+    status = pj_ssl_sock_create(listener->factory.pool, &ssock_param, 
+				&listener->ssock);
+    if (status != PJ_SUCCESS)
+	return status;
+
+    if (listener->cert) {
+	status = pj_ssl_sock_set_certificate(listener->ssock, 
+				       listener->factory.pool, listener->cert);
+	if (status != PJ_SUCCESS)
+	    return status;
+    }
+
+    /* Start accepting incoming connections. Note that some TLS/SSL
+     * backends may not support for SSL socket server.
+     */    
+    pj_memcpy(&newsock_param, &ssock_param, sizeof(newsock_param));
+    newsock_param.async_cnt = 1;
+    newsock_param.cb.on_data_read = &on_data_read;
+    newsock_param.cb.on_data_sent = &on_data_sent;
+    status = pj_ssl_sock_start_accept2(listener->ssock, listener->factory.pool,
+			    (pj_sockaddr_t*)listener_addr,
+			    pj_sockaddr_get_len((pj_sockaddr_t*)listener_addr),
+			    &newsock_param);
+
+    if (status == PJ_SUCCESS || status == PJ_EPENDING) {
+	pj_ssl_sock_info info;	
+
+	/* Retrieve the bound address */
+	status = pj_ssl_sock_get_info(listener->ssock, &info);
+	if (status == PJ_SUCCESS)
+	    pj_sockaddr_cp(listener_addr, (pj_sockaddr_t*)&info.local_addr);
+
+    }
+    status = update_factory_addr(listener, a_name);
+    if (status != PJ_SUCCESS)
+	return status;
+
+    update_transport_info(listener);
+
+    return status;    
+}
+
+
 PJ_DEF(pj_status_t) pjsip_tls_transport_start2( pjsip_endpoint *endpt,
  					        const pjsip_tls_setting *opt,
 					        const pj_sockaddr *local,
 					        const pjsip_host_port *a_name,
 					        unsigned async_cnt,
 					        pjsip_tpfactory **p_factory)
-{
+{        
     pj_pool_t *pool;
-    pj_bool_t is_ipv6;
-    int af, sip_ssl_method;
-    pj_uint32_t sip_ssl_proto;
-    struct tls_listener *listener;
-    pj_ssl_sock_param ssock_param;
-    pj_sockaddr *listener_addr;
-    pj_bool_t has_listener;
+    pj_bool_t is_ipv6;    
+    struct tls_listener *listener;    
     pj_status_t status;
 
     /* Sanity check */
     PJ_ASSERT_RETURN(endpt && async_cnt, PJ_EINVAL);
 
-    is_ipv6 = (local && local->addr.sa_family == pj_AF_INET6());
-    af = is_ipv6 ? pj_AF_INET6() : pj_AF_INET();
+    is_ipv6 = (local && local->addr.sa_family == pj_AF_INET6());    
 
-    /* Verify that address given in a_name (if any) is valid */
-    if (a_name && a_name->host.slen) {
-	pj_sockaddr tmp;
-
-	status = pj_sockaddr_init(af, &tmp, &a_name->host,
-				  (pj_uint16_t)a_name->port);
-	if (status != PJ_SUCCESS || !pj_sockaddr_has_addr(&tmp) ||
-	    (!is_ipv6 && tmp.ipv4.sin_addr.s_addr == PJ_INADDR_NONE))
-	{
-	    /* Invalid address */
-	    return PJ_EINVAL;
-	}
-    }
-
-    pool = pjsip_endpt_create_pool(endpt, "tlslis", POOL_LIS_INIT, 
+    pool = pjsip_endpt_create_pool(endpt, "tlstp", POOL_LIS_INIT, 
 				   POOL_LIS_INC);
     PJ_ASSERT_RETURN(pool, PJ_ENOMEM);
 
@@ -352,9 +557,10 @@ PJ_DEF(pj_status_t) pjsip_tls_transport_start2( pjsip_endpoint *endpt,
     listener->factory.type_name = (char*)
 		pjsip_transport_get_type_name(listener->factory.type);
     listener->factory.flag = 
-	pjsip_transport_get_flag_from_type(listener->factory.type);
+		pjsip_transport_get_flag_from_type(listener->factory.type);
+    listener->endpt = endpt;
 
-    pj_ansi_strcpy(listener->factory.obj_name, "tlslis");
+    pj_ansi_strcpy(listener->factory.obj_name, "tlstp");
     if (is_ipv6)
 	pj_ansi_strcat(listener->factory.obj_name, "6");
 
@@ -371,78 +577,23 @@ PJ_DEF(pj_status_t) pjsip_tls_transport_start2( pjsip_endpoint *endpt,
     if (async_cnt > MAX_ASYNC_CNT) 
 	async_cnt = MAX_ASYNC_CNT;
 
-    /* Build SSL socket param */
-    pj_ssl_sock_param_default(&ssock_param);
-    ssock_param.sock_af = af;
-    ssock_param.cb.on_accept_complete = &on_accept_complete;
-    ssock_param.cb.on_data_read = &on_data_read;
-    ssock_param.cb.on_data_sent = &on_data_sent;
-    ssock_param.async_cnt = async_cnt;
-    ssock_param.ioqueue = pjsip_endpt_get_ioqueue(endpt);
-    ssock_param.require_client_cert = listener->tls_setting.require_client_cert;
-    ssock_param.timeout = listener->tls_setting.timeout;
-    ssock_param.user_data = listener;
-    ssock_param.verify_peer = PJ_FALSE; /* avoid SSL socket closing the socket
-					 * due to verification error */
-    if (ssock_param.send_buffer_size < PJSIP_MAX_PKT_LEN)
-	ssock_param.send_buffer_size = PJSIP_MAX_PKT_LEN;
-    if (ssock_param.read_buffer_size < PJSIP_MAX_PKT_LEN)
-	ssock_param.read_buffer_size = PJSIP_MAX_PKT_LEN;
-    ssock_param.ciphers_num = listener->tls_setting.ciphers_num;
-    ssock_param.ciphers = listener->tls_setting.ciphers;
-    ssock_param.reuse_addr = listener->tls_setting.reuse_addr;
-    ssock_param.qos_type = listener->tls_setting.qos_type;
-    ssock_param.qos_ignore_error = listener->tls_setting.qos_ignore_error;
-    pj_memcpy(&ssock_param.qos_params, &listener->tls_setting.qos_params,
-	      sizeof(ssock_param.qos_params));
-
-    ssock_param.sockopt_ignore_error = 
-				     listener->tls_setting.sockopt_ignore_error;
-    /* Copy the sockopt */
-    pj_memcpy(&ssock_param.sockopt_params, 
-	      &listener->tls_setting.sockopt_params,
-	      sizeof(listener->tls_setting.sockopt_params));
-
-    has_listener = PJ_FALSE;
-
-    sip_ssl_method = listener->tls_setting.method;
-    sip_ssl_proto = listener->tls_setting.proto;
-    ssock_param.proto = ssl_get_proto(sip_ssl_method, sip_ssl_proto); 
+    listener->async_cnt = async_cnt;    
 
     /* Create group lock */
     status = pj_grp_lock_create(pool, NULL, &listener->grp_lock);
     if (status != PJ_SUCCESS)
-	return status;
+	goto on_error;
 
     /* Setup group lock handler */
     pj_grp_lock_add_ref(listener->grp_lock);
     pj_grp_lock_add_handler(listener->grp_lock, pool, listener,
 			    &lis_on_destroy);
 
-    ssock_param.grp_lock = listener->grp_lock;
-
-    /* Create SSL socket */
-    status = pj_ssl_sock_create(pool, &ssock_param, &listener->ssock);
-    if (status != PJ_SUCCESS)
-	goto on_error;
-
-    /* Bind address may be different than factory.local_addr because
-     * factory.local_addr will be resolved below.
-     */
-    listener_addr = &listener->factory.local_addr;
-    if (local) {
-	pj_sockaddr_cp((pj_sockaddr_t*)listener_addr, 
-		       (const pj_sockaddr_t*)local);
-	pj_sockaddr_cp(&listener->bound_addr, local);
-    } else {
-	pj_sockaddr_init(af, listener_addr, NULL, 0);
-	pj_sockaddr_init(af, &listener->bound_addr, NULL, 0);
-    }
-
     /* Check if certificate/CA list for SSL socket is set */
     if (listener->tls_setting.cert_file.slen ||
 	listener->tls_setting.ca_list_file.slen ||
-	listener->tls_setting.ca_list_path.slen) 
+	listener->tls_setting.ca_list_path.slen || 
+	listener->tls_setting.privkey_file.slen) 
     {
 	status = pj_ssl_cert_load_from_files2(pool,
 			&listener->tls_setting.ca_list_file,
@@ -453,78 +604,45 @@ PJ_DEF(pj_status_t) pjsip_tls_transport_start2( pjsip_endpoint *endpt,
 			&listener->cert);
 	if (status != PJ_SUCCESS)
 	    goto on_error;
-
-	status = pj_ssl_sock_set_certificate(listener->ssock, pool, 
-					     listener->cert);
+    } else if (listener->tls_setting.ca_buf.slen ||
+	       listener->tls_setting.cert_buf.slen||
+	       listener->tls_setting.privkey_buf.slen)
+    {
+	status = pj_ssl_cert_load_from_buffer(pool,
+			&listener->tls_setting.ca_buf,
+			&listener->tls_setting.cert_buf,
+			&listener->tls_setting.privkey_buf,
+			&listener->tls_setting.password,
+			&listener->cert);
 	if (status != PJ_SUCCESS)
-	    goto on_error;
+	    goto on_error;    
     }
 
-    /* Start accepting incoming connections. Note that some TLS/SSL backends
-     * may not support for SSL socket server.
-     */
-    has_listener = PJ_FALSE;
-
-    status = pj_ssl_sock_start_accept(listener->ssock, pool, 
-			  (pj_sockaddr_t*)listener_addr, 
-			  pj_sockaddr_get_len((pj_sockaddr_t*)listener_addr));
-    if (status == PJ_SUCCESS || status == PJ_EPENDING) {
-	pj_ssl_sock_info info;
-	has_listener = PJ_TRUE;
-
-	/* Retrieve the bound address */
-	status = pj_ssl_sock_get_info(listener->ssock, &info);
-	if (status == PJ_SUCCESS)
-	    pj_sockaddr_cp(listener_addr, (pj_sockaddr_t*)&info.local_addr);
-    } else if (status != PJ_ENOTSUP) {
-	goto on_error;
-    }
-
-    /* If published host/IP is specified, then use that address as the
-     * listener advertised address.
-     */
-    if (a_name && a_name->host.slen) {
-	/* Copy the address */
-	listener->factory.addr_name = *a_name;
-	pj_strdup(listener->factory.pool, &listener->factory.addr_name.host, 
-		  &a_name->host);
-	listener->factory.addr_name.port = a_name->port;
-
-    } else {
-	/* No published address is given, use the bound address */
-
-	/* If the address returns 0.0.0.0, use the default
-	 * interface address as the transport's address.
-	 */
-	if (!pj_sockaddr_has_addr(listener_addr)) {
-	    pj_sockaddr hostip;
-
-	    status = pj_gethostip(af, &hostip);
-	    if (status != PJ_SUCCESS)
-		goto on_error;
-
-	    pj_sockaddr_copy_addr(listener_addr, &hostip);
-	}
-
-	/* Save the address name */
-	sockaddr_to_host_port(listener->factory.pool, 
-			      &listener->factory.addr_name, listener_addr);
-    }
-
-    /* If port is zero, get the bound port */
-    if (listener->factory.addr_name.port == 0) {
-	listener->factory.addr_name.port = pj_sockaddr_get_port(listener_addr);
-    }
-
-    pj_ansi_snprintf(listener->factory.obj_name, 
-		     sizeof(listener->factory.obj_name),
-		     "tlslis:%d",  listener->factory.addr_name.port);
-
     /* Register to transport manager */
     listener->endpt = endpt;
     listener->tpmgr = pjsip_endpt_get_tpmgr(endpt);
     listener->factory.create_transport2 = lis_create_transport;
     listener->factory.destroy = lis_destroy;
+
+#if !(defined(PJSIP_TLS_TRANSPORT_DONT_CREATE_LISTENER) && \
+    PJSIP_TLS_TRANSPORT_DONT_CREATE_LISTENER != 0)
+    /* Start listener. */
+    status = pjsip_tls_transport_lis_start(&listener->factory, local, a_name);
+    if (status != PJ_SUCCESS)
+	goto on_error;
+#else
+    update_bound_addr(listener, local);
+    /* If published host/IP is specified, then use that address as the
+     * listener advertised address.
+     */
+    status = update_factory_addr(listener, a_name);
+    if (status != PJ_SUCCESS)
+	goto on_error;
+
+    /* Set transport info. */
+    update_transport_info(listener);
+#endif
+
     listener->is_registered = PJ_TRUE;
     status = pjsip_tpmgr_register_tpfactory(listener->tpmgr,
 					    &listener->factory);
@@ -533,18 +651,6 @@ PJ_DEF(pj_status_t) pjsip_tls_transport_start2( pjsip_endpoint *endpt,
 	goto on_error;
     }
 
-    if (has_listener) {
-	PJ_LOG(4,(listener->factory.obj_name, 
-		 "SIP TLS listener is ready for incoming connections "
-		 "at %.*s:%d",
-		 (int)listener->factory.addr_name.host.slen,
-		 listener->factory.addr_name.host.ptr,
-		 listener->factory.addr_name.port));
-    } else {
-	PJ_LOG(4,(listener->factory.obj_name, "SIP TLS is ready "
-		  "(client only)"));
-    }
-
     /* Return the pointer to user */
     if (p_factory) *p_factory = &listener->factory;
 
@@ -561,27 +667,25 @@ static void lis_on_destroy(void *arg)
 {
     struct tls_listener *listener = (struct tls_listener*)arg;
 
+    if (listener->cert) {
+	pj_ssl_cert_wipe_keys(listener->cert);
+	listener->cert = NULL;
+    }
+
     if (listener->factory.lock) {
 	pj_lock_destroy(listener->factory.lock);
 	listener->factory.lock = NULL;
     }
 
     if (listener->factory.pool) {
-	pj_pool_t *pool = listener->factory.pool;
-
-	PJ_LOG(4,(listener->factory.obj_name,  "SIP TLS listener destroyed"));
-
-	listener->factory.pool = NULL;
-	pj_pool_release(pool);
+	PJ_LOG(4,(listener->factory.obj_name,  "SIP TLS transport destroyed"));
+	pj_pool_secure_release(&listener->factory.pool);
     }
 }
 
 
-/* This callback is called by transport manager to destroy listener */
-static pj_status_t lis_destroy(pjsip_tpfactory *factory)
+static void lis_close(struct tls_listener *listener)
 {
-    struct tls_listener *listener = (struct tls_listener *)factory;
-
     if (listener->is_registered) {
 	pjsip_tpmgr_unregister_tpfactory(listener->tpmgr, &listener->factory);
 	listener->is_registered = PJ_FALSE;
@@ -591,6 +695,15 @@ static pj_status_t lis_destroy(pjsip_tpfactory *factory)
 	pj_ssl_sock_close(listener->ssock);
 	listener->ssock = NULL;
     }
+}
+
+
+/* This callback is called by transport manager to destroy listener */
+static pj_status_t lis_destroy(pjsip_tpfactory *factory)
+{
+    struct tls_listener *listener = (struct tls_listener *)factory;
+
+    lis_close(listener);
 
     if (listener->grp_lock) {
 	pj_grp_lock_t *grp_lock = listener->grp_lock;
@@ -605,6 +718,38 @@ static pj_status_t lis_destroy(pjsip_tpfactory *factory)
 }
 
 
+PJ_DEF(pj_status_t) pjsip_tls_transport_restart(pjsip_tpfactory *factory,
+						const pj_sockaddr *local,
+						const pjsip_host_port *a_name)
+{
+    pj_status_t status = PJ_SUCCESS;
+    struct tls_listener *listener = (struct tls_listener *)factory;
+
+    lis_close(listener);
+
+    status = pjsip_tls_transport_lis_start(factory, local, a_name);
+    if (status != PJ_SUCCESS) {	
+	tls_perror(listener->factory.obj_name, 
+		   "Unable to start listener after closing it", status);
+
+	return status;
+    }
+    
+    status = pjsip_tpmgr_register_tpfactory(listener->tpmgr,
+					    &listener->factory);
+    if (status != PJ_SUCCESS) {
+	tls_perror(listener->factory.obj_name,
+		    "Unable to register the transport listener", status);
+
+	listener->is_registered = PJ_FALSE;	
+    } else {
+	listener->is_registered = PJ_TRUE;	
+    }    
+
+    return status;
+}
+
+
 /***************************************************************************/
 /*
  * TLS Transport
@@ -649,6 +794,7 @@ static pj_status_t tls_create( struct tls_listener *listener,
 			       const pj_sockaddr *local,
 			       const pj_sockaddr *remote,
 			       const pj_str_t *remote_name,
+			       pj_grp_lock_t *glock,
 			       struct tls_transport **p_tls)
 {
     struct tls_transport *tls;
@@ -729,9 +875,15 @@ static pj_status_t tls_create( struct tls_listener *listener,
     tls->base.send_msg = &tls_send_msg;
     tls->base.do_shutdown = &tls_shutdown;
     tls->base.destroy = &tls_destroy_transport;
+    tls->base.factory = &listener->factory;
 
     tls->ssock = ssock;
 
+    /* Set up the group lock */
+    tls->grp_lock = tls->base.grp_lock = glock;
+    pj_grp_lock_add_ref(tls->grp_lock);
+    pj_grp_lock_add_handler(tls->grp_lock, pool, tls, &tls_on_destroy);
+
     /* Register transport to transport manager */
     status = pjsip_transport_register(listener->tpmgr, &tls->base);
     if (status != PJ_SUCCESS) {
@@ -755,7 +907,11 @@ static pj_status_t tls_create( struct tls_listener *listener,
     return PJ_SUCCESS;
 
 on_error:
-    tls_destroy(&tls->base, status);
+    if (tls->grp_lock && pj_grp_lock_get_ref(tls->grp_lock))
+	tls_destroy(&tls->base, status);
+    else
+    	tls_on_destroy(tls);
+
     return status;
 }
 
@@ -821,8 +977,7 @@ static void tls_on_destroy(void *arg)
     struct tls_transport *tls = (struct tls_transport*)arg;
 
     if (tls->rdata.tp_info.pool) {
-	pj_pool_release(tls->rdata.tp_info.pool);
-	tls->rdata.tp_info.pool = NULL;
+	pj_pool_secure_release(&tls->rdata.tp_info.pool);
     }
 
     if (tls->base.lock) {
@@ -836,8 +991,6 @@ static void tls_on_destroy(void *arg)
     }
 
     if (tls->base.pool) {
-	pj_pool_t *pool;
-
 	if (tls->close_reason != PJ_SUCCESS) {
 	    char errmsg[PJ_ERR_MSG_SIZE];
 
@@ -852,10 +1005,7 @@ static void tls_on_destroy(void *arg)
 		      "TLS transport destroyed normally"));
 
 	}
-
-	pool = tls->base.pool;
-	tls->base.pool = NULL;
-	pj_pool_release(pool);
+	pj_pool_secure_release(&tls->base.pool);
     }
 }
 
@@ -910,8 +1060,6 @@ static pj_status_t tls_destroy(pjsip_transport *transport,
 	tls->grp_lock = NULL;
 	pj_grp_lock_dec_ref(grp_lock);
 	/* Transport may have been deleted at this point */
-    } else {
-	tls_on_destroy(tls);
     }
 
     return PJ_SUCCESS;
@@ -961,9 +1109,8 @@ static pj_status_t tls_start_read(struct tls_transport *tls)
     status = pj_ssl_sock_start_read2(tls->ssock, tls->base.pool, size,
 				     readbuf, 0);
     if (status != PJ_SUCCESS && status != PJ_EPENDING) {
-	PJ_LOG(4, (tls->base.obj_name, 
-		   "pj_ssl_sock_start_read() error, status=%d", 
-		   status));
+	PJ_PERROR(4, (tls->base.obj_name, status,
+		     "pj_ssl_sock_start_read() error"));
 	return status;
     }
 
@@ -1026,6 +1173,7 @@ static pj_status_t lis_create_transport(pjsip_tpfactory *factory,
     ssock_param.cb.on_data_sent = &on_data_sent;
     ssock_param.async_cnt = 1;
     ssock_param.ioqueue = pjsip_endpt_get_ioqueue(listener->endpt);
+    ssock_param.timer_heap = pjsip_endpt_get_timer_heap(listener->endpt);
     ssock_param.server_name = remote_name;
     ssock_param.timeout = listener->tls_setting.timeout;
     ssock_param.user_data = NULL; /* pending, must be set later */
@@ -1037,6 +1185,11 @@ static pj_status_t lis_create_transport(pjsip_tpfactory *factory,
 	ssock_param.read_buffer_size = PJSIP_MAX_PKT_LEN;
     ssock_param.ciphers_num = listener->tls_setting.ciphers_num;
     ssock_param.ciphers = listener->tls_setting.ciphers;
+    ssock_param.curves_num = listener->tls_setting.curves_num;
+    ssock_param.curves = listener->tls_setting.curves;
+    ssock_param.sigalgs = listener->tls_setting.sigalgs;
+    ssock_param.entropy_type = listener->tls_setting.entropy_type;
+    ssock_param.entropy_path = listener->tls_setting.entropy_path;
     ssock_param.qos_type = listener->tls_setting.qos_type;
     ssock_param.qos_ignore_error = listener->tls_setting.qos_ignore_error;
     pj_memcpy(&ssock_param.qos_params, &listener->tls_setting.qos_params,
@@ -1081,20 +1234,13 @@ static pj_status_t lis_create_transport(pjsip_tpfactory *factory,
 
     /* Create the transport descriptor */
     status = tls_create(listener, pool, ssock, PJ_FALSE, &local_addr, 
-			rem_addr, &remote_name, &tls);
-    if (status != PJ_SUCCESS) {
-	pj_grp_lock_destroy(glock);
+			rem_addr, &remote_name, glock, &tls);
+    if (status != PJ_SUCCESS)
 	return status;
-    }
 
     /* Set the "pending" SSL socket user data */
     pj_ssl_sock_set_user_data(tls->ssock, tls);
 
-    /* Set up the group lock */
-    tls->grp_lock = glock;
-    pj_grp_lock_add_ref(tls->grp_lock);
-    pj_grp_lock_add_handler(tls->grp_lock, pool, tls, &tls_on_destroy);
-
     /* Start asynchronous connect() operation */
     tls->has_pending_connect = PJ_TRUE;
     status = pj_ssl_sock_start_connect(tls->ssock, tls->base.pool, 
@@ -1110,6 +1256,8 @@ static pj_status_t lis_create_transport(pjsip_tpfactory *factory,
 
     if (tls->has_pending_connect) {
 	pj_ssl_sock_info info;
+	char local_addr_buf[PJ_INET6_ADDRSTRLEN+10];
+	char remote_addr_buf[PJ_INET6_ADDRSTRLEN+10];
 
 	/* Update local address, just in case local address currently set is 
 	 * different now that asynchronous connect() is started.
@@ -1139,13 +1287,13 @@ static pj_status_t lis_create_transport(pjsip_tpfactory *factory,
 	}
 
 	PJ_LOG(4,(tls->base.obj_name, 
-		  "TLS transport %.*s:%d is connecting to %.*s:%d...",
-		  (int)tls->base.local_name.host.slen,
-		  tls->base.local_name.host.ptr,
-		  tls->base.local_name.port,
-		  (int)tls->base.remote_name.host.slen,
-		  tls->base.remote_name.host.ptr,
-		  tls->base.remote_name.port));
+		  "TLS transport %s is connecting to %s...",
+		  pj_addr_str_print(&tls->base.local_name.host, 
+				    tls->base.local_name.port, 
+				    local_addr_buf, sizeof(local_addr_buf), 1),
+		  pj_addr_str_print(&tls->base.remote_name.host, 
+			        tls->base.remote_name.port, 
+				remote_addr_buf, sizeof(remote_addr_buf), 1)));
     }
 
     /* Done */
@@ -1155,15 +1303,12 @@ static pj_status_t lis_create_transport(pjsip_tpfactory *factory,
 }
 
 
-/*
- * This callback is called by SSL socket when pending accept() operation
- * has completed.
- */
-static pj_bool_t on_accept_complete(pj_ssl_sock_t *ssock,
-				    pj_ssl_sock_t *new_ssock,
-				    const pj_sockaddr_t *src_addr,
-				    int src_addr_len)
-{
+static pj_bool_t on_accept_complete2(pj_ssl_sock_t *ssock,
+				     pj_ssl_sock_t *new_ssock,
+				     const pj_sockaddr_t *src_addr,
+				     int src_addr_len, 
+				     pj_status_t accept_status)
+{    
     struct tls_listener *listener;
     struct tls_transport *tls;
     pj_ssl_sock_info ssl_info;
@@ -1172,22 +1317,53 @@ static pj_bool_t on_accept_complete(pj_ssl_sock_t *ssock,
     pj_sockaddr tmp_src_addr;
     pj_bool_t is_shutdown;
     pj_status_t status;
+    char addr_buf[PJ_INET6_ADDRSTRLEN+10];        
 
     PJ_UNUSED_ARG(src_addr_len);
 
     listener = (struct tls_listener*) pj_ssl_sock_get_user_data(ssock);
 
+    if (accept_status != PJ_SUCCESS) {
+	if (listener && listener->tls_setting.on_accept_fail_cb) {
+	    pjsip_tls_on_accept_fail_param param;
+	    pj_ssl_sock_info ssi;
+
+	    pj_bzero(&param, sizeof(param));
+	    param.status = accept_status;
+	    param.local_addr = &listener->factory.local_addr;
+	    param.remote_addr = src_addr;
+	    if (new_ssock &&
+		pj_ssl_sock_get_info(new_ssock, &ssi) == PJ_SUCCESS)
+	    {
+		param.last_native_err = ssi.last_native_err;
+	    }
+
+	    (*listener->tls_setting.on_accept_fail_cb) (&param);
+	}
+
+	return PJ_FALSE;
+    }
+
     PJ_ASSERT_RETURN(new_ssock, PJ_TRUE);
 
-    if (!listener->is_registered)
+    if (!listener->is_registered) {
+	if (listener->tls_setting.on_accept_fail_cb) {
+	    pjsip_tls_on_accept_fail_param param;
+	    pj_bzero(&param, sizeof(param));
+	    param.status = PJSIP_TLS_EACCEPT;
+	    param.local_addr = &listener->factory.local_addr;
+	    param.remote_addr = src_addr;
+	    (*listener->tls_setting.on_accept_fail_cb) (&param);
+	}
 	return PJ_FALSE;
+    }	
 
     PJ_LOG(4,(listener->factory.obj_name, 
-	      "TLS listener %.*s:%d: got incoming TLS connection "
+	      "TLS listener %s: got incoming TLS connection "
 	      "from %s, sock=%d",
-	      (int)listener->factory.addr_name.host.slen,
-	      listener->factory.addr_name.host.ptr,
-	      listener->factory.addr_name.port,
+	      pj_addr_str_print(&listener->factory.addr_name.host, 
+				listener->factory.addr_name.port, addr_buf, 
+				sizeof(addr_buf), 1),
 	      pj_sockaddr_print(src_addr, addr, sizeof(addr), 3),
 	      new_ssock));
 
@@ -1197,6 +1373,15 @@ static pj_bool_t on_accept_complete(pj_ssl_sock_t *ssock,
     status = pj_ssl_sock_get_info(new_ssock, &ssl_info);
     if (status != PJ_SUCCESS) {
 	pj_ssl_sock_close(new_ssock);
+
+	if (listener->tls_setting.on_accept_fail_cb) {
+	    pjsip_tls_on_accept_fail_param param;
+	    pj_bzero(&param, sizeof(param));
+	    param.status = status;
+	    param.local_addr = &listener->factory.local_addr;
+	    param.remote_addr = src_addr;
+	    (*listener->tls_setting.on_accept_fail_cb) (&param);
+	}
 	return PJ_TRUE;
     }
 
@@ -1209,23 +1394,24 @@ static pj_bool_t on_accept_complete(pj_ssl_sock_t *ssock,
      * Create TLS transport for the new socket.
      */
     status = tls_create( listener, NULL, new_ssock, PJ_TRUE,
-			 &listener->factory.local_addr,
-			 &tmp_src_addr, NULL, &tls);
+			 &ssl_info.local_addr, &tmp_src_addr, NULL,
+			 ssl_info.grp_lock, &tls);
     
-    if (status != PJ_SUCCESS)
+    if (status != PJ_SUCCESS) {
+	if (listener->tls_setting.on_accept_fail_cb) {
+	    pjsip_tls_on_accept_fail_param param;
+	    pj_bzero(&param, sizeof(param));
+	    param.status = status;
+	    param.local_addr = &listener->factory.local_addr;
+	    param.remote_addr = src_addr;
+	    (*listener->tls_setting.on_accept_fail_cb) (&param);
+	}
 	return PJ_TRUE;
+    }
 
     /* Set the "pending" SSL socket user data */
     pj_ssl_sock_set_user_data(new_ssock, tls);
 
-    /* Set up the group lock */
-    if (ssl_info.grp_lock) {
-	tls->grp_lock = ssl_info.grp_lock;
-	pj_grp_lock_add_ref(tls->grp_lock);
-	pj_grp_lock_add_handler(tls->grp_lock, tls->base.pool, tls,
-				&tls_on_destroy);
-    }
-
     /* Prevent immediate transport destroy as application may access it 
      * (getting info, etc) in transport state notification callback.
      */
@@ -1281,8 +1467,9 @@ static pj_bool_t on_accept_complete(pj_ssl_sock_t *ssock,
 	tls_destroy(&tls->base, status);
     } else {
 	/* Start keep-alive timer */
-	if (PJSIP_TLS_KEEP_ALIVE_INTERVAL) {
-	    pj_time_val delay = {PJSIP_TLS_KEEP_ALIVE_INTERVAL, 0};
+	if (pjsip_cfg()->tls.keep_alive_interval) {
+	    pj_time_val delay = {0};	    
+	    delay.sec = pjsip_cfg()->tls.keep_alive_interval;
 	    pjsip_endpt_schedule_timer(listener->endpt, 
 				       &tls->ka_timer, 
 				       &delay);
@@ -1294,6 +1481,18 @@ static pj_bool_t on_accept_complete(pj_ssl_sock_t *ssock,
     return PJ_TRUE;
 }
 
+/*
+ * This callback is called by SSL socket when pending accept() operation
+ * has completed.
+ */
+//static pj_bool_t on_accept_complete(pj_ssl_sock_t *ssock,
+//				    pj_ssl_sock_t *new_ssock,
+//				    const pj_sockaddr_t *src_addr,
+//				    int src_addr_len)
+//{
+//    PJ_UNUSED_ARG(src_addr_len);
+//}
+
 
 /* 
  * Callback from ioqueue when packet is sent.
@@ -1554,11 +1753,21 @@ static pj_bool_t on_connect_complete(pj_ssl_sock_t *ssock,
     pj_sockaddr addr, *tp_addr;
     pjsip_tp_state_callback state_cb;
     pj_bool_t is_shutdown;
+    char local_addr_buf[PJ_INET6_ADDRSTRLEN+10];
+    char remote_addr_buf[PJ_INET6_ADDRSTRLEN+10];
 
     tls = (struct tls_transport*) pj_ssl_sock_get_user_data(ssock);
 
-    if (tls->base.is_shutdown || tls->base.is_destroying) 
-	return PJ_FALSE;
+    /* If transport is being shutdown/destroyed, proceed as error connect.
+     * Note that it is important to notify application via on_data_sent()
+     * as otherwise the transport reference counter may never reach zero
+     * (see #1898).
+     */
+    if ((tls->base.is_shutdown || tls->base.is_destroying) &&
+	status == PJ_SUCCESS)
+    {
+	status = PJ_ECANCELLED;
+    }
 
     /* Check connect() status */
     if (status != PJ_SUCCESS) {
@@ -1730,13 +1939,13 @@ static pj_bool_t on_connect_complete(pj_ssl_sock_t *ssock,
     tls->has_pending_connect = PJ_FALSE;
 
     PJ_LOG(4,(tls->base.obj_name, 
-	      "TLS transport %.*s:%d is connected to %.*s:%d",
-	      (int)tls->base.local_name.host.slen,
-	      tls->base.local_name.host.ptr,
-	      tls->base.local_name.port,
-	      (int)tls->base.remote_name.host.slen,
-	      tls->base.remote_name.host.ptr,
-	      tls->base.remote_name.port));
+	      "TLS transport %s is connected to %s",
+	      pj_addr_str_print(&tls->base.local_name.host, 
+				tls->base.local_name.port, local_addr_buf, 
+				sizeof(local_addr_buf), 1),
+	      pj_addr_str_print(&tls->base.remote_name.host, 
+				tls->base.remote_name.port, remote_addr_buf, 
+				sizeof(remote_addr_buf), 1)));
 
     /* Start pending read */
     status = tls_start_read(tls);
@@ -1747,8 +1956,9 @@ static pj_bool_t on_connect_complete(pj_ssl_sock_t *ssock,
     tls_flush_pending_tx(tls);
 
     /* Start keep-alive timer */
-    if (PJSIP_TLS_KEEP_ALIVE_INTERVAL) {
-	pj_time_val delay = { PJSIP_TLS_KEEP_ALIVE_INTERVAL, 0 };
+    if (pjsip_cfg()->tls.keep_alive_interval) {
+	pj_time_val delay = {0};	    
+	delay.sec = pjsip_cfg()->tls.keep_alive_interval;
 	pjsip_endpt_schedule_timer(tls->base.endpt, &tls->ka_timer, 
 				   &delay);
 	tls->ka_timer.id = PJ_TRUE;
@@ -1772,6 +1982,7 @@ static void tls_keep_alive_timer(pj_timer_heap_t *th, pj_timer_entry *e)
     pj_time_val now;
     pj_ssize_t size;
     pj_status_t status;
+    char addr[PJ_INET6_ADDRSTRLEN+10];    
 
     PJ_UNUSED_ARG(th);
 
@@ -1780,9 +1991,9 @@ static void tls_keep_alive_timer(pj_timer_heap_t *th, pj_timer_entry *e)
     pj_gettimeofday(&now);
     PJ_TIME_VAL_SUB(now, tls->last_activity);
 
-    if (now.sec > 0 && now.sec < PJSIP_TLS_KEEP_ALIVE_INTERVAL) {
+    if (now.sec > 0 && now.sec < pjsip_cfg()->tls.keep_alive_interval) {
 	/* There has been activity, so don't send keep-alive */
-	delay.sec = PJSIP_TLS_KEEP_ALIVE_INTERVAL - now.sec;
+	delay.sec = pjsip_cfg()->tls.keep_alive_interval - now.sec;
 	delay.msec = 0;
 
 	pjsip_endpt_schedule_timer(tls->base.endpt, &tls->ka_timer, 
@@ -1791,10 +2002,11 @@ static void tls_keep_alive_timer(pj_timer_heap_t *th, pj_timer_entry *e)
 	return;
     }
 
-    PJ_LOG(5,(tls->base.obj_name, "Sending %d byte(s) keep-alive to %.*s:%d", 
-	      (int)tls->ka_pkt.slen, (int)tls->base.remote_name.host.slen,
-	      tls->base.remote_name.host.ptr,
-	      tls->base.remote_name.port));
+    PJ_LOG(5,(tls->base.obj_name, "Sending %d byte(s) keep-alive to %s", 
+	      (int)tls->ka_pkt.slen, 
+	      pj_addr_str_print(&tls->base.remote_name.host, 
+				tls->base.remote_name.port, addr, 
+				sizeof(addr), 1)));
 
     /* Send the data */
     size = tls->ka_pkt.slen;
@@ -1810,7 +2022,7 @@ static void tls_keep_alive_timer(pj_timer_heap_t *th, pj_timer_entry *e)
     }
 
     /* Register next keep-alive */
-    delay.sec = PJSIP_TLS_KEEP_ALIVE_INTERVAL;
+    delay.sec = pjsip_cfg()->tls.keep_alive_interval;
     delay.msec = 0;
 
     pjsip_endpt_schedule_timer(tls->base.endpt, &tls->ka_timer, 
@@ -1818,4 +2030,30 @@ static void tls_keep_alive_timer(pj_timer_heap_t *th, pj_timer_entry *e)
     tls->ka_timer.id = PJ_TRUE;
 }
 
+
+static void wipe_buf(pj_str_t *buf)
+{
+    volatile char *p = buf->ptr;
+    pj_ssize_t len = buf->slen;
+    while (len--) *p++ = 0;
+    buf->slen = 0;
+}
+
+/*
+ * Wipe out certificates and keys in the TLS setting buffer.
+ */
+PJ_DEF(void) pjsip_tls_setting_wipe_keys(pjsip_tls_setting *opt)
+{
+    wipe_buf(&opt->ca_list_file);
+    wipe_buf(&opt->ca_list_path);
+    wipe_buf(&opt->cert_file);
+    wipe_buf(&opt->privkey_file);
+    wipe_buf(&opt->password);
+    wipe_buf(&opt->sigalgs);
+    wipe_buf(&opt->entropy_path);
+    wipe_buf(&opt->ca_buf);
+    wipe_buf(&opt->cert_buf);
+    wipe_buf(&opt->privkey_buf);    
+}
+
 #endif /* PJSIP_HAS_TLS_TRANSPORT */
