@@ -29,7 +29,200 @@ using namespace std;
 
 ///////////////////////////////////////////////////////////////////////////////
 
-void AccountRegConfig::readObject(const ContainerNode &node) throw(Error)
+void RtcpFbCap::fromPj(const pjmedia_rtcp_fb_cap &prm)
+{
+    this->codecId   = pj2Str(prm.codec_id);
+    this->type	    = prm.type;
+    this->typeName  = pj2Str(prm.type_name);
+    this->param	    = pj2Str(prm.param);
+}
+
+pjmedia_rtcp_fb_cap RtcpFbCap::toPj() const
+{
+    pjmedia_rtcp_fb_cap cap;
+
+    pj_bzero(&cap, sizeof(cap));
+    cap.codec_id    = str2Pj(this->codecId);
+    cap.type	    = this->type;
+    cap.type_name   = str2Pj(this->typeName);
+    cap.param	    = str2Pj(this->param);
+
+    return cap;
+}
+
+///////////////////////////////////////////////////////////////////////////////
+
+RtcpFbConfig::RtcpFbConfig()
+{
+    pjmedia_rtcp_fb_setting setting;
+    pjmedia_rtcp_fb_setting_default(&setting);
+    fromPj(setting);
+}
+
+void RtcpFbConfig::fromPj(const pjmedia_rtcp_fb_setting &prm)
+{
+    this->dontUseAvpf = PJ2BOOL(prm.dont_use_avpf);
+    this->caps.clear();
+    for (unsigned i = 0; i < prm.cap_count; ++i) {
+	RtcpFbCap cap;
+	cap.fromPj(prm.caps[i]);
+	this->caps.push_back(cap);
+    }
+}
+
+pjmedia_rtcp_fb_setting RtcpFbConfig::toPj() const
+{
+    pjmedia_rtcp_fb_setting setting;
+
+    pj_bzero(&setting, sizeof(setting));
+    setting.dont_use_avpf   = this->dontUseAvpf;
+    setting.cap_count	    = this->caps.size();
+    for (unsigned i = 0; i < setting.cap_count; ++i) {
+	setting.caps[i] = this->caps[i].toPj();
+    }
+
+    return setting;
+}
+
+void RtcpFbConfig::readObject(const ContainerNode &node) PJSUA2_THROW(Error)
+{
+    ContainerNode this_node = node.readContainer("RtcpFbConfig");
+    NODE_READ_BOOL	(this_node, dontUseAvpf);
+
+    ContainerNode cap_node = this_node.readArray("caps");
+    this->caps.clear();
+    while (cap_node.hasUnread()) {
+	RtcpFbCap cap;
+	NODE_READ_STRING	(cap_node, cap.codecId);
+	NODE_READ_NUM_T		(cap_node, pjmedia_rtcp_fb_type, cap.type);
+	NODE_READ_STRING	(cap_node, cap.typeName);
+	NODE_READ_STRING	(cap_node, cap.param);
+	this->caps.push_back(cap);
+    }
+}
+
+void RtcpFbConfig::writeObject(ContainerNode &node) const PJSUA2_THROW(Error)
+{
+    ContainerNode this_node = node.writeNewContainer("RtcpFbConfig");
+    NODE_WRITE_BOOL	(this_node, dontUseAvpf);
+
+    ContainerNode cap_node = this_node.writeNewArray("caps");
+    for (unsigned i=0; i<this->caps.size(); ++i) {
+	NODE_WRITE_STRING	(cap_node, this->caps[i].codecId);
+	NODE_WRITE_NUM_T	(cap_node, pjmedia_rtcp_fb_type,
+				 this->caps[i].type);
+	NODE_WRITE_STRING	(cap_node, this->caps[i].typeName);
+	NODE_WRITE_STRING	(cap_node, this->caps[i].param);
+    }
+}
+
+///////////////////////////////////////////////////////////////////////////////
+
+void SrtpCrypto::fromPj(const pjmedia_srtp_crypto &prm)
+{
+    this->key	    = pj2Str(prm.key);
+    this->name	    = pj2Str(prm.name);
+    this->flags	    = prm.flags;
+}
+
+pjmedia_srtp_crypto SrtpCrypto::toPj() const
+{
+    pjmedia_srtp_crypto crypto;
+    
+    crypto.key	    = str2Pj(this->key);
+    crypto.name	    = str2Pj(this->name);
+    crypto.flags    = this->flags;
+
+    return crypto;
+}
+
+///////////////////////////////////////////////////////////////////////////////
+
+SrtpOpt::SrtpOpt()
+{
+    pjsua_srtp_opt opt;
+    pjsua_srtp_opt_default(&opt);
+    fromPj(opt);
+}
+
+void SrtpOpt::fromPj(const pjsua_srtp_opt &prm)
+{
+    this->cryptos.clear();
+    for (unsigned i = 0; i < prm.crypto_count; ++i) {
+	SrtpCrypto crypto;
+	crypto.fromPj(prm.crypto[i]);
+	this->cryptos.push_back(crypto);
+    }
+
+    this->keyings.clear();
+    for (unsigned i = 0; i < prm.keying_count; ++i) {
+	this->keyings.push_back(prm.keying[i]);
+    }
+}
+
+pjsua_srtp_opt SrtpOpt::toPj() const
+{
+    pjsua_srtp_opt opt;
+
+    pj_bzero(&opt, sizeof(opt));
+
+    opt.crypto_count = this->cryptos.size();
+    for (unsigned i = 0; i < opt.crypto_count; ++i) {
+	opt.crypto[i] = this->cryptos[i].toPj();
+    }
+
+    opt.keying_count = this->keyings.size();
+    for (unsigned i = 0; i < opt.keying_count; ++i) {
+	opt.keying[i] = (pjmedia_srtp_keying_method)this->keyings[i];
+    }
+
+    return opt;
+}
+
+void SrtpOpt::readObject(const ContainerNode &node) PJSUA2_THROW(Error)
+{
+    ContainerNode this_node = node.readContainer("SrtpOpt");
+
+    ContainerNode crypto_node = this_node.readArray("cryptos");
+    this->cryptos.clear();
+    while (crypto_node.hasUnread()) {
+	SrtpCrypto crypto;
+	NODE_READ_STRING	(crypto_node, crypto.key);
+	NODE_READ_STRING	(crypto_node, crypto.name);
+	NODE_READ_UNSIGNED	(crypto_node, crypto.flags);
+	this->cryptos.push_back(crypto);
+    }
+
+    ContainerNode keying_node = this_node.readArray("keyings");
+    this->keyings.clear();
+    while (keying_node.hasUnread()) {
+	unsigned keying;
+	NODE_READ_UNSIGNED	(keying_node, keying);
+	this->keyings.push_back(keying);
+    }
+}
+
+void SrtpOpt::writeObject(ContainerNode &node) const PJSUA2_THROW(Error)
+{
+    ContainerNode this_node = node.writeNewContainer("SrtpOpt");
+
+    ContainerNode crypto_node = this_node.writeNewArray("cryptos");
+    for (unsigned i=0; i<this->cryptos.size(); ++i) {
+	NODE_WRITE_STRING	(crypto_node, this->cryptos[i].key);
+	NODE_WRITE_STRING	(crypto_node, this->cryptos[i].name);
+	NODE_WRITE_UNSIGNED	(crypto_node, this->cryptos[i].flags);
+    }
+
+    ContainerNode keying_node = this_node.writeNewArray("keyings");
+    for (unsigned i=0; i<this->keyings.size(); ++i) {
+	NODE_WRITE_UNSIGNED	(keying_node, this->keyings[i]);
+    }
+}
+
+///////////////////////////////////////////////////////////////////////////////
+
+void AccountRegConfig::readObject(const ContainerNode &node)
+				  PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("AccountRegConfig");
 
@@ -43,11 +236,13 @@ void AccountRegConfig::readObject(const ContainerNode &node) throw(Error)
     NODE_READ_BOOL	(this_node, dropCallsOnFail);
     NODE_READ_UNSIGNED	(this_node, unregWaitMsec);
     NODE_READ_UNSIGNED	(this_node, proxyUse);
+    NODE_READ_STRING	(this_node, contactParams);
 
     readSipHeaders(this_node, "headers", headers);
 }
 
-void AccountRegConfig::writeObject(ContainerNode &node) const throw(Error)
+void AccountRegConfig::writeObject(ContainerNode &node) const
+				   PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("AccountRegConfig");
 
@@ -61,13 +256,15 @@ void AccountRegConfig::writeObject(ContainerNode &node) const throw(Error)
     NODE_WRITE_BOOL	(this_node, dropCallsOnFail);
     NODE_WRITE_UNSIGNED	(this_node, unregWaitMsec);
     NODE_WRITE_UNSIGNED	(this_node, proxyUse);
+    NODE_WRITE_STRING	(this_node, contactParams);
 
     writeSipHeaders(this_node, "headers", headers);
 }
 
 ///////////////////////////////////////////////////////////////////////////////
 
-void AccountSipConfig::readObject(const ContainerNode &node) throw(Error)
+void AccountSipConfig::readObject(const ContainerNode &node)
+				  PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("AccountSipConfig");
 
@@ -88,7 +285,8 @@ void AccountSipConfig::readObject(const ContainerNode &node) throw(Error)
     }
 }
 
-void AccountSipConfig::writeObject(ContainerNode &node) const throw(Error)
+void AccountSipConfig::writeObject(ContainerNode &node) const
+				   PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("AccountSipConfig");
 
@@ -108,7 +306,8 @@ void AccountSipConfig::writeObject(ContainerNode &node) const throw(Error)
 
 ///////////////////////////////////////////////////////////////////////////////
 
-void AccountCallConfig::readObject(const ContainerNode &node) throw(Error)
+void AccountCallConfig::readObject(const ContainerNode &node)
+				   PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("AccountCallConfig");
 
@@ -119,7 +318,8 @@ void AccountCallConfig::readObject(const ContainerNode &node) throw(Error)
     NODE_READ_UNSIGNED( this_node, timerSessExpiresSec);
 }
 
-void AccountCallConfig::writeObject(ContainerNode &node) const throw(Error)
+void AccountCallConfig::writeObject(ContainerNode &node) const
+				    PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("AccountCallConfig");
 
@@ -132,7 +332,8 @@ void AccountCallConfig::writeObject(ContainerNode &node) const throw(Error)
 
 ///////////////////////////////////////////////////////////////////////////////
 
-void AccountPresConfig::readObject(const ContainerNode &node) throw(Error)
+void AccountPresConfig::readObject(const ContainerNode &node)
+				   PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("AccountPresConfig");
 
@@ -144,7 +345,8 @@ void AccountPresConfig::readObject(const ContainerNode &node) throw(Error)
     readSipHeaders(this_node, "headers", headers);
 }
 
-void AccountPresConfig::writeObject(ContainerNode &node) const throw(Error)
+void AccountPresConfig::writeObject(ContainerNode &node) const
+				    PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("AccountPresConfig");
 
@@ -158,7 +360,8 @@ void AccountPresConfig::writeObject(ContainerNode &node) const throw(Error)
 
 ///////////////////////////////////////////////////////////////////////////////
 
-void AccountMwiConfig::readObject(const ContainerNode &node) throw(Error)
+void AccountMwiConfig::readObject(const ContainerNode &node)
+				  PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("AccountMwiConfig");
 
@@ -166,7 +369,8 @@ void AccountMwiConfig::readObject(const ContainerNode &node) throw(Error)
     NODE_READ_UNSIGNED( this_node, expirationSec);
 }
 
-void AccountMwiConfig::writeObject(ContainerNode &node) const throw(Error)
+void AccountMwiConfig::writeObject(ContainerNode &node) const
+				   PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("AccountMwiConfig");
 
@@ -176,12 +380,14 @@ void AccountMwiConfig::writeObject(ContainerNode &node) const throw(Error)
 
 ///////////////////////////////////////////////////////////////////////////////
 
-void AccountNatConfig::readObject(const ContainerNode &node) throw(Error)
+void AccountNatConfig::readObject(const ContainerNode &node)
+				  PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("AccountNatConfig");
 
     NODE_READ_NUM_T   ( this_node, pjsua_stun_use, sipStunUse);
     NODE_READ_NUM_T   ( this_node, pjsua_stun_use, mediaStunUse);
+    NODE_READ_NUM_T   ( this_node, pjsua_nat64_opt, nat64Opt);
     NODE_READ_BOOL    ( this_node, iceEnabled);
     NODE_READ_INT     ( this_node, iceMaxHostCands);
     NODE_READ_BOOL    ( this_node, iceAggressiveNomination);
@@ -207,12 +413,14 @@ void AccountNatConfig::readObject(const ContainerNode &node) throw(Error)
     NODE_READ_INT     ( this_node, contactUseSrcPort);
 }
 
-void AccountNatConfig::writeObject(ContainerNode &node) const throw(Error)
+void AccountNatConfig::writeObject(ContainerNode &node) const
+				   PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("AccountNatConfig");
 
     NODE_WRITE_NUM_T   ( this_node, pjsua_stun_use, sipStunUse);
     NODE_WRITE_NUM_T   ( this_node, pjsua_stun_use, mediaStunUse);
+    NODE_WRITE_NUM_T   ( this_node, pjsua_nat64_opt, nat64Opt);
     NODE_WRITE_BOOL    ( this_node, iceEnabled);
     NODE_WRITE_INT     ( this_node, iceMaxHostCands);
     NODE_WRITE_BOOL    ( this_node, iceAggressiveNomination);
@@ -240,7 +448,8 @@ void AccountNatConfig::writeObject(ContainerNode &node) const throw(Error)
 
 ///////////////////////////////////////////////////////////////////////////////
 
-void AccountMediaConfig::readObject(const ContainerNode &node) throw(Error)
+void AccountMediaConfig::readObject(const ContainerNode &node)
+				    PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("AccountMediaConfig");
 
@@ -248,11 +457,14 @@ void AccountMediaConfig::readObject(const ContainerNode &node) throw(Error)
     NODE_READ_BOOL    ( this_node, streamKaEnabled);
     NODE_READ_NUM_T   ( this_node, pjmedia_srtp_use, srtpUse);
     NODE_READ_INT     ( this_node, srtpSecureSignaling);
+    NODE_READ_OBJ     ( this_node, srtpOpt);
     NODE_READ_NUM_T   ( this_node, pjsua_ipv6_use, ipv6Use);
     NODE_READ_OBJ     ( this_node, transportConfig);
+    NODE_READ_BOOL    ( this_node, rtcpMuxEnabled);
 }
 
-void AccountMediaConfig::writeObject(ContainerNode &node) const throw(Error)
+void AccountMediaConfig::writeObject(ContainerNode &node) const
+				     PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("AccountMediaConfig");
 
@@ -260,36 +472,71 @@ void AccountMediaConfig::writeObject(ContainerNode &node) const throw(Error)
     NODE_WRITE_BOOL    ( this_node, streamKaEnabled);
     NODE_WRITE_NUM_T   ( this_node, pjmedia_srtp_use, srtpUse);
     NODE_WRITE_INT     ( this_node, srtpSecureSignaling);
+    NODE_WRITE_OBJ     ( this_node, srtpOpt);
     NODE_WRITE_NUM_T   ( this_node, pjsua_ipv6_use, ipv6Use);
     NODE_WRITE_OBJ     ( this_node, transportConfig);
+    NODE_WRITE_BOOL    ( this_node, rtcpMuxEnabled);
 }
 
 ///////////////////////////////////////////////////////////////////////////////
 
-void AccountVideoConfig::readObject(const ContainerNode &node) throw(Error)
+void AccountVideoConfig::readObject(const ContainerNode &node)
+				    PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("AccountVideoConfig");
 
     NODE_READ_BOOL    ( this_node, autoShowIncoming);
     NODE_READ_BOOL    ( this_node, autoTransmitOutgoing);
     NODE_READ_UNSIGNED( this_node, windowFlags);
-    NODE_READ_NUM_T   ( this_node, pjmedia_vid_dev_index, defaultCaptureDevice);
-    NODE_READ_NUM_T   ( this_node, pjmedia_vid_dev_index, defaultRenderDevice);
-    NODE_READ_NUM_T   ( this_node, pjmedia_vid_stream_rc_method, rateControlMethod);
+    NODE_READ_NUM_T   ( this_node, pjmedia_vid_dev_index,
+			defaultCaptureDevice);
+    NODE_READ_NUM_T   ( this_node, pjmedia_vid_dev_index,
+			defaultRenderDevice);
+    NODE_READ_NUM_T   ( this_node, pjmedia_vid_stream_rc_method,
+			rateControlMethod);
     NODE_READ_UNSIGNED( this_node, rateControlBandwidth);
+    NODE_READ_UNSIGNED( this_node, startKeyframeCount);
+    NODE_READ_UNSIGNED( this_node, startKeyframeInterval);
 }
 
-void AccountVideoConfig::writeObject(ContainerNode &node) const throw(Error)
+void AccountVideoConfig::writeObject(ContainerNode &node) const
+				     PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("AccountVideoConfig");
 
     NODE_WRITE_BOOL    ( this_node, autoShowIncoming);
     NODE_WRITE_BOOL    ( this_node, autoTransmitOutgoing);
     NODE_WRITE_UNSIGNED( this_node, windowFlags);
-    NODE_WRITE_NUM_T   ( this_node, pjmedia_vid_dev_index, defaultCaptureDevice);
-    NODE_WRITE_NUM_T   ( this_node, pjmedia_vid_dev_index, defaultRenderDevice);
-    NODE_WRITE_NUM_T   ( this_node, pjmedia_vid_stream_rc_method, rateControlMethod);
+    NODE_WRITE_NUM_T   ( this_node, pjmedia_vid_dev_index,
+			 defaultCaptureDevice);
+    NODE_WRITE_NUM_T   ( this_node, pjmedia_vid_dev_index,
+			 defaultRenderDevice);
+    NODE_WRITE_NUM_T   ( this_node, pjmedia_vid_stream_rc_method,
+			 rateControlMethod);
     NODE_WRITE_UNSIGNED( this_node, rateControlBandwidth);
+    NODE_WRITE_UNSIGNED( this_node, startKeyframeCount);
+    NODE_WRITE_UNSIGNED( this_node, startKeyframeInterval);
+}
+///////////////////////////////////////////////////////////////////////////////
+
+void AccountIpChangeConfig::readObject(const ContainerNode &node)
+				       PJSUA2_THROW(Error)
+{
+    ContainerNode this_node = node.readContainer("AccountIpChangeConfig");
+
+    NODE_READ_BOOL    ( this_node, shutdownTp);
+    NODE_READ_BOOL    ( this_node, hangupCalls);
+    NODE_READ_UNSIGNED( this_node, reinviteFlags);
+}
+
+void AccountIpChangeConfig::writeObject(ContainerNode &node) const
+					PJSUA2_THROW(Error)
+{
+    ContainerNode this_node = node.writeNewContainer("AccountIpChangeConfig");
+
+    NODE_WRITE_BOOL    ( this_node, shutdownTp);
+    NODE_WRITE_BOOL    ( this_node, hangupCalls);
+    NODE_WRITE_UNSIGNED( this_node, reinviteFlags);
 }
 
 ///////////////////////////////////////////////////////////////////////////////
@@ -325,6 +572,7 @@ void AccountConfig::toPj(pjsua_acc_config &ret) const
     ret.drop_calls_on_reg_fail	= regConfig.dropCallsOnFail;
     ret.unreg_timeout		= regConfig.unregWaitMsec;
     ret.reg_use_proxy		= regConfig.proxyUse;
+    ret.reg_contact_params	= str2Pj(regConfig.contactParams);
     for (i=0; i<regConfig.headers.size(); ++i) {
 	pj_list_push_back(&ret.reg_hdr_list, &regConfig.headers[i].toPj());
     }
@@ -377,15 +625,22 @@ void AccountConfig::toPj(pjsua_acc_config &ret) const
     ret.unpublish_max_wait_time_msec = presConfig.publishShutdownWaitMsec;
     ret.pidf_tuple_id		= str2Pj(presConfig.pidfTupleId);
 
+    // AccountMwiConfig
+    ret.mwi_enabled 		= mwiConfig.enabled;
+    ret.mwi_expires 		= mwiConfig.expirationSec;
+
     // AccountNatConfig
     ret.sip_stun_use		= natConfig.sipStunUse;
     ret.media_stun_use		= natConfig.mediaStunUse;
+    ret.nat64_opt		= natConfig.nat64Opt;
     ret.ice_cfg_use		= PJSUA_ICE_CONFIG_USE_CUSTOM;
     ret.ice_cfg.enable_ice	= natConfig.iceEnabled;
     ret.ice_cfg.ice_max_host_cands = natConfig.iceMaxHostCands;
     ret.ice_cfg.ice_opt.aggressive = natConfig.iceAggressiveNomination;
-    ret.ice_cfg.ice_opt.nominated_check_delay = natConfig.iceNominatedCheckDelayMsec;
-    ret.ice_cfg.ice_opt.controlled_agent_want_nom_timeout = natConfig.iceWaitNominationTimeoutMsec;
+    ret.ice_cfg.ice_opt.nominated_check_delay =
+			    natConfig.iceNominatedCheckDelayMsec;
+    ret.ice_cfg.ice_opt.controlled_agent_want_nom_timeout =
+			    natConfig.iceWaitNominationTimeoutMsec;
     ret.ice_cfg.ice_no_rtcp	= natConfig.iceNoRtcp;
     ret.ice_cfg.ice_always_update = natConfig.iceAlwaysUpdate;
 
@@ -394,9 +649,12 @@ void AccountConfig::toPj(pjsua_acc_config &ret) const
     ret.turn_cfg.turn_server	= str2Pj(natConfig.turnServer);
     ret.turn_cfg.turn_conn_type	= natConfig.turnConnType;
     ret.turn_cfg.turn_auth_cred.type = PJ_STUN_AUTH_CRED_STATIC;
-    ret.turn_cfg.turn_auth_cred.data.static_cred.username = str2Pj(natConfig.turnUserName);
-    ret.turn_cfg.turn_auth_cred.data.static_cred.data_type = (pj_stun_passwd_type)natConfig.turnPasswordType;
-    ret.turn_cfg.turn_auth_cred.data.static_cred.data = str2Pj(natConfig.turnPassword);
+    ret.turn_cfg.turn_auth_cred.data.static_cred.username =
+			    str2Pj(natConfig.turnUserName);
+    ret.turn_cfg.turn_auth_cred.data.static_cred.data_type =
+			    (pj_stun_passwd_type)natConfig.turnPasswordType;
+    ret.turn_cfg.turn_auth_cred.data.static_cred.data =
+			    str2Pj(natConfig.turnPassword);
     ret.turn_cfg.turn_auth_cred.data.static_cred.realm = pj_str((char*)"");
     ret.turn_cfg.turn_auth_cred.data.static_cred.nonce = pj_str((char*)"");
 
@@ -419,7 +677,10 @@ void AccountConfig::toPj(pjsua_acc_config &ret) const
 #endif
     ret.use_srtp		= mediaConfig.srtpUse;
     ret.srtp_secure_signaling	= mediaConfig.srtpSecureSignaling;
+    ret.srtp_opt		= mediaConfig.srtpOpt.toPj();
     ret.ipv6_media_use		= mediaConfig.ipv6Use;
+    ret.enable_rtcp_mux		= mediaConfig.rtcpMuxEnabled;
+    ret.rtcp_fb_cfg		= mediaConfig.rtcpFbConfig.toPj();
 
     // AccountVideoConfig
     ret.vid_in_auto_show	= videoConfig.autoShowIncoming;
@@ -429,6 +690,13 @@ void AccountConfig::toPj(pjsua_acc_config &ret) const
     ret.vid_rend_dev		= videoConfig.defaultRenderDevice;
     ret.vid_stream_rc_cfg.method= videoConfig.rateControlMethod;
     ret.vid_stream_rc_cfg.bandwidth = videoConfig.rateControlBandwidth;
+    ret.vid_stream_sk_cfg.count = videoConfig.startKeyframeCount;
+    ret.vid_stream_sk_cfg.interval = videoConfig.startKeyframeInterval;
+
+    // AccountIpChangeConfig
+    ret.ip_change_cfg.shutdown_tp = ipChangeConfig.shutdownTp;
+    ret.ip_change_cfg.hangup_calls = ipChangeConfig.hangupCalls;
+    ret.ip_change_cfg.reinvite_flags = ipChangeConfig.reinviteFlags;
 }
 
 /* Initialize from pjsip. */
@@ -453,6 +721,7 @@ void AccountConfig::fromPj(const pjsua_acc_config &prm,
     regConfig.dropCallsOnFail	= PJ2BOOL(prm.drop_calls_on_reg_fail);
     regConfig.unregWaitMsec	= prm.unreg_timeout;
     regConfig.proxyUse		= prm.reg_use_proxy;
+    regConfig.contactParams	= pj2Str(prm.reg_contact_params);
     regConfig.headers.clear();
     hdr = prm.reg_hdr_list.next;
     while (hdr != &prm.reg_hdr_list) {
@@ -520,12 +789,16 @@ void AccountConfig::fromPj(const pjsua_acc_config &prm,
     // AccountNatConfig
     natConfig.sipStunUse	= prm.sip_stun_use;
     natConfig.mediaStunUse	= prm.media_stun_use;
+    natConfig.nat64Opt		= prm.nat64_opt;
     if (prm.ice_cfg_use == PJSUA_ICE_CONFIG_USE_CUSTOM) {
 	natConfig.iceEnabled = PJ2BOOL(prm.ice_cfg.enable_ice);
 	natConfig.iceMaxHostCands = prm.ice_cfg.ice_max_host_cands;
-	natConfig.iceAggressiveNomination = PJ2BOOL(prm.ice_cfg.ice_opt.aggressive);
-	natConfig.iceNominatedCheckDelayMsec = prm.ice_cfg.ice_opt.nominated_check_delay;
-	natConfig.iceWaitNominationTimeoutMsec = prm.ice_cfg.ice_opt.controlled_agent_want_nom_timeout;
+	natConfig.iceAggressiveNomination =
+			PJ2BOOL(prm.ice_cfg.ice_opt.aggressive);
+	natConfig.iceNominatedCheckDelayMsec =
+			prm.ice_cfg.ice_opt.nominated_check_delay;
+	natConfig.iceWaitNominationTimeoutMsec =
+			prm.ice_cfg.ice_opt.controlled_agent_want_nom_timeout;
 	natConfig.iceNoRtcp	= PJ2BOOL(prm.ice_cfg.ice_no_rtcp);
 	natConfig.iceAlwaysUpdate = PJ2BOOL(prm.ice_cfg.ice_always_update);
     } else {
@@ -537,8 +810,10 @@ void AccountConfig::fromPj(const pjsua_acc_config &prm,
 	natConfig.iceEnabled	= PJ2BOOL(mcfg->enable_ice);
 	natConfig.iceMaxHostCands= mcfg->ice_max_host_cands;
 	natConfig.iceAggressiveNomination = PJ2BOOL(mcfg->ice_opt.aggressive);
-	natConfig.iceNominatedCheckDelayMsec = mcfg->ice_opt.nominated_check_delay;
-	natConfig.iceWaitNominationTimeoutMsec = mcfg->ice_opt.controlled_agent_want_nom_timeout;
+	natConfig.iceNominatedCheckDelayMsec =
+			mcfg->ice_opt.nominated_check_delay;
+	natConfig.iceWaitNominationTimeoutMsec =
+			mcfg->ice_opt.controlled_agent_want_nom_timeout;
 	natConfig.iceNoRtcp	= PJ2BOOL(mcfg->ice_no_rtcp);
 	natConfig.iceAlwaysUpdate = PJ2BOOL(mcfg->ice_always_update);
     }
@@ -547,9 +822,12 @@ void AccountConfig::fromPj(const pjsua_acc_config &prm,
 	natConfig.turnEnabled	= PJ2BOOL(prm.turn_cfg.enable_turn);
 	natConfig.turnServer	= pj2Str(prm.turn_cfg.turn_server);
 	natConfig.turnConnType	= prm.turn_cfg.turn_conn_type;
-	natConfig.turnUserName	= pj2Str(prm.turn_cfg.turn_auth_cred.data.static_cred.username);
-	natConfig.turnPasswordType = prm.turn_cfg.turn_auth_cred.data.static_cred.data_type;
-	natConfig.turnPassword	= pj2Str(prm.turn_cfg.turn_auth_cred.data.static_cred.data);
+	natConfig.turnUserName	=
+		pj2Str(prm.turn_cfg.turn_auth_cred.data.static_cred.username);
+	natConfig.turnPasswordType =
+		prm.turn_cfg.turn_auth_cred.data.static_cred.data_type;
+	natConfig.turnPassword	=
+		pj2Str(prm.turn_cfg.turn_auth_cred.data.static_cred.data);
     } else {
 	pjsua_media_config default_mcfg;
 	if (!mcfg) {
@@ -559,9 +837,12 @@ void AccountConfig::fromPj(const pjsua_acc_config &prm,
 	natConfig.turnEnabled	= PJ2BOOL(mcfg->enable_turn);
 	natConfig.turnServer	= pj2Str(mcfg->turn_server);
 	natConfig.turnConnType	= mcfg->turn_conn_type;
-	natConfig.turnUserName	= pj2Str(mcfg->turn_auth_cred.data.static_cred.username);
-	natConfig.turnPasswordType = mcfg->turn_auth_cred.data.static_cred.data_type;
-	natConfig.turnPassword	= pj2Str(mcfg->turn_auth_cred.data.static_cred.data);
+	natConfig.turnUserName	=
+		pj2Str(mcfg->turn_auth_cred.data.static_cred.username);
+	natConfig.turnPasswordType =
+		mcfg->turn_auth_cred.data.static_cred.data_type;
+	natConfig.turnPassword	=
+		pj2Str(mcfg->turn_auth_cred.data.static_cred.data);
     }
     natConfig.contactRewriteUse	= prm.allow_contact_rewrite;
     natConfig.contactRewriteMethod = prm.contact_rewrite_method;
@@ -584,7 +865,10 @@ void AccountConfig::fromPj(const pjsua_acc_config &prm,
 #endif
     mediaConfig.srtpUse		= prm.use_srtp;
     mediaConfig.srtpSecureSignaling = prm.srtp_secure_signaling;
+    mediaConfig.srtpOpt.fromPj(prm.srtp_opt);
     mediaConfig.ipv6Use		= prm.ipv6_media_use;
+    mediaConfig.rtcpMuxEnabled	= PJ2BOOL(prm.enable_rtcp_mux);
+    mediaConfig.rtcpFbConfig.fromPj(prm.rtcp_fb_cfg);
 
     // AccountVideoConfig
     videoConfig.autoShowIncoming 	= PJ2BOOL(prm.vid_in_auto_show);
@@ -594,9 +878,16 @@ void AccountConfig::fromPj(const pjsua_acc_config &prm,
     videoConfig.defaultRenderDevice	= prm.vid_rend_dev;
     videoConfig.rateControlMethod	= prm.vid_stream_rc_cfg.method;
     videoConfig.rateControlBandwidth	= prm.vid_stream_rc_cfg.bandwidth;
+    videoConfig.startKeyframeCount	= prm.vid_stream_sk_cfg.count;
+    videoConfig.startKeyframeInterval	= prm.vid_stream_sk_cfg.interval;
+
+    // AccountIpChangeConfig
+    ipChangeConfig.shutdownTp = PJ2BOOL(prm.ip_change_cfg.shutdown_tp);
+    ipChangeConfig.hangupCalls = PJ2BOOL(prm.ip_change_cfg.hangup_calls);
+    ipChangeConfig.reinviteFlags = prm.ip_change_cfg.reinvite_flags;
 }
 
-void AccountConfig::readObject(const ContainerNode &node) throw(Error)
+void AccountConfig::readObject(const ContainerNode &node) PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("AccountConfig");
 
@@ -612,7 +903,7 @@ void AccountConfig::readObject(const ContainerNode &node) throw(Error)
     NODE_READ_OBJ     ( this_node, videoConfig);
 }
 
-void AccountConfig::writeObject(ContainerNode &node) const throw(Error)
+void AccountConfig::writeObject(ContainerNode &node) const PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("AccountConfig");
 
@@ -659,20 +950,11 @@ Account::~Account()
     /* If this instance is deleted, also delete the corresponding account in
      * PJSUA library.
      */
-    if (isValid() && pjsua_get_state() < PJSUA_STATE_CLOSING) {
-        // Cleanup buddies in the buddy list
-	while(buddyList.size() > 0) {
-	    Buddy *b = buddyList[0];
-	    delete b; /* this will remove itself from the list */
-	}
-
-	pjsua_acc_set_user_data(id, NULL);
-	pjsua_acc_del(id);
-    }
+    shutdown();
 }
 
 void Account::create(const AccountConfig &acc_cfg,
-                     bool make_default) throw(Error)
+                     bool make_default) PJSUA2_THROW(Error)
 {
     pjsua_acc_config pj_acc_cfg;
     
@@ -681,7 +963,24 @@ void Account::create(const AccountConfig &acc_cfg,
     PJSUA2_CHECK_EXPR( pjsua_acc_add(&pj_acc_cfg, make_default, &id) );
 }
 
-void Account::modify(const AccountConfig &acc_cfg) throw(Error)
+void Account::shutdown()
+{
+    if (isValid() && pjsua_get_state() < PJSUA_STATE_CLOSING) {
+        // Cleanup buddies in the buddy list
+	while(buddyList.size() > 0) {
+	    Buddy *b = buddyList[0];
+	    delete b; /* this will remove itself from the list */
+	}
+
+	// This caused error message of "Error: cannot find Account.."
+	// when Endpoint::on_reg_started() is called for unregistration.
+	//pjsua_acc_set_user_data(id, NULL);
+
+	pjsua_acc_del(id);
+    }
+}
+
+void Account::modify(const AccountConfig &acc_cfg) PJSUA2_THROW(Error)
 {
     pjsua_acc_config pj_acc_cfg;
     
@@ -695,7 +994,7 @@ bool Account::isValid() const
     return pjsua_acc_is_valid(id) != 0;
 }
 
-void Account::setDefault() throw(Error)
+void Account::setDefault() PJSUA2_THROW(Error)
 {
     PJSUA2_CHECK_EXPR( pjsua_acc_set_default(id) );
 }
@@ -715,7 +1014,7 @@ Account *Account::lookup(int acc_id)
     return (Account*)pjsua_acc_get_user_data(acc_id);
 }
 
-AccountInfo Account::getInfo() const throw(Error)
+AccountInfo Account::getInfo() const PJSUA2_THROW(Error)
 {
     pjsua_acc_info pj_ai;
     AccountInfo ai;
@@ -725,13 +1024,13 @@ AccountInfo Account::getInfo() const throw(Error)
     return ai;
 }
 
-void Account::setRegistration(bool renew) throw(Error)
+void Account::setRegistration(bool renew) PJSUA2_THROW(Error)
 {
     PJSUA2_CHECK_EXPR( pjsua_acc_set_registration(id, renew) );
 }
 
 void
-Account::setOnlineStatus(const PresenceStatus &pres_st) throw(Error)
+Account::setOnlineStatus(const PresenceStatus &pres_st) PJSUA2_THROW(Error)
 {
     pjrpid_element pj_rpid;
 
@@ -746,12 +1045,12 @@ Account::setOnlineStatus(const PresenceStatus &pres_st) throw(Error)
 			    &pj_rpid) );
 }
 
-void Account::setTransport(TransportId tp_id) throw(Error)
+void Account::setTransport(TransportId tp_id) PJSUA2_THROW(Error)
 {
     PJSUA2_CHECK_EXPR( pjsua_acc_set_transport(id, tp_id) );
 }
 
-void Account::presNotify(const PresNotifyParam &prm) throw(Error)
+void Account::presNotify(const PresNotifyParam &prm) PJSUA2_THROW(Error)
 {
     pj_str_t pj_state_str   = str2Pj(prm.stateStr);
     pj_str_t pj_reason	    = str2Pj(prm.reason);
@@ -764,13 +1063,27 @@ void Account::presNotify(const PresNotifyParam &prm) throw(Error)
 					 &msg_data) );
 }
 
-const BuddyVector& Account::enumBuddies() const throw(Error)
+const BuddyVector& Account::enumBuddies() const PJSUA2_THROW(Error)
 {
     return buddyList;
 }
 
+BuddyVector2 Account::enumBuddies2() const PJSUA2_THROW(Error)
+{
+    BuddyVector2 bv2;
+    pjsua_buddy_id ids[PJSUA_MAX_BUDDIES];
+    unsigned i, count = PJSUA_MAX_BUDDIES;
+
+    PJSUA2_CHECK_EXPR( pjsua_enum_buddies(ids, &count) );
+    for (i = 0; i < count; ++i) {
+	bv2.push_back(Buddy(ids[i]));
+    }
+
+    return bv2;
+}
+
 Buddy* Account::findBuddy(string uri, FindBuddyMatch *buddy_match) const
-		throw(Error)
+		PJSUA2_THROW(Error)
 {
     if (!buddy_match) {
 	static FindBuddyMatch def_bm;
@@ -784,6 +1097,22 @@ Buddy* Account::findBuddy(string uri, FindBuddyMatch *buddy_match) const
     PJSUA2_RAISE_ERROR(PJ_ENOTFOUND);
 }
 
+Buddy Account::findBuddy2(string uri) const PJSUA2_THROW(Error)
+{
+    pj_str_t pj_uri;
+    pjsua_buddy_id bud_id;
+
+    pj_strset2(&pj_uri, (char*)uri.c_str());
+
+    bud_id = pjsua_buddy_find(&pj_uri);
+    if (id == PJSUA_INVALID_ID) {
+	PJSUA2_RAISE_ERROR(PJ_ENOTFOUND);
+    }
+
+    Buddy buddy(bud_id);
+    return buddy;
+}
+
 void Account::addBuddy(Buddy *buddy)
 {
     pj_assert(buddy);
