@@ -30,7 +30,7 @@ namespace pj
 {
 void readIntVector( ContainerNode &node,
                     const string &array_name,
-                    IntVector &v) throw(Error)
+                    IntVector &v) PJSUA2_THROW(Error)
 {
     ContainerNode array_node = node.readArray(array_name);
     v.resize(0);
@@ -41,7 +41,7 @@ void readIntVector( ContainerNode &node,
 
 void writeIntVector(ContainerNode &node,
                     const string &array_name,
-                    const IntVector &v) throw(Error)
+                    const IntVector &v) PJSUA2_THROW(Error)
 {
     ContainerNode array_node = node.writeNewArray(array_name);
     for (unsigned i=0; i<v.size(); ++i) {
@@ -50,7 +50,7 @@ void writeIntVector(ContainerNode &node,
 }
 
 void readQosParams( ContainerNode &node,
-                    pj_qos_params &qos) throw(Error)
+                    pj_qos_params &qos) PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("qosParams");
 
@@ -61,7 +61,7 @@ void readQosParams( ContainerNode &node,
 }
 
 void writeQosParams( ContainerNode &node,
-                     const pj_qos_params &qos) throw(Error)
+                     const pj_qos_params &qos) PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("qosParams");
 
@@ -73,7 +73,7 @@ void writeQosParams( ContainerNode &node,
 
 void readSipHeaders( const ContainerNode &node,
                      const string &array_name,
-                     SipHeaderVector &headers) throw(Error)
+                     SipHeaderVector &headers) PJSUA2_THROW(Error)
 {
     ContainerNode headers_node = node.readArray(array_name);
     headers.resize(0);
@@ -89,7 +89,7 @@ void readSipHeaders( const ContainerNode &node,
 
 void writeSipHeaders(ContainerNode &node,
                      const string &array_name,
-                     const SipHeaderVector &headers) throw(Error)
+                     const SipHeaderVector &headers) PJSUA2_THROW(Error)
 {
     ContainerNode headers_node = node.writeNewArray(array_name);
     for (unsigned i=0; i<headers.size(); ++i) {
@@ -117,7 +117,7 @@ AuthCredInfo::AuthCredInfo(const string &param_scheme,
 {
 }
 
-void AuthCredInfo::readObject(const ContainerNode &node) throw(Error)
+void AuthCredInfo::readObject(const ContainerNode &node) PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("AuthCredInfo");
 
@@ -131,7 +131,7 @@ void AuthCredInfo::readObject(const ContainerNode &node) throw(Error)
     NODE_READ_STRING( this_node, akaAmf);
 }
 
-void AuthCredInfo::writeObject(ContainerNode &node) const throw(Error)
+void AuthCredInfo::writeObject(ContainerNode &node) const PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("AuthCredInfo");
 
@@ -163,6 +163,9 @@ pjsip_tls_setting TlsConfig::toPj() const
     ts.cert_file	= str2Pj(this->certFile);
     ts.privkey_file	= str2Pj(this->privKeyFile);
     ts.password		= str2Pj(this->password);
+    ts.ca_buf		= str2Pj(this->CaBuf);
+    ts.cert_buf		= str2Pj(this->certBuf);
+    ts.privkey_buf	= str2Pj(this->privKeyBuf);
     ts.method		= this->method;
     ts.ciphers_num	= (unsigned)this->ciphers.size();
     ts.proto		= this->proto;
@@ -188,6 +191,9 @@ void TlsConfig::fromPj(const pjsip_tls_setting &prm)
     this->certFile 	= pj2Str(prm.cert_file);
     this->privKeyFile 	= pj2Str(prm.privkey_file);
     this->password 	= pj2Str(prm.password);
+    this->CaBuf		= pj2Str(prm.ca_buf);
+    this->certBuf	= pj2Str(prm.cert_buf);
+    this->privKeyBuf	= pj2Str(prm.privkey_buf);
     this->method 	= (pjsip_ssl_method)prm.method;
     this->proto 	= prm.proto;
     // The following will only work if sizeof(enum)==sizeof(int)
@@ -202,7 +208,7 @@ void TlsConfig::fromPj(const pjsip_tls_setting &prm)
     this->qosIgnoreError = PJ2BOOL(prm.qos_ignore_error);
 }
 
-void TlsConfig::readObject(const ContainerNode &node) throw(Error)
+void TlsConfig::readObject(const ContainerNode &node) PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("TlsConfig");
 
@@ -210,6 +216,9 @@ void TlsConfig::readObject(const ContainerNode &node) throw(Error)
     NODE_READ_STRING  ( this_node, certFile);
     NODE_READ_STRING  ( this_node, privKeyFile);
     NODE_READ_STRING  ( this_node, password);
+    NODE_READ_STRING  ( this_node, CaBuf);
+    NODE_READ_STRING  ( this_node, certBuf);
+    NODE_READ_STRING  ( this_node, privKeyBuf);
     NODE_READ_NUM_T   ( this_node, pjsip_ssl_method, method);
     readIntVector     ( this_node, "ciphers", ciphers);
     NODE_READ_BOOL    ( this_node, verifyServer);
@@ -221,7 +230,7 @@ void TlsConfig::readObject(const ContainerNode &node) throw(Error)
     NODE_READ_BOOL    ( this_node, qosIgnoreError);
 }
 
-void TlsConfig::writeObject(ContainerNode &node) const throw(Error)
+void TlsConfig::writeObject(ContainerNode &node) const PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("TlsConfig");
 
@@ -229,6 +238,9 @@ void TlsConfig::writeObject(ContainerNode &node) const throw(Error)
     NODE_WRITE_STRING  ( this_node, certFile);
     NODE_WRITE_STRING  ( this_node, privKeyFile);
     NODE_WRITE_STRING  ( this_node, password);
+    NODE_WRITE_STRING  ( this_node, CaBuf);
+    NODE_WRITE_STRING  ( this_node, certBuf);
+    NODE_WRITE_STRING  ( this_node, privKeyBuf);
     NODE_WRITE_NUM_T   ( this_node, pjsip_ssl_method, method);
     writeIntVector     ( this_node, "ciphers", ciphers);
     NODE_WRITE_BOOL    ( this_node, verifyServer);
@@ -276,7 +288,7 @@ pjsua_transport_config TransportConfig::toPj() const
     return tc;
 }
 
-void TransportConfig::readObject(const ContainerNode &node) throw(Error)
+void TransportConfig::readObject(const ContainerNode &node) PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.readContainer("TransportConfig");
 
@@ -289,7 +301,8 @@ void TransportConfig::readObject(const ContainerNode &node) throw(Error)
     NODE_READ_OBJ       ( this_node, tlsConfig);
 }
 
-void TransportConfig::writeObject(ContainerNode &node) const throw(Error)
+void TransportConfig::writeObject(ContainerNode &node) const
+				  PJSUA2_THROW(Error)
 {
     ContainerNode this_node = node.writeNewContainer("TransportConfig");
 
@@ -304,24 +317,29 @@ void TransportConfig::writeObject(ContainerNode &node) const throw(Error)
 
 ///////////////////////////////////////////////////////////////////////////////
 
-void TransportInfo::fromPj(const pjsua_transport_info &info)
+TransportInfo::TransportInfo()
+: id(), type(PJSIP_TRANSPORT_UNSPECIFIED), flags(), usageCount()
 {
-    this->id = info.id;
-    this->type = info.type;
-    this->typeName = pj2Str(info.type_name);
-    this->info = pj2Str(info.info);
-    this->flags = info.flag;
+}
+
+void TransportInfo::fromPj(const pjsua_transport_info &tinfo)
+{
+    this->id = tinfo.id;
+    this->type = tinfo.type;
+    this->typeName = pj2Str(tinfo.type_name);
+    this->info = pj2Str(tinfo.info);
+    this->flags = tinfo.flag;
 
     char straddr[PJ_INET6_ADDRSTRLEN+10];
-    pj_sockaddr_print(&info.local_addr, straddr, sizeof(straddr), 3);
+    pj_sockaddr_print(&tinfo.local_addr, straddr, sizeof(straddr), 3);
     this->localAddress = straddr;
 
     pj_ansi_snprintf(straddr, sizeof(straddr), "%.*s:%d",
-                     (int)info.local_name.host.slen,
-                     info.local_name.host.ptr,
-                     info.local_name.port);
+                     (int)tinfo.local_name.host.slen,
+                     tinfo.local_name.host.ptr,
+                     tinfo.local_name.port);
     this->localName = straddr;
-    this->usageCount = info.usage_count;
+    this->usageCount = tinfo.usage_count;
 }
 
 ///////////////////////////////////////////////////////////////////////////////
@@ -361,18 +379,38 @@ pjsip_media_type SipMediaType::toPj() const
 
 ///////////////////////////////////////////////////////////////////////////////
 
-void SipHeader::fromPj(const pjsip_hdr *hdr) throw(Error)
+void SipHeader::fromPj(const pjsip_hdr *hdr) PJSUA2_THROW(Error)
 {
-    char buf[256];
+    char *buf = NULL;
+    int len = 0;
+    unsigned buf_size = 256>>1;
 
-    int len = pjsip_hdr_print_on((void*)hdr, buf, sizeof(buf)-1);
-    if (len <= 0)
+    /* Print header to a 256 bytes buffer first.
+     * If buffer is not sufficient, try 512, 1024, soon
+     * until > PJSIP_MAX_PKT_LEN
+     */
+    do {
+        buf_size <<= 1;
+	buf = (char*)malloc(buf_size);
+	if (!buf)
+	    PJSUA2_RAISE_ERROR(PJ_ENOMEM);
+
+	len = pjsip_hdr_print_on((void*)hdr, buf, buf_size-1);
+        if (len < 0)
+            free(buf);
+
+    } while ((buf_size < PJSIP_MAX_PKT_LEN) && (len < 0));
+    
+    if (len < 0)
 	PJSUA2_RAISE_ERROR(PJ_ETOOSMALL);
+
     buf[len] = '\0';
 
     char *pos = strchr(buf, ':');
-    if (!pos)
+    if (!pos) {
+	free(buf);
 	PJSUA2_RAISE_ERROR(PJSIP_EINVALIDHDR);
+    }
 
     // Trim white space after header name
     char *end_name = pos;
@@ -384,6 +422,7 @@ void SipHeader::fromPj(const pjsip_hdr *hdr) throw(Error)
 
     hName = string(buf, end_name);
     hValue = string(start_val);
+    free(buf);
 }
 
 pjsip_generic_string_hdr &SipHeader::toPj() const
@@ -397,7 +436,8 @@ pjsip_generic_string_hdr &SipHeader::toPj() const
 
 ///////////////////////////////////////////////////////////////////////////////
 
-void SipMultipartPart::fromPj(const pjsip_multipart_part &prm) throw(Error)
+void SipMultipartPart::fromPj(const pjsip_multipart_part &prm)
+			      PJSUA2_THROW(Error)
 {
     headers.clear();
     pjsip_hdr* pj_hdr = prm.hdr.next;
@@ -495,7 +535,7 @@ void SipTxData::fromPj(pjsip_tx_data &tdata)
     
     info	= pjsip_tx_data_get_info(&tdata);
     pjsip_tx_data_encode(&tdata);
-    wholeMsg	= string(tdata.buf.start, tdata.buf.end - tdata.buf.start);
+    wholeMsg	= string(tdata.buf.start, tdata.buf.cur - tdata.buf.start);
     if (pj_sockaddr_has_addr(&tdata.tp_info.dst_addr)) {
 	pj_sockaddr_print(&tdata.tp_info.dst_addr, straddr, sizeof(straddr), 3);
 	dstAddress  = straddr;
@@ -506,7 +546,7 @@ void SipTxData::fromPj(pjsip_tx_data &tdata)
 }
 
 SipTransaction::SipTransaction()
-: role(PJSIP_ROLE_UAC), statusCode(0), pjTransaction(NULL)
+: role(PJSIP_ROLE_UAC), statusCode(PJSIP_SC_NULL), state(PJSIP_TSX_STATE_NULL), pjTransaction(NULL)
 {
 }
 
@@ -516,6 +556,7 @@ void SipTransaction::fromPj(pjsip_transaction &tsx)
     this->method        = pj2Str(tsx.method.name);
     this->statusCode    = tsx.status_code;
     this->statusText    = pj2Str(tsx.status_text);
+    this->state		= tsx.state;
     if (tsx.last_tx)
 	this->lastTx.fromPj(*tsx.last_tx);
     else
@@ -523,6 +564,11 @@ void SipTransaction::fromPj(pjsip_transaction &tsx)
     this->pjTransaction = (void *)&tsx;
 }
 
+TsxStateEvent::TsxStateEvent()
+: prevState(PJSIP_TSX_STATE_NULL), type(PJSIP_EVENT_UNKNOWN)
+{
+}
+
 bool SipTxOption::isEmpty() const
 {
     return (targetUri == "" && headers.size() == 0 && contentType == "" &&
@@ -530,7 +576,7 @@ bool SipTxOption::isEmpty() const
             multipartContentType.subType == "" && multipartParts.size() == 0);
 }
 
-void SipTxOption::fromPj(const pjsua_msg_data &prm) throw(Error)
+void SipTxOption::fromPj(const pjsua_msg_data &prm) PJSUA2_THROW(Error)
 {
     targetUri = pj2Str(prm.target_uri);
 
